#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_arrow.h"


/******************************************************************************
  arrow_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

arrow_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      arrow_help (Console->formats[CONTROL_ARROW]);
      break;

    case FUNCTION_FORMAT:
      arrow_format (Console->formats[CONTROL_ARROW]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_ARROW], line);
      break;

    case FUNCTION_STORE:
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      arrow_initialize (widget);
      break;

    case FUNCTION_PARSE:
      arrow_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      break;

    case FUNCTION_SETUP:
      arrow_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      arrow_execute (widget);
      break;

    case FUNCTION_DRAW:
      arrow_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to arrow", NULL);
      break;
  }
}



/******************************************************************************
  arrow_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

arrow_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("ARROW\n");
  printf ("    ALTER   RIGHT                  right facing arrow\n");
  printf ("    ALTER   LEFT                   left facing arrow\n");
  printf ("    ALTER   UP                     up facing arrow\n");
  printf ("    ALTER   DOWN                   down facing arrow\n");
  printf ("    ALTER   SIZE    <int>          size of arrow (varies)\n");
  printf ("    ALTER   THICK   <int>          thickness of arrow shadow (%d)\n",
			format->data[ARROW_D_THICKNESS]);
}



/******************************************************************************
  arrow_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

arrow_format (format)
  Widget *format;
{
  /** CHECK: Probably want it to be motion active at some point **/

  format->status[IS_SELECTABLE] =	TRUE;
  format->status[IS_ACTIVATABLE] =	TRUE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	FALSE;
  format->status[IS_TIMER_ACTIVE] =	TRUE;
  format->status[IS_BUFFERABLE] =	BUFFER_NORMAL;

  format->size.x =			DEFAULT_ARROW_DIMENSION_X;
  format->size.y =			DEFAULT_ARROW_DIMENSION_Y;
  format->spacing.x =			DEFAULT_ARROW_SPACING_X;
  format->spacing.y =			DEFAULT_ARROW_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	1;

  format->gcs[GC_BODY] =		COLOR_CTRL_SURFACE;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_SURFACE;
  format->gcs[ARROW_GC_ARROW] =		COLOR_CTRL_BACK;
  format->gcs[ARROW_GC_ARROW_INV] =	COLOR_CTRL_SURFACE_INV;

  format->data[ARROW_D_DIRECTION] =	ARROW_DIRECTION_RIGHT;
  format->data[ARROW_D_THICKNESS] =	2;
  format->data[ARROW_D_DIMENSION] =	-1;

  format->format = CONTROL_ARROW;
}



/******************************************************************************
  arrow_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

arrow_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_ARROW;
  widget->subformat = CONTROL_NONE;
}



/******************************************************************************
  arrow_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

arrow_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100];

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (tres_line);
  int_factor = atoi (tres_line);

  if (line[0] == 'A')
  {
    switch (dos_line[0])
    {
      /** RIGHT **/

      case 'R':

        widget->data[ARROW_D_DIRECTION] = ARROW_DIRECTION_RIGHT;
        break;

      /** LEFT **/

      case 'L':

        widget->data[ARROW_D_DIRECTION] = ARROW_DIRECTION_LEFT;
        break;

      /** UP **/

      case 'U':

        widget->data[ARROW_D_DIRECTION] = ARROW_DIRECTION_UP;
        break;

      /** DOWN **/

      case 'D':

        widget->data[ARROW_D_DIRECTION] = ARROW_DIRECTION_DOWN;
        break;

      /** THICK **/

      case 'T':

        widget->data[ARROW_D_THICKNESS] = int_factor;
        break;

      /** SIZE **/

      case 'S':

        widget->data[ARROW_D_DIMENSION] = int_factor;
        break;
    }
    line[0] = '#';
  }
}



/******************************************************************************
  arrow_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

arrow_setup (widget)
  Widget *widget;
{
  int i,
      dimension,
      thickness,
      verbose_offset, verbose_box;

  /** Set base dimension **/

  if (widget->size.x > widget->size.y)
    dimension = widget->size.y-2;
  else
    dimension = widget->size.x-2;

  /** Make sure dimension is odd **/

  if (dimension%2 == 0)
    dimension -= 1;

  if (widget->data[ARROW_D_DIMENSION] <= 0)
    widget->data[ARROW_D_DIMENSION] = dimension;
  else
    dimension = widget->data[ARROW_D_DIMENSION];

  thickness = widget->data[ARROW_D_THICKNESS];

  /** Based on direction, determine x and y arrow sizes **/

  switch (widget->data[ARROW_D_DIRECTION])
  {
    case ARROW_DIRECTION_RIGHT:
      widget->text_offset[0].x = widget->size.x/2 - dimension/2;
      widget->text_offset[0].y = widget->size.y/2 - dimension/2;

      widget->text_offset[1].x = widget->size.x/2 - dimension/2;
      widget->text_offset[1].y = widget->size.y/2 + dimension/2 -
					thickness + 1;
      break;

    case ARROW_DIRECTION_LEFT:
      widget->text_offset[0].x = widget->size.x/2 + dimension/2 - 2;
      widget->text_offset[0].y = widget->size.y/2 - dimension/2;

      widget->text_offset[1].x = widget->size.x/2 + dimension/2 - 2;
      widget->text_offset[1].y = widget->size.y/2 + dimension/2 -
					thickness + 1;
      break;

    case ARROW_DIRECTION_UP:
      widget->text_offset[0].x = widget->size.x/2 - dimension/2;
      widget->text_offset[0].y = widget->size.y/2 + dimension/2 - 2;

      widget->text_offset[1].x = widget->size.x/2 + dimension/2 -
					thickness + 1;
      widget->text_offset[1].y = widget->size.y/2 + dimension/2 - 2;
      break;

    case ARROW_DIRECTION_DOWN:
      widget->text_offset[0].x = widget->size.x/2 - dimension/2;
      widget->text_offset[0].y = widget->size.y/2 - dimension/2;

      widget->text_offset[1].x = widget->size.x/2 + dimension/2 -
					thickness + 1;
      widget->text_offset[1].y = widget->size.y/2 - dimension/2;
      break;
  }

  /** Set arrow bounds **/

  widget->bound_offset.x = widget->thickness[THICKNESS_SHADOW] +
		widget->thickness[THICKNESS_ACTIVE];
  widget->bound_offset.y = widget->thickness[THICKNESS_SHADOW] +
		widget->thickness[THICKNESS_ACTIVE];
  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;
  widget->bound_size.y = 2*widget->bound_offset.y + widget->size.y;
}



/******************************************************************************
  arrow_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

arrow_execute (widget)
  Widget *widget;
{
  widget->flag = TRUE;
}



/******************************************************************************
  arrow_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

arrow_draw (widget)
  Widget *widget;
{
  short i,
        color_arrow,
        color_arrow_inv,
        color_top,
        color_bottom,
        color_temp,
        xoffset,
        yoffset,
        thickness,
        temp_thickness,
        dimension;

  XPoint start,
         end;

  /** Set some convenience variables **/

  color_arrow = widget->gcs[ARROW_GC_ARROW];
  thickness = widget->data[ARROW_D_THICKNESS];
  dimension = widget->data[ARROW_D_DIMENSION];

  /** Draw box in which arrow will be drawn **/

  draw_generic_box (widget, BULGE_IN, BULGE_KEEP, ACTIVE_SHOW);

  /** Set colors for appropriate raised/inset appearance **/

  if (widget->mode == MODE_SELECT)
  {
    color_bottom = COLOR_CTRL_HIGH;
    color_top = COLOR_CTRL_SHADOW;
  }
  else
  {
    color_top = COLOR_CTRL_HIGH;
    color_bottom = COLOR_CTRL_SHADOW;
  }

  /** Draw arrow of specified direction **/ 

  switch (widget->data[ARROW_D_DIRECTION])
  {
    case ARROW_DIRECTION_RIGHT:
    case ARROW_DIRECTION_LEFT:

      if (widget->data[ARROW_D_DIRECTION] == ARROW_DIRECTION_RIGHT)
        xoffset = 2;
      else
        xoffset = -2;

      start.x = widget->position.x + widget->text_offset[0].x;
      start.y = widget->position.y + widget->text_offset[0].y;

      /** Draw arrow inards **/

      for (i=0; i<dimension/2; i++)
      {
        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_arrow],
			start.x+xoffset*i, start.y+i, 2, dimension-2*i);
      }

      /** Draw top highlight **/

      for (i=0; i<thickness-1; i++)
      {
        /** kludge to fix alignment error of unknown origin **/

        if (widget->data[ARROW_D_DIRECTION] == ARROW_DIRECTION_LEFT)
          XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x + i + 1, start.y + i/2,
			1, dimension - 2*(i/2));
        else
          XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x + i, start.y + i/2,
			1, dimension - 2*(i/2));
      }

      /** Draw side highlight **/

      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
       if (i+temp_thickness >= dimension/2+2)
          temp_thickness--;

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x, start.y, 2, temp_thickness);

        start.x += xoffset;
        start.y += 1;
      }

      /** Draw bottom shadow **/

      start.x = widget->position.x + widget->text_offset[1].x;
      start.y = widget->position.y + widget->text_offset[1].y;

      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness > dimension/2)
        {
          temp_thickness--;
          start.y++;
        }

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x, start.y, 2, temp_thickness);
        start.x += xoffset;
        start.y -= 1;
      }
      break;

    case ARROW_DIRECTION_UP:
    case ARROW_DIRECTION_DOWN:

      if (widget->data[ARROW_D_DIRECTION] == ARROW_DIRECTION_DOWN)
      {
        yoffset = 2;
      }
      else
      {
        color_temp = color_bottom;
        color_bottom = color_top;
        color_top = color_temp;

        yoffset = -2;
      }

      start.x = widget->position.x + widget->text_offset[0].x;
      start.y = widget->position.y + widget->text_offset[0].y;

      /** Draw arrow innards **/

      for (i=0; i<dimension/2; i++)
      {
        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_arrow],
			start.x+i, start.y+yoffset*i, dimension-2*i, 2);
      }

      /** Draw side highlight **/

      for (i=0; i<thickness-1; i++)
        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x + i/2, start.y + i,
			dimension - 2*(i/2), 1);

      /** Draw bottom shadow **/

      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness >= dimension/2+2)
          temp_thickness--;

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x, start.y, temp_thickness, 2);
        start.x += 1;
        start.y += yoffset;
      }

      start.x = widget->position.x + widget->text_offset[1].x;
      start.y = widget->position.y + widget->text_offset[1].y;

      /** Draw side shadow **/

      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness > dimension/2)
        {
          temp_thickness--;
          start.x++;
        }

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x, start.y, temp_thickness, 2);
        start.x -= 1;
        start.y += yoffset;
      }
      break;
  }
}
