#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_bar.h"


/******************************************************************************
  bar_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

bar_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      bar_help (Console->formats[CONTROL_BAR]);
      break;

    case FUNCTION_FORMAT:
      bar_format (Console->formats[CONTROL_BAR]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_BAR], line);
      break;

    case FUNCTION_STORE:
      bar_store (widget, line);
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      bar_initialize (widget);
      break;

    case FUNCTION_PARSE:
      bar_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      bar_clean (widget);
      break;

    case FUNCTION_SETUP:
      bar_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      bar_select (widget);
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      bar_execute (widget);
      break;

    case FUNCTION_DRAW:
      bar_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to bar", NULL);
      break;
  }
}


/******************************************************************************
  bar_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

bar_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("BAR\n");
  printf ("    MIN    <float>                 minimum value\n");
  printf ("    MAX    <float>                 maximum value\n");
  printf ("    INIT   <float>                 initial value\n");
  printf ("    ALTER   LENGTH  <int>          width of vertical bar (%d)\n",
			format->data[BAR_D_LENGTH]);
  printf ("    ALTER   THICK   <int>          width of vertical bar (%d)\n",
			format->data[BAR_D_THICK]);
  printf ("    ALTER   COLOR   <gc>           gc index for bar color (%d)\n",
			format->gcs[BAR_G_BAR]);
  printf ("    ALTER   CLASSIC                use classic format\n");
  printf ("\n"); 
}



/******************************************************************************
  bar_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

bar_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	TRUE;
  format->status[IS_ACTIVATABLE] =	TRUE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	TRUE;
  format->status[IS_BUFFERABLE] =	BUFFER_NORMAL;

  format->size.x =			DEFAULT_BAR_DIMENSION_X;
  format->size.y =			DEFAULT_BAR_DIMENSION_Y;
  format->spacing.x =  			DEFAULT_BAR_SPACING_X;
  format->spacing.y =		   	DEFAULT_BAR_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	1;

  format->data[BAR_D_THICK] =		DEFAULT_BAR_THICK;
  format->data[BAR_D_LENGTH] =		DEFAULT_BAR_LENGTH;

  format->gcs[GC_BODY] =		COLOR_CTRL_AUX_2;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_DARK_INV;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_DARK;
  format->gcs[BAR_G_BAR] =		COLOR_CTRL_SURFACE;

  format->format =			CONTROL_BAR;
}



/******************************************************************************
  bar_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

bar_store (widget, line)
  Widget *widget;
  char *line;
{
  char subline[100];

  line[0] = '\0';
  sprintf (subline, "INIT\t%f\n", widget->value);
  strcat (line, subline);
}



/******************************************************************************
  bar_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

bar_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_BAR;
}



/******************************************************************************
  bar_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

bar_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100],
       *ptr,
       ch;  

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (tres_line);
  int_factor = atoi (tres_line);
 
  if (line[0] == 'A')
  {
    switch (dos_line[0])
    {
      /** LENGTH **/

      case 'L':

        widget->data[BAR_D_LENGTH] = int_factor;
        break;

      /** THICK **/

      case 'T':

        widget->data[BAR_D_THICK] = int_factor;
        break;

      /** COLOR **/

      case 'C':

        widget->gcs[BAR_G_BAR] = int_factor;
        break;
    }

    line[0] = '#';
  }
}



/******************************************************************************
  bar_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

bar_clean (widget)
  Widget *widget;
{
  if (widget->label == NULL)
    widget->label = strdup (widget->name);
}



/******************************************************************************
  bar_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

bar_setup (widget)
  Widget *widget;
{
  int height,
      limit1, limit2;

  char line[100],
       number[20];

  widget->bound_offset.x = widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];
  widget->bound_offset.y = widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];

  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;
  widget->bound_size.y = 2*widget->bound_offset.y + widget->size.y;
}



/******************************************************************************
  bar_select (widget)

  Handle type-specific function of widget <widget> when it is first selected
  by the user.
******************************************************************************/

bar_select (widget)
  Widget *widget;
{
  int xpoint,
      bar_width,
      half_bar_width,
      active_length;

  bar_width = widget->data[BAR_D_LENGTH];
  half_bar_width = bar_width/2;
  active_length = widget->size.x - bar_width;

  xpoint = (int)((widget->value - widget->minimum)*(active_length)/
		(widget->maximum - widget->minimum) + half_bar_width + 0.5);

  if (Win->mouse.x - widget->position.x < xpoint-half_bar_width ||
	Win->mouse.x - widget->position.x > xpoint+half_bar_width)
    widget->points[BAR_P_OFFSET].x = 0;
  else
  {
    widget->points[BAR_P_OFFSET].x = 
	Win->mouse.x - widget->position.x - xpoint;
  }
}



/******************************************************************************
  bar_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

bar_execute (widget)
  Widget *widget;
{
  int i, j,
      bar_width,
      half_bar_width,
      active_length;

  double value;

  widget->flag = TRUE;
  bar_width = widget->data[BAR_D_LENGTH];
  half_bar_width = bar_width/2;
  active_length = widget->size.x - bar_width;

  value = widget->value;

  if (Win->mouse.x - (widget->position.x+half_bar_width) -
			widget->points[BAR_P_OFFSET].x > active_length)
    value = widget->maximum;
  else
    value = (widget->maximum - widget->minimum)*
		(Win->mouse.x - (widget->position.x + half_bar_width) -
		widget->points[BAR_P_OFFSET].x)/
		(active_length) + widget->minimum;
 
  Master->last_value = value;
 
  widget->value = value;
  widget->values[0] = value;
}



/******************************************************************************
  bar_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

bar_draw (widget)
  Widget *widget;
{
  int i, xpoint;

  short color_bar,
        bar_width,
        half_bar_width,
        shadow_thickness,
        active_length;

  if (widget->enable == FALSE)
    return;

  color_bar = widget->gcs[BAR_G_BAR];

  bar_width = widget->data[BAR_D_LENGTH];
  shadow_thickness = widget->data[BAR_D_THICK];
  half_bar_width = bar_width/2;
  active_length = widget->size.x - bar_width;

  draw_generic_box (widget, BULGE_IN, BULGE_KEEP, ACTIVE_SHOW);

  xpoint = (int)((widget->value - widget->minimum)*(active_length)/
		(widget->maximum - widget->minimum) + 0.5);

  if (xpoint < 0)
    xpoint = 0;
  else if (xpoint > active_length)
    xpoint = active_length;

  XFillRectangle (Win->display, widget->drawable,
	Win->hue_widget[color_bar],
	widget->position.x + xpoint,
	widget->position.y,
	bar_width, widget->size.y);

  draw_border (widget->drawable,
	widget->position.x + xpoint + shadow_thickness,
	widget->position.y + shadow_thickness,
	bar_width-2*shadow_thickness, widget->size.y-2*shadow_thickness,
	shadow_thickness,
	Win->hue_widget[COLOR_CTRL_HIGH],
	Win->hue_widget[COLOR_CTRL_SHADOW]);
}
