#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_button.h"


/******************************************************************************
  button_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

button_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      button_help (Console->formats[CONTROL_BUTTON]);
      break;

    case FUNCTION_FORMAT:
      button_format (Console->formats[CONTROL_BUTTON]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_BUTTON], line);
      break;

    case FUNCTION_STORE:
      button_store (widget, line);
      break;

    case FUNCTION_DUMP:
      button_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      button_initialize (widget);
      break;

    case FUNCTION_PARSE:
      button_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      button_clean (widget);
      break;

    case FUNCTION_SETUP:
      button_setup (widget);
      break;

    case FUNCTION_EDIT:
      button_edit (widget, line);
      break;

    case FUNCTION_SELECT:
      button_select (widget);
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      button_execute (widget);
      break;

    case FUNCTION_DRAW:
      button_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to button", NULL);
      break;
  }
}



/******************************************************************************
  button_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

button_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("BUTTON\n");
  printf ("    TEXT   <string>                text for state 1\n");
  printf ("      :\n"); 
  printf ("    TEXT   <string>                text for state n\n");
  printf ("    INIT   <int>                   initial state\n");
  printf (
 "    PIXMAP  <int><a,b> <file>      <file> is <a,b> bitmap for state <int>\n");
  printf (
 "    GRAPHIC <int>      <file>      <file> is pixmap for state <int>\n");
  printf ("    VERBOSE SQUARE                 checkbox style, square in box\n");
  printf ("            CROSS                                  cross in box\n");
  printf ("            CHECK                                  check in box\n");
  printf ("            NONE                                   no boxes\n");
  printf ("    EXPLIC  <as above>             checkbox style, multiple on\n");
  printf ("    ALTER   OFFSET  <int>          distance from box to text (%d)\n",
			format->data[BUTTON_D_CHECK_OFFSET]);
  printf ("    ALTER   BOX     <int>          size of check box (%d)\n",
			format->data[BUTTON_D_CHECK_BOX]);
  printf ("    ALTER   SPACE   <float>        vertical text spacing (%4.2f)\n",
			format->data_float[BUTTON_DF_TEXT_SPACING]);
  printf ("    ALTER   CHECK   <gc>           index of check gc (%d)\n",
			format->gcs[BUTTON_G_CHECK]);
  printf ("    ALTER   TOGGLE                 use in/out toggle button\n");
  printf ("\n");
}



/******************************************************************************
  button_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

button_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	TRUE;
  format->status[IS_ACTIVATABLE] =	TRUE;
  format->status[IS_EDITABLE] =		EDITABLE_PARTIAL;
  format->status[IS_MOTION_ACTIVE] =	FALSE;
  format->status[IS_BUFFERABLE] =	BUFFER_NORMAL;

  format->size.x =			DEFAULT_BUTTON_DIMENSION_X;
  format->size.y =			DEFAULT_BUTTON_DIMENSION_Y;
  format->spacing.x =			DEFAULT_BUTTON_SPACING_X;
  format->spacing.y =			DEFAULT_BUTTON_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	3;
  format->thickness[THICKNESS_ACTIVE] =	2;

  format->gcs[GC_BODY] =		COLOR_CTRL_SURFACE;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_SURFACE_INV;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_TEXT_1;
  format->gcs[GC_FONT_SECOND] =		COLOR_CTRL_TEXT_1;
  format->gcs[BUTTON_G_CHECK] =		COLOR_CTRL_TEXT_3;

  format->fonts[FONT_FIRST] =		FONT_MEDIUM;
  format->fonts[FONT_SECOND] =		FONT_MEDIUM;

  format->data[BUTTON_D_CHECK_OFFSET] =	DEFAULT_BUTTON_CHECK_OFFSET;
  format->data[BUTTON_D_CHECK_BOX] =	DEFAULT_BUTTON_CHECK_BOX;
  format->data[BUTTON_D_CHECK] =	CHECK_SQUARE;

  format->data_float[BUTTON_DF_TEXT_SPACING] =
					DEFAULT_BUTTON_TEXT_SPACING;

  format->format =			CONTROL_BUTTON;
}



/******************************************************************************
  button_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

button_store (widget, line)
  Widget *widget;
  char *line;
{
  int i;
  char subline[100];

  line[0] = '\0';
  if (widget->subformat == CONTROL_BUTTON_EXPLICIT)
  {
    for (i=0; i<widget->state_count; i++)
    {
      sprintf (subline, "INIT\tSTATE\t%d\t%d\n", i, widget->states[i]);
      strcat (line, subline);
    }
  }
  else
  {
    sprintf (subline, "INIT\t%d\n", widget->state);
    strcat (line, subline);
  }
}



/******************************************************************************
  button_dump (widget)

  Dump type-specific information for widget <widget> to stdout for the
  purposes of debugging.
******************************************************************************/

button_dump (widget)
  Widget *widget;
{
  int i;

  generic_dump (widget);

  printf ("  states: %d\n", widget->state_count);
  for (i=0; i<widget->state_count; i++)
    printf ("    %d: %s\n", i, widget->labels[i]);
}



/******************************************************************************
  button_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

button_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_BUTTON;
  widget->subformat = CONTROL_NONE;
}



/******************************************************************************
  button_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

button_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100],
       ch;

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (tres_line);
  int_factor = atoi (tres_line);

  switch (line[0])
  {
    /** ALTER **/

    case 'A':

      switch (dos_line[0])
      {
        /** OFFSET **/

        case 'O':

          widget->data[BUTTON_D_CHECK_OFFSET] = int_factor;
          break;

        /** BOX **/

        case 'B':

          widget->data[BUTTON_D_CHECK_BOX] = int_factor;
          break;

        /** SPACE **/

        case 'S':

          widget->data_float[BUTTON_DF_TEXT_SPACING] = factor;
          break;

        /** CHECK **/

        case 'C':

          widget->gcs[BUTTON_G_CHECK] = int_factor;
          break;

        /** TOGGLE **/

        case 'T':

          widget->subformat = CONTROL_BUTTON_TOGGLE;
          break;
      }
      line[0] = '#';
      break;

    /** EXPLICIT or VERBOSE **/

    case 'E':
    case 'e':
    case 'V':
    case 'v':

      widget->gcs[GC_FONT_SECOND] = COLOR_CTRL_TEXT_2;

      if (line[0] == 'E' || line[0] == 'e')
        widget->subformat = CONTROL_BUTTON_EXPLICIT;
      else
        widget->subformat = CONTROL_BUTTON_VERBOSE;
      widget->status[IS_EDITABLE] = EDITABLE_NEVER;

      strip_first (line);
      switch (line[0])
      {
        /** SQUARE **/

        case 'S':
        case 's':

          widget->data[BUTTON_D_CHECK] = CHECK_SQUARE;
          break;

        /** X (cross) **/

        case 'X':
        case 'x':

          widget->data[BUTTON_D_CHECK] = CHECK_CROSS;
          break;

        /** CHECK **/

        case 'C':
        case 'c':

          widget->data[BUTTON_D_CHECK] = CHECK_CHECK;
          break;

        /** NONE **/

        case 'N':
        case 'n':

          widget->data[BUTTON_D_CHECK] = CHECK_NONE;
          break;

        default:

          break;
      }
      line[0] = '#';
      break;
  }
}



/******************************************************************************
  button_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

button_clean (widget)
  Widget *widget;
{
  double text_spacing;
  short font_0;

  if (!widget->state_count)
    printf ("BUTTON: unspecified number of states\n");

  if (widget->graphic_count)
    widget->status[IS_EDITABLE] = EDITABLE_NEVER;
}



/******************************************************************************
  button_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

button_setup (widget)
  Widget *widget;
{
  int i, height,
      verbose_offset, verbose_box;

  double text_spacing;
  short font_0;


  if (widget->graphic_count)
  {
    load_graphics (widget);
  }
  else
  {
    font_0 = widget->fonts[FONT_FIRST];

    /** If verbose or explicit, must figure out text spacings **/

    if (widget->subformat == CONTROL_BUTTON_VERBOSE || 
		widget->subformat == CONTROL_BUTTON_EXPLICIT)
    {
      text_spacing = widget->data_float[BUTTON_DF_TEXT_SPACING];
      font_0 = widget->fonts[FONT_0];

      /** If flagged by main program, set widget height **/

      if (widget->user)
      {
        widget->size.y +=
		(int)(((Win->fonts[font_0]->dimension.y) * text_spacing) *
		(widget->state_count-1) + 0.5);
        widget->user = FALSE;
      }

      verbose_offset = widget->data[BUTTON_D_CHECK_OFFSET];
      verbose_box = widget->data[BUTTON_D_CHECK_BOX];
      height = Win->fonts[font_0]->dimension.y;

      /** For each state, set text offsets **/

      for (i=0; i<widget->state_count; i++)
      {
        widget->text_offset[i].x = 2*verbose_offset + verbose_box;

        widget->text_offset[i].y = (int)((widget->size.y -
		((widget->state_count-1)*height*text_spacing +
		height))/2.0 + height + height*text_spacing*i + 0.5);

        flush_text (widget->labels[i],
			&(widget->text_offset[i].x), font_0, TRUE);
      }
    }
    else
      for (i=0; i<widget->state_count; i++)
        center_text (0, 0, widget->size.x, widget->size.y, widget->labels[i],
		&(widget->text_offset[i].x), &(widget->text_offset[i].y),
		font_0);
  }

  /** If toggle button **/

  if (widget->subformat == CONTROL_BUTTON_TOGGLE)
  {
    widget->status[IS_EDITABLE] = EDITABLE_NEVER;
    widget->state_count = 2;
    widget->text_offset[1].x = widget->text_offset[0].x+1;
    widget->text_offset[1].y = widget->text_offset[0].y+1;
  }

  /** Set widget bounds **/

  widget->bound_offset.x = widget->thickness[THICKNESS_SHADOW] +
		widget->thickness[THICKNESS_ACTIVE];
  widget->bound_offset.y = widget->thickness[THICKNESS_SHADOW] +
		widget->thickness[THICKNESS_ACTIVE];
  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;
  widget->bound_size.y = 2*widget->bound_offset.y + widget->size.y;
}



/******************************************************************************
  button_edit (widget, line)

  Manage editing of widget <widget>, based on input string <line>.  The
  internal value <widget->reserve> determines editing status.
******************************************************************************/

button_edit (widget, line)
  Widget *widget;
  char *line;
{
  static short state;

  switch (widget->reserve)
  {
    case EDIT_BEGIN:

      state = widget->state;
      widget->reserve = EDIT_PROCESS;
      edit_text (widget);
      break;

    case EDIT_PROCESS:

      if (widget->labels[state] != NULL)
        free (widget->labels[state]);
      widget->labels[state] = strdup (line);
      set_label (widget, line);
      center_text (0, 0, widget->size.x, widget->size.y, line,
		&(widget->text_offset[state].x),
		&(widget->text_offset[state].y),
		widget->fonts[FONT_FIRST]);
      break;

    case EDIT_FINISH:

      if (widget->labels[state] != NULL)
        free (widget->labels[state]);
      widget->labels[state] = strdup (line);
      set_label (widget, line); 
      center_text (0, 0, widget->size.x, widget->size.y, line,
		&(widget->text_offset[state].x),
		&(widget->text_offset[state].y),
		widget->fonts[FONT_FIRST]);
      widget->reserve = EDIT_BEGIN;
      break;
  }
}



/******************************************************************************
  button_select (widget)

  Handle type-specific function of widget <widget> when it is first selected
  by the user.
******************************************************************************/

button_select (widget)
  Widget *widget;
{
  short i,
        check_box,
	box_size;

  XPoint offset,
         cursor;

  /** If explicit or verbose **/

  if (widget->subformat == CONTROL_BUTTON_VERBOSE ||
		widget->subformat == CONTROL_BUTTON_EXPLICIT)
  {
    offset.x = widget->data[BUTTON_D_CHECK_OFFSET];
    box_size = widget->data[BUTTON_D_CHECK_BOX];

    check_box = -1;

    cursor.x = widget->cursor.x - widget->position.x;
    cursor.y = widget->cursor.y - widget->position.y;

    /** Determine which check box cursor is in **/

    if (offset.x <= cursor.x && offset.x + box_size >= cursor.x)
    {
      offset.y = Win->fonts[widget->fonts[FONT_FIRST]]->dimension.y +
		(box_size-Win->fonts[widget->fonts[FONT_FIRST]]->dimension.y)/2;

      for (i=0; i<widget->state_count; i++)
      {
        if (widget->text_offset[i].y - offset.y <= cursor.y &&
		widget->text_offset[i].y - offset.y + box_size >= cursor.y)
          check_box = i;
      }
    }

    if (check_box != -1)   
    {
      widget->next_state = check_box;
      return;
    }
    else if (widget->subformat == CONTROL_BUTTON_EXPLICIT)
    {
      widget->next_state = -1;
      return;
    }
    /** Else verbose, treat like normal button (increment/decrement) **/
  }

  /** Increment or decrement button state **/

  if (Win->event.xbutton.button == Button1)
  {
    widget->next_state = widget->state + 1;
    if (widget->next_state >= widget->state_count)
      widget->next_state = 0;
  }
  else if (Win->event.xbutton.button == Button2 ||
			Win->event.xbutton.button == Button3)
  {
    widget->next_state = widget->state - 1;
    if (widget->next_state < 0)
      widget->next_state = widget->state_count-1;
  }
}



/******************************************************************************
  button_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

button_execute (widget)
  Widget *widget;
{
  if (widget->next_state < 0)
    return;
   
  if (widget->subformat == CONTROL_BUTTON_EXPLICIT)
    widget->states[widget->next_state] =
		 !(widget->states[widget->next_state]);
  else
    widget->state = widget->next_state;
  widget->flag = TRUE;
}



/******************************************************************************
  button_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

button_draw (widget)
  Widget *widget;
{
  int i, k,
      height,
      font_gc,
      gc_index,
      mode,
      state_selected,
      state_on,
      verbose_offset, verbose_box,
      hold_shadow;

  XPoint offset;

  GC use_gc;

  Graphic *graphic;

  short color_font_0,
	color_font_1,
	shadow_thickness,
	label_state,
        font_0,
        font_1;

  if (widget->enable == FALSE)
    return;

  /** Set convenience variables **/

  color_font_0 = widget->gcs[GC_FONT_FIRST];
  color_font_1 = widget->gcs[GC_FONT_SECOND];
  font_0 = widget->fonts[FONT_FIRST];
  font_1 = widget->fonts[FONT_SECOND];
  hold_shadow = widget->thickness[THICKNESS_SHADOW];

  label_state = widget->state;

  state_selected = -1;
  if (widget->subformat == CONTROL_BUTTON_VERBOSE ||
		widget->subformat == CONTROL_BUTTON_EXPLICIT)
  {
    if (widget->mode == MODE_SELECT)
      state_selected = widget->next_state;
  }

  /** Draw button base **/

  if (widget->subformat == CONTROL_BUTTON_TOGGLE && widget->state)
  {
    draw_generic_box (widget, BULGE_IN, BULGE_SWAP, ACTIVE_SHOW);
    label_state = 0;
  }
  else
    draw_generic_box (widget, BULGE_OUT, BULGE_SWAP, ACTIVE_SHOW);

  /** If verbose or explicit **/

  if (widget->subformat == CONTROL_BUTTON_VERBOSE ||
		widget->subformat == CONTROL_BUTTON_EXPLICIT)
  {
    verbose_offset = widget->data[BUTTON_D_CHECK_OFFSET];
    verbose_box = widget->data[BUTTON_D_CHECK_BOX];

    widget->thickness[THICKNESS_SHADOW] = 1;

    height = Win->fonts[font_0]->dimension.y;
    offset.x = verbose_offset;
    offset.y = -height - (verbose_box - height)/2;

    if (widget->mode == MODE_SELECT && hold_shadow)
    {
      offset.x += 1;
      offset.y += 1;
    }

    for (i=0; i<widget->state_count; i++)
    {
      if (widget->subformat == CONTROL_BUTTON_EXPLICIT)
      {
        if (widget->states[i])
          state_on = TRUE;
        else
          state_on = FALSE;
      }
      else
      {
        if (widget->state == i)
          state_on = TRUE;
        else
          state_on = FALSE;
      }

      /** Determine which color to use **/

      if (widget->data[BUTTON_D_CHECK] == CHECK_NONE)
        use_gc = Win->hue_widget[color_font_0];
      else if (state_on)
        use_gc = Win->hue_widget[color_font_0];
      else
        use_gc = Win->hue_widget[color_font_1];

      /** Draw text **/

      if (widget->mode == MODE_NORMAL ||
			widget->mode == MODE_ACTIVE || !hold_shadow)
        draw_string (widget->drawable,
		widget->position.x, widget->position.y,
		widget->text_offset[i].x, widget->text_offset[i].y,
		widget->labels[i], font_0, use_gc, FALSE);
      else if (widget->mode == MODE_SELECT)
        draw_string (widget->drawable,
		widget->position.x, widget->position.y,
		widget->text_offset[i].x+1, widget->text_offset[i].y+1,
		widget->labels[i], font_0, use_gc, FALSE);

      /** Draw checkbox **/

      if (widget->data[BUTTON_D_CHECK] != CHECK_NONE)
      {
        button_draw_subbutton (widget, state_on,
		widget->position.x + offset.x,
		widget->position.y + widget->text_offset[i].y + offset.y,
		verbose_box, verbose_box, (i==state_selected),
		widget->drawable);
      }
    }
    widget->thickness[THICKNESS_SHADOW] = hold_shadow;
  }

  /** Else normal button **/

  else
  {
    if (widget->subformat == CONTROL_BUTTON_TOGGLE && widget->state)
    {
      offset.x = -1;
      offset.y = -1;
    }
    else
    {
      offset.x = 1;
      offset.y = 1;
    }

    /** If button uses graphics, draw them **/

    if (widget->graphic_count)
    {
      if (widget->mode != MODE_SELECT || !hold_shadow)
      {
        offset.x = 0;
        offset.y = 0;
      }

      graphic = widget->graphics[label_state];
      if (graphic->is_bitmap)
      {
        XCopyPlane (Win->display, graphic->pixmap, widget->drawable,
			Win->hue_widget[graphic->gc],
			0, 0, graphic->width, graphic->height,
			widget->position.x +
			widget->text_offset[widget->state].x+offset.x,
			widget->position.y +
			widget->text_offset[widget->state].y+offset.y, 1L);
      }
      else
      {
        XCopyArea (Win->display, graphic->pixmap, widget->drawable,
			Win->hue_widget[0],
			0, 0, graphic->width, graphic->height,
			widget->position.x +
			widget->text_offset[widget->state].x+offset.x,
			widget->position.y +
			widget->text_offset[widget->state].y+offset.y);
      }
    }

    /** Else button just has text, draw it **/

    else
    {
      if (widget->state != 0 && widget->gcs[GC_FONT_SECOND] != -1)
        font_gc = color_font_1;
      else
        font_gc = color_font_0;

      if (widget->mode == MODE_NORMAL ||
			widget->mode == MODE_ACTIVE || !hold_shadow)
        draw_string (widget->drawable,
			widget->position.x, widget->position.y,
			widget->text_offset[widget->state].x,
			widget->text_offset[widget->state].y,
			widget->labels[label_state], font_0,
			Win->hue_widget[font_gc], FALSE);
      else if (widget->mode == MODE_SELECT)
        draw_string (widget->drawable,
			widget->position.x, widget->position.y,
			widget->text_offset[widget->state].x+offset.x,
			widget->text_offset[widget->state].y+offset.y,
			widget->labels[label_state], font_0,
			Win->hue_widget[font_gc], FALSE);
    }
  }
}



/******************************************************************************
  button_draw_subbutton (widget, state_on, x_position, y_position,
				x_size, y_size, invert, drawable)

  Draw a small button-like icon which serves the role of a checkbox in verbose
  and explicit buttons.
******************************************************************************/

button_draw_subbutton (widget, state_on, x_position, y_position,
				x_size, y_size, invert, drawable)
  Widget *widget;
  int state_on,
      x_position, y_position,
      x_size, y_size,
      invert;
  Window drawable;
{
  int i;

  GC use_gc;

  int src_x_position, src_y_position,
      dest_x_position, dest_y_position,
      shadow_thickness;

  float vert[3][2];

  short color_surface,
	color_surface_inv,
	color_check;

  if (widget->mode == MODE_SELECT)
    color_surface = widget->gcs[GC_BODY_INV];
  else
    color_surface = widget->gcs[GC_BODY];

  color_surface_inv = widget->gcs[GC_BODY_INV];
  color_check = widget->gcs[BUTTON_G_CHECK];

  /** Draw innards **/

  XFillRectangle (Win->display, drawable,
			Win->hue_widget[color_surface],
			x_position, y_position, x_size, y_size);

  shadow_thickness = 1;

  /** Draw shadowed border **/

  if (invert)
    draw_border (drawable, x_position, y_position, x_size, y_size,
		shadow_thickness,
		Win->hue_widget[COLOR_CTRL_HIGH],
		Win->hue_widget[COLOR_CTRL_SHADOW]);
  else
    draw_border (drawable, x_position, y_position, x_size, y_size,
		shadow_thickness,
		Win->hue_widget[COLOR_CTRL_SHADOW],
		Win->hue_widget[COLOR_CTRL_HIGH]);

  /** Draw check marker **/

  use_gc = Win->hue_widget[color_check];
  XSetLineAttributes (Win->display, use_gc, 2, LineSolid, CapButt, JoinMiter);

  if (!state_on)
    return;

  switch (widget->data[BUTTON_D_CHECK])
  {
    case CHECK_CROSS:
      XDrawLine (Win->display, drawable, use_gc, x_position, y_position,
				 x_position+x_size-1, y_position+y_size-1);
      XDrawLine (Win->display, drawable, use_gc, x_position+x_size-1,
				 y_position, x_position, y_position+y_size-1);
      break;

    case CHECK_SQUARE:
      XFillRectangle (Win->display, drawable, use_gc, x_position, y_position,
					 x_size, y_size);
      break;

    case CHECK_CHECK:
      XDrawLine (Win->display, drawable, use_gc,
			x_position+x_size/2, y_position+3*y_size/4,
			x_position+x_size/4, y_position+y_size/2);
      XDrawLine (Win->display, drawable, use_gc,
			x_position+x_size/2, y_position+3*y_size/4,
			x_position+x_size, y_position+y_size/4);
      break;
  }
  XSetLineAttributes (Win->display, use_gc, 1, LineSolid, CapButt, JoinMiter);
}
