#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_dial.h"


/******************************************************************************
  dial_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

dial_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      dial_help (Console->formats[CONTROL_DIAL]);
      break;

    case FUNCTION_FORMAT:
      dial_format (Console->formats[CONTROL_DIAL]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_DIAL], line);
      break;

    case FUNCTION_STORE:
      dial_store (widget, line);
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      dial_initialize (widget);
      break;

    case FUNCTION_PARSE:
      dial_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      dial_clean (widget);
      break;

    case FUNCTION_SETUP:
      dial_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      dial_execute (widget);
      break;

    case FUNCTION_DRAW:
      dial_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to dial", NULL);
      break;
  }
}



/******************************************************************************
  dial_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

dial_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("DIAL\n");
  printf ("    TEXT   <string>                dial label\n");
  printf ("    MIN    <float>                 minimum value\n");
  printf ("    MAX    <float>                 maximum value\n");
  printf ("    INIT   <float>                 initial value\n");
  printf ("    LOCATE <locate>                relative position of label\n");
  printf ("    ALTER   WIDTH   <int>          width of dial bar (%d)\n",
                        format->data[DIAL_D_BAR_WIDTH]);
  printf ("    ALTER   TICKS   <bool>         (don't) use tickmarks (%d)\n",
                        format->data[DIAL_D_TICKS]);
  printf ("    ALTER   DIRECT  <bool>         (counter)clockwise turn (%d)\n",
                        format->data[DIAL_D_DIRECTION]);
  printf ("    ALTER   FRINGE  <int>          width of fringe (%d)\n",
                        format->data[DIAL_D_FRINGE]);
  printf ("    ALTER   ANGLE   <float>        radian extent of range (%4.2f)\n",
                        format->data_float[DIAL_DF_ANGLE]);
  printf ("    ALTER   ZERO    <float>        radian start of range (%4.2f)\n",
                        format->data_float[DIAL_DF_ZERO_ANGLE]);
  printf ("\n");
}



/******************************************************************************
  dial_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

dial_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	TRUE;
  format->status[IS_ACTIVATABLE] =	TRUE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	TRUE;
  format->status[IS_BUFFERABLE] =	BUFFER_CLEAR;

  format->size.x =			DEFAULT_DIAL_DIMENSION_X;
  format->size.y =			DEFAULT_DIAL_DIMENSION_Y;
  format->spacing.x =			DEFAULT_DIAL_SPACING_X;
  format->spacing.y =			DEFAULT_DIAL_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	2;

  format->gcs[GC_BODY] =		COLOR_CTRL_SURFACE;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_SURFACE_INV;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_DARK;
  format->gcs[DIAL_G_BAR] =		COLOR_CTRL_TEXT_3;
  format->gcs[DIAL_G_FRINGE] =		COLOR_CTRL_BACK;

  format->data[DIAL_D_BAR_WIDTH] =	3;
  format->data[DIAL_D_DIRECTION] =	DIAL_CLOCKWISE;
  format->data[DIAL_D_FRINGE] =		12;
  format->data[DIAL_D_TICKS] =		TRUE;

  format->data_float[DIAL_DF_ZERO_ANGLE] = DIAL_PI;
  format->data_float[DIAL_DF_ANGLE] =	DIAL_THREE_HALF_PI;

  format->fonts[FONT_FIRST] =		FONT_LARGE;
  format->fonts[FONT_SECOND] =		FONT_MEDIUM;

  format->locate =			LABEL_NORTH;
  format->locate_offset.y =		format->data[DIAL_D_FRINGE] + 5;
  format->locate_offset.x =		format->data[DIAL_D_FRINGE] + 5;

  format->format =			CONTROL_DIAL;
}



/******************************************************************************
  dial_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

dial_store (widget, line)
  Widget *widget;
  char *line;
{
  char subline[100];

  line[0] = '\0';
  sprintf (subline, "INIT\t%f\n", widget->value);
  strcat (line, subline);
}



/******************************************************************************
  dial_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

dial_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_DIAL;
}



/******************************************************************************
  dial_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

dial_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100],
       *ptr,
       ch;  

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (tres_line);
  int_factor = atoi (tres_line);
 
  if (line[0] == 'A')
  {
    switch (dos_line[0])
    {
      /** WIDTH **/

      case 'W':

        widget->data[DIAL_D_BAR_WIDTH] = int_factor;
        break;

      /** DIRECTION **/

      case 'D':

        widget->data[DIAL_D_DIRECTION] = int_factor;
        break;

      /** FRINGE **/

      case 'F':

        widget->data[DIAL_D_FRINGE] = int_factor;
        break;

      /** TICKS **/

      case 'T':

        widget->data[DIAL_D_TICKS] = int_factor;
        break;

      /** ANGLE **/

      case 'A':

        widget->data_float[DIAL_DF_ANGLE] = factor;
        break;

      /** ZERO **/

      case 'Z':

        widget->data_float[DIAL_DF_ZERO_ANGLE] = factor;
        break;
    }

    line[0] = '#';
  }
}



/******************************************************************************
  dial_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

dial_clean (widget)
  Widget *widget;
{
  if (widget->label == NULL)
    widget->label = strdup (widget->name);
}



/******************************************************************************
  dial_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

dial_setup (widget)
  Widget *widget;
{
  char line[20];

  if (widget->label == NULL)
    widget->locate = LABEL_NONE;
  else
    locate_text (widget, widget->label, 0, FONT_FIRST, FALSE);

  widget->bound_offset.x = widget->data[DIAL_D_FRINGE];
  widget->bound_offset.y = widget->data[DIAL_D_FRINGE];
  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;    
  widget->bound_size.y = 2*widget->bound_offset.y + widget->size.y;       
}



/******************************************************************************
  dial_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

dial_execute (widget)
  Widget *widget;
{
  double value,
         angle, angle_diff, angle_fraction,
         radius,
         pos_x, pos_y;

  widget->flag = TRUE;

  /** If right button, just set value to last value **/

  if (Win->event.xbutton.button == Button3)
  {
    value = Master->last_value;
 
    if (value < widget->minimum)
      value = widget->minimum;
    else if (value > widget->maximum)
      value = widget->maximum;
  }

  /** Else normal button, figure out value **/

  else
  {
    radius = widget->size.x/2;
 
    pos_x = (double)(Win->mouse.x - (widget->position.x + widget->size.x/2));
    pos_y = -(double)(Win->mouse.y - (widget->position.y + widget->size.y/2));
 
    if (pos_x == 0 && pos_y == 0) 
      value = widget->value;   
    else 
    {  
      angle = atan2 (pos_y, pos_x);

      if (widget->data[DIAL_D_DIRECTION] == DIAL_CLOCKWISE)
      {
        if (angle > widget->data_float[DIAL_DF_ZERO_ANGLE])
          angle_diff = DIAL_TWO_PI -
			(angle - widget->data_float[DIAL_DF_ZERO_ANGLE]);
        else
          angle_diff = widget->data_float[DIAL_DF_ZERO_ANGLE] - angle;
      }
      else
      {
        if (angle < widget->data_float[DIAL_DF_ZERO_ANGLE])
          angle_diff = DIAL_TWO_PI -
			(widget->data_float[DIAL_DF_ZERO_ANGLE] - angle);
        else
          angle_diff = angle - widget->data_float[DIAL_DF_ZERO_ANGLE];
      }

      if (widget->data_float[DIAL_DF_ANGLE] < 0.0001)
        angle_fraction = angle_diff/0.0000001;
      else
        angle_fraction = angle_diff/widget->data_float[DIAL_DF_ANGLE];

      if (angle_fraction > 1.0)
      {
        if (widget->data_float[DIAL_DF_ANGLE] == DIAL_TWO_PI)
          angle_fraction = (angle_diff - widget->data_float[DIAL_DF_ANGLE])/
		0.0000001;
        else
          angle_fraction = (angle_diff - widget->data_float[DIAL_DF_ANGLE])/
		(DIAL_TWO_PI - widget->data_float[DIAL_DF_ANGLE]);

        if (angle_fraction < 0.5)
          angle_fraction = 1.0;
        else
          angle_fraction = 0.0;
      }

      value = widget->minimum +
			angle_fraction * (widget->maximum - widget->minimum);
    }
  }
 
  widget->value = value;
  Master->last_value = value;
}



/******************************************************************************
  dial_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

dial_draw (widget)
  Widget *widget;
{
  int i;

  int dial_bar_width,
      shadow_thickness,
      fringe,
      arc_start, arc_end,
      offset;

  XPoint center,
         arrow,
         tick_position;

  double radius, angle,
         fraction,
         tick_radius;

  short color_surface,
        color_surface_inv,
        color_background,
        color_bar,
        color_fringe,
	color_font_0,
        font_0;

  if (widget->enable == FALSE)
    return;

  /** Set convenience variables **/

  color_surface = widget->gcs[GC_BODY];
  color_surface_inv = widget->gcs[GC_BODY_INV];
  color_font_0 = widget->gcs[GC_FONT_FIRST];
  color_background = widget->gcs[GC_BACKGROUND];
  color_bar = widget->gcs[DIAL_G_BAR];
  color_fringe = widget->gcs[DIAL_G_FRINGE];
  font_0 = widget->fonts[FONT_FIRST];

  dial_bar_width = widget->data[DIAL_D_BAR_WIDTH];
  shadow_thickness = widget->thickness[THICKNESS_SHADOW];
  fringe = widget->data[DIAL_D_FRINGE];

  center.x = widget->position.x + widget->size.x/2;
  center.y = widget->position.y + widget->size.y/2;
  radius = widget->size.x/2;

/**
  if (widget->mode == MODE_SELECT || widget->mode == MODE_ACTIVE)
    XFillArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_ACTIVE],
		widget->position.x-1, widget->position.y-1,
		widget->size.x+2, widget->size.y+2, 0, 23040);
  else
    XFillArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_SURFACE_1],
		widget->position.x-1, widget->position.y-1,
		widget->size.x+2, widget->size.y+2, 0, 23040);
**/

  /** Draw fringe and markers **/

  angle = widget->data_float[DIAL_DF_ZERO_ANGLE];

  if (widget->data[DIAL_D_DIRECTION] == DIAL_CLOCKWISE)
    angle -= widget->data_float[DIAL_DF_ANGLE];

  if (angle < 0)
    angle = DIAL_TWO_PI + angle;
  else if (angle > DIAL_TWO_PI)
    angle = angle - DIAL_TWO_PI;

  if (widget->data[DIAL_D_DIRECTION] == DIAL_CLOCKWISE)
  {
    arc_start = (int)(DIAL_ARC_FACTOR * angle + 0.5);
    arc_end = (int)(DIAL_ARC_FACTOR * widget->data_float[DIAL_DF_ANGLE] + 0.5);
  }

  XFillArc (Win->display, widget->drawable,
		Win->hue_widget[color_fringe],
		widget->position.x-fringe, widget->position.y-fringe,
		widget->size.x+2*fringe, widget->size.y+2*fringe,
		arc_start, arc_end);

  /** Draw tick marks **/

  if (widget->data[DIAL_D_TICKS])
  {
    fraction = widget->data_float[DIAL_DF_ANGLE]/8.0;
    angle = widget->data_float[DIAL_DF_ZERO_ANGLE];

    for (i=0; i<=8; i++)
    {
      if (i%4 == 0)
        tick_radius = (double)(radius + fringe);
      else if (i%2 == 0)
        tick_radius = (double)(radius + fringe/2);
      else
        tick_radius = (double)(radius + fringe/2);

      tick_position.x = center.x + (int)(tick_radius * cos (angle) + 0.5); 
      tick_position.y = center.y - (int)(tick_radius * sin (angle) + 0.5); 

      XDrawLine (Win->display, widget->drawable,
                Win->hue_widget[COLOR_CTRL_DARK],
                center.x, center.y,
                tick_position.x, tick_position.y);

      if (widget->data[DIAL_D_DIRECTION] == DIAL_CLOCKWISE)
        angle -= fraction;
      else
        angle += fraction;
    }
  }

  /** Draw shadows **/

  XFillArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_HIGH],
		widget->position.x, widget->position.y-shadow_thickness,
		widget->size.x, widget->size.y, 0, 23040);
  XFillArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_SHADOW],
		widget->position.x, widget->position.y + shadow_thickness,
		widget->size.x, widget->size.y, 0, 23040);

  /** Draw dial itself **/

  if (widget->mode == MODE_SELECT)
    XFillArc (Win->display, widget->drawable,
		Win->hue_widget[color_surface_inv],
		widget->position.x, widget->position.y,
		widget->size.x, widget->size.y, 0, 23040);
  else
    XFillArc (Win->display, widget->drawable,
		Win->hue_widget[color_surface],
		widget->position.x, widget->position.y,
		widget->size.x, widget->size.y, 0, 23040);

  /** Draw dial line **/

  if (widget->maximum < 0.0000001)
    angle = widget->data_float[DIAL_DF_ANGLE]*
			(widget->value - widget->minimum)/0.0000001;
  else
    angle = widget->data_float[DIAL_DF_ANGLE]*
			(widget->value - widget->minimum)/widget->maximum;

  if (widget->data[DIAL_D_DIRECTION] == DIAL_CLOCKWISE)
    angle = widget->data_float[DIAL_DF_ZERO_ANGLE] - angle;
  else
    angle = widget->data_float[DIAL_DF_ZERO_ANGLE] + angle;

  arrow.x = (long)(center.x + radius * cos(angle) + 0.5);
  arrow.y = (long)(center.y - radius * sin(angle) + 0.5);

  XSetLineAttributes (Win->display, Win->hue_widget[color_bar],
			dial_bar_width, LineSolid, CapButt, JoinMiter);

  XDrawLine (Win->display, widget->drawable,
		Win->hue_widget[color_bar],
		center.x, center.y, arrow.x, arrow.y);

  XSetLineAttributes (Win->display, Win->hue_widget[color_bar],
			1, LineSolid, CapButt, JoinMiter);

  arrow.x = center.x - widget->size.x/4;
  arrow.y = center.y - widget->size.y/4;

  XFillArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_HIGH],
		arrow.x, arrow.y-shadow_thickness*2,
		widget->size.x/2, widget->size.y/2, 0, 23040);
  XDrawArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_DARK],
		arrow.x, arrow.y-shadow_thickness*2,
		widget->size.x/2, widget->size.y/2, 0, 23040);

  XFillArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_SHADOW],
		arrow.x, arrow.y+shadow_thickness*2,
		widget->size.x/2, widget->size.y/2, 0, 23040);

  /** Draw dial itself **/

  if (widget->mode == MODE_SELECT)
    XFillArc (Win->display, widget->drawable,
		Win->hue_widget[color_surface_inv],
		arrow.x, arrow.y,
		widget->size.x/2, widget->size.y/2, 0, 23040);
  else
    XFillArc (Win->display, widget->drawable,
		Win->hue_widget[color_surface],
		arrow.x, arrow.y,
		widget->size.x/2, widget->size.y/2, 0, 23040);

  XDrawArc (Win->display, widget->drawable,
		Win->hue_widget[COLOR_CTRL_DARK],
		arrow.x, arrow.y,
		widget->size.x/2, widget->size.y/2, 0, 23040);

  /** Draw label **/

  if (widget->label != NULL && widget->locate != LABEL_NONE)
    draw_string (Win->window,
                widget->real_position.x, widget->real_position.y,
                widget->text_offset[0].x, widget->text_offset[0].y,
                widget->label,
                font_0, Win->hue_widget[color_font_0], FALSE);
}
