#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_label.h"


/******************************************************************************
  label_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

label_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      label_help (Console->formats[CONTROL_LABEL]);
      break;

    case FUNCTION_FORMAT:
      label_format (Console->formats[CONTROL_LABEL]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_LABEL], line);
      break;

    case FUNCTION_STORE:
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      label_initialize (widget);
      break;

    case FUNCTION_PARSE:
      break;

    case FUNCTION_CLEAN:
      break;

    case FUNCTION_SETUP:
      label_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      label_execute (widget);
      break;

    case FUNCTION_DRAW:
      label_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to text", NULL);
      break;
  }
}



/******************************************************************************
  label_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

label_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("LABEL\n");
  printf ("    STRING <string>                label string\n");
  printf ("\n");
}



/******************************************************************************
  label_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

label_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	FALSE;
  format->status[IS_ACTIVATABLE] =	FALSE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	FALSE;
  format->status[IS_BUFFERABLE] =	BUFFER_CLEAR;

  format->size.x =			DEFAULT_LABEL_DIMENSION_X;
  format->size.y =			DEFAULT_LABEL_DIMENSION_Y;
  format->spacing.x =			DEFAULT_LABEL_SPACING_X;
  format->spacing.y =			DEFAULT_LABEL_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	2;

  format->gcs[GC_BODY] =		COLOR_CTRL_SURFACE_2;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_SURFACE_INV_2;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_TEXT_1;

  format->fonts[FONT_FIRST] =		FONT_MEDIUM;
  format->fonts[FONT_SECOND] =		FONT_MEDIUM;

  format->locate_offset.x =		0;
  format->locate_offset.y =		0;

  format->locate =			LABEL_WEST;

  format->format =			CONTROL_LABEL;
}



/******************************************************************************
  label_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

label_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_LABEL;
}



/******************************************************************************
  label_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

label_setup (widget)
  Widget *widget;
{
  int junk;

  XCharStruct text_information;

  XPoint text_start;

  if (widget->graphic_count)
  {
    load_graphics (widget);
    widget->size.x = widget->graphics[0]->width;
    widget->size.y = widget->graphics[0]->height;

    widget->bound_offset.x = -widget->text_offset[0].x;
    widget->bound_offset.y = -widget->text_offset[0].y;
    widget->bound_size.x = widget->size.x;    
    widget->bound_size.y = widget->size.y;       
  }
  else
  {
    locate_text (widget, widget->string, 0, FONT_FIRST, FALSE);

    XTextExtents (Win->fonts[widget->fonts[FONT_FIRST]]->font,
		widget->string, strlen (widget->string),
		&junk, &junk, &junk, &text_information);

    text_start.x = widget->position.x + widget->text_offset[0].x -
			text_information.lbearing;
    text_start.y = widget->position.y + widget->text_offset[0].y -
			text_information.ascent;

    widget->bound_offset.x = widget->position.x - text_start.x;
    widget->bound_offset.y = widget->position.y - text_start.y;

    widget->bound_size.x = text_information.lbearing +
				text_information.rbearing;
    widget->bound_size.y = text_information.ascent +
				text_information.descent;
  }
}



/******************************************************************************
  label_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

label_execute (widget)
  Widget *widget;
{ 
  widget->flag = TRUE; 
}



/******************************************************************************
  label_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

label_draw (widget)
  Widget *widget;
{
  int i;

  GC use_gc;

  float vert[3][2];

  short color_font_0,
        font_0;

  Graphic *graphic;

  color_font_0 = widget->gcs[GC_FONT_FIRST];
  font_0 = widget->fonts[FONT_FIRST];

  if (widget->graphic_count)
  {
    graphic = widget->graphics[0];

    if (graphic->is_bitmap)
    {  
      XCopyPlane (Win->display, graphic->pixmap, widget->drawable,
			Win->hue_widget[graphic->gc],
			0, 0, graphic->width, graphic->height,
			widget->position.x +
				widget->text_offset[widget->state].x,
			widget->position.y +
				widget->text_offset[widget->state].y, 1L);
    }  
    else
    {   
      XCopyArea (Win->display, graphic->pixmap, widget->drawable,
			Win->hue_widget[0],
			0, 0, graphic->width, graphic->height,
			widget->position.x +
				widget->text_offset[widget->state].x,
			widget->position.y +
				widget->text_offset[widget->state].y);
    }
  }
  else
  {
    draw_string (widget->drawable,
		widget->position.x, widget->position.y,
		widget->text_offset[0].x, widget->text_offset[0].y,
		widget->string, font_0,
		Win->hue_widget[color_font_0], FALSE);
  }
}
