#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_data.h"
#include "pip_c_scroll.h"
#include "pip_c_list.h"


/******************************************************************************
  list_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

list_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      list_help (Console->formats[CONTROL_LIST]);
      break;

    case FUNCTION_FORMAT:
      list_format (Console->formats[CONTROL_LIST]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_LIST], line);
      break;

    case FUNCTION_STORE:
      list_store (widget, line);
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      list_initialize (widget);
      break;

    case FUNCTION_PARSE:
      list_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      list_clean (widget);
      break;

    case FUNCTION_SETUP:
      list_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      list_execute (widget, line);
      break;

    case FUNCTION_DRAW:
      list_draw (widget, line);
      break;

    default:
      throw_warning ("Function %d unknown to list", NULL);
      break;
  }
}



/******************************************************************************
  list_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

list_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("LIST\n");
  printf ("    HORIZ                          show horizontal scroll\n");
  printf ("    TEXT   <string>                slider label\n");
  printf ("    MIN    <float>                 minimum value\n");
  printf ("    MAX    <float>                 maximum value\n");
  printf ("    INIT   <float>                 initial value\n");
  printf ("    LOCATE <locate>                relative position of label\n");
  printf ("    ALTER   SPACE   <int>          space before scrollbar (%d)\n",
                        format->data[LIST_D_SPACE]);
  printf ("\n");
}



/******************************************************************************
  list_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

list_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	FALSE;
  format->status[IS_ACTIVATABLE] =	FALSE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	FALSE;
  format->status[IS_BUFFERABLE] =	FALSE;
  format->status[IS_COMPOSITE] =	TRUE;

  format->size.x =			DEFAULT_LIST_DIMENSION_X;
  format->size.y =			DEFAULT_LIST_DIMENSION_Y;
  format->spacing.x =			DEFAULT_LIST_SPACING_X;
  format->spacing.y =			DEFAULT_LIST_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] = 2;
  format->thickness[THICKNESS_ACTIVE] = 1;

  format->data[LIST_D_SPACE] =		DEFAULT_LIST_SPACE;
  format->data[LIST_D_LINE_COUNT] =	DEFAULT_LIST_LINE_COUNT;
  format->data[LIST_D_HORIZONTAL] =	DEFAULT_LIST_HORIZONTAL;

  format->data_float[LIST_DF_LINE_SPACING] =	DEFAULT_LIST_LINE_SPACING;

  format->gcs[GC_BODY] =		COLOR_CTRL_AUX_2;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_DARK_INV;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_TEXT_2;

  format->fonts[FONT_FIRST] =		FONT_MEDIUM;
  format->fonts[FONT_SECOND] =		FONT_MEDIUM;

  format->locate =			LABEL_WEST;

  format->format = CONTROL_LIST;
}



/******************************************************************************
  list_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

list_store (widget, line)
  Widget *widget;
  char *line;
{
  char subline[100];

  line[0] = '\0';
  sprintf (subline, "INIT\t%f\n", widget->value);
  strcat (line, subline);
}



/******************************************************************************
  list_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

list_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_LIST;
}



/******************************************************************************
  list_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

list_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100],
       *ptr,
       ch;  

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);
 
  factor = atof (tres_line);
  int_factor = atoi (tres_line);
 
  if (line[0] == 'A')
  {
    switch (dos_line[0])
    {
      /** HORIZ **/

      case 'H':

        widget->data[LIST_D_HORIZONTAL] = TRUE;
        break;

      /** SPACE **/

      case 'S':

        widget->data[LIST_D_SPACE] = int_factor;
        break;

      /** LINES **/

      case 'L':

        widget->data[LIST_D_LINE_COUNT] = int_factor;
        break;

      /** INTER **/

      case 'I':

        widget->data_float[LIST_DF_LINE_SPACING] = factor;
        break;
    }

    line[0] = '#';
  }
}



/******************************************************************************
  list_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

list_clean (widget)
  Widget *widget;
{
  int i;
  char line[100];

  Widget *widget1,
         *widget2,
         *widget3;

  Panel *panel;

  panel = Console->panels[Console->panel_count];

  /** Initialize data widget **/

  widget1 = create_widget();
  widget1->panel = panel;
  panel->widgets[panel->widget_count] = widget1;
  panel->widget_count++;

  widget1->format = CONTROL_DATA;
  sprintf (line, "%s_text", widget->name);
  widget1->name = strdup (line);
  copy_defaults (widget1, NULL);
  widget1->superformat = CONTROL_LIST;
  widget1->relatives[0] = (void *)(widget);

  /** Initialize vertical scroll widget **/

  widget2 = create_widget();
  widget2->panel = panel;
  panel->widgets[panel->widget_count] = widget2;
  panel->widget_count++;

  widget2->format = CONTROL_SCROLL;
  sprintf (line, "%s_scroll", widget->name);
  widget2->name = strdup (line);
  copy_defaults (widget2, NULL);
  widget2->superformat = CONTROL_LIST;
  widget2->relatives[0] = (void *)(widget);

  /** Initialize horizontal scroll widget **/

  if (widget->data[LIST_D_HORIZONTAL])
  {
    widget3 = create_widget();
    widget3->panel = panel;
    panel->widgets[panel->widget_count] = widget3;
    panel->widget_count++;

    widget3->format = CONTROL_SCROLL;
    sprintf (line, "%s_h_scroll", widget->name);
    widget3->name = strdup (line);
    copy_defaults (widget3, NULL);
    widget3->superformat = CONTROL_LIST;
    widget3->relatives[0] = (void *)(widget);
  }
  else
    widget3 = NULL;

  /** Initialize relatives (for superformat) **/

  widget->relatives[0] = (void *)(widget1);
  widget->relatives[1] = (void *)(widget2);
  widget->relatives[2] = (void *)(widget3);

  /** Set up data widget **/

  load_partial_widget (widget1, widget);
  for (i=0; i<widget->state_count; i++)
    widget1->labels[i] = widget->labels[i];
  set_string (widget1, widget->string);
  widget1->size.x = widget->size.x - widget2->data[SCROLL_D_BREADTH] -
			widget->data[LIST_D_SPACE];
  widget1->data[DATA_D_DIGIT] =	'N';
  widget1->data[DATA_D_TEXT] =		'L';
  widget1->fonts[FONT_FIRST] =		widget->fonts[FONT_FIRST];
  widget1->fonts[FONT_SECOND] =		widget->fonts[FONT_SECOND];
  widget1->locate =			widget->locate;
  widget1->data[DATA_D_LINE_COUNT] =	widget->data[LIST_D_LINE_COUNT];
  widget1->status[IS_EDITABLE] =	EDITABLE_NEVER;
  widget1->gcs[GC_BODY_INV] =		widget1->gcs[GC_BODY];
  widget1->state = 			0;
  widget1->next_state = 		-1;
  widget1->data_float[DATA_DF_LINE_SPACING] =
			widget->data_float[LIST_DF_LINE_SPACING];

  /** CHECK: should GCs be set too? **/

  /** Set up vertical scroll widget **/

  free (widget2->link_names[LINK_BASIC]);
  widget2->link_names[LINK_BASIC] =		strdup (widget1->name);
  widget2->data[SCROLL_D_ORIENTATION] =		SCROLL_VERTICAL;
  widget2->spacing.x =				widget->data[LIST_D_SPACE];
  widget2->position.x =				LOCATE_RIGHT;
  widget2->position.y =				LOCATE_SAME;

  /** Set up horizontal scroll widget **/

  if (widget3)
  {
    free (widget3->link_names[LINK_BASIC]);
    widget3->link_names[LINK_BASIC] =		strdup (widget1->name);
    widget3->data[SCROLL_D_ORIENTATION] =	SCROLL_HORIZONTAL;
    widget3->spacing.y =			widget->data[LIST_D_SPACE];
    widget3->position.x =			LOCATE_SAME;
    widget3->position.y =			LOCATE_BELOW;
  }

  /** Run type-specific cleaning operation **/

  widget_function (widget1, NULL, FUNCTION_CLEAN);
  widget_function (widget2, NULL, FUNCTION_CLEAN);
  if (widget3)
    widget_function (widget3, NULL, FUNCTION_CLEAN);
}



/******************************************************************************
  list_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

list_setup (widget)
  Widget *widget;
{
  Widget *widget1, *widget2, *widget3;

  widget1 = widget->relatives[0];
  widget2 = widget->relatives[1];
  widget3 = widget->relatives[2];

  widget->size.y = widget->data[LIST_D_LINE_COUNT] *
	((int)(widget->data_float[LIST_DF_LINE_SPACING] *
		Win->fonts[widget->fonts[FONT_FIRST]]->dimension.y)) +
		DATA_TOP_OFFSET + DATA_BOTTOM_OFFSET;

  /** Set up data widget **/

  widget1->size.y = widget->size.y;

  /** Set up vertical scroll widget **/

  widget2->size.y = widget->size.y;

  /** Set up horizontal scroll widget **/

  if (widget3)
    widget3->size.x = widget1->size.x;

  /** Load text list **/

  list_load (widget, widget1->labels, widget1->state_count, FALSE, FALSE);
}



/******************************************************************************
  list_load (widget, strings, count, do_draw, to_end)

  Load a list of <count> text strings specified by <strings> into list widget
  <widget>.  If <do_draw>, draw the widget after loading.  If <to_end>,
  load only as many of the final strings as will fit in the list widget.
******************************************************************************/

list_load (widget, strings, count, do_draw, to_end)
  Widget *widget;
  char *strings[];
  int count,
      do_draw,
      to_end;
{
  int i,
      length,
      max_length,
      max_width,
      junk;

  XCharStruct text;
  Widget *widget1, *widget2, *widget3;

  /** Set data and scroll widget pointers **/

  widget1 = (Widget *)(widget->relatives[0]);
  widget2 = (Widget *)(widget->relatives[1]);
  widget3 = (Widget *)(widget->relatives[2]);

  /** Free old superlabels **/

  if (widget1->superlabels != NULL)
  {
    for (i=0; i<widget1->state_count; i++)
      free (widget1->superlabels[i]);
    free (widget1->superlabels);
    widget1->superlabels = NULL;
  }

  /** Determine start state **/

  if (to_end)
  {
    widget1->state = count - widget1->data[DATA_D_LINE_COUNT] - (to_end-1);
    if (widget1->state < 0)
      widget1->state = 0;
  }
  else
    widget1->state = 0;

  /** Set up strings **/

  widget1->next_state =	-1;
  widget1->state_count =	count;
  widget1->superlabels =	(char **)(malloc(sizeof(char *)*count));
  for (i=0; i<widget1->state_count; i++)
    widget1->superlabels[i] =	strdup (strings[i]);

  /** Set up scroll bar bounds **/

  widget2->minimum =		0.0;
  widget2->maximum =		(double)
		(widget1->state_count - widget1->data[DATA_D_LINE_COUNT]);
  if (widget2->maximum != 0.0)
    widget2->data_float[SCROLL_DF_ARROW_MOVE] = 1.0/widget2->maximum;
  else
    widget2->data_float[SCROLL_DF_ARROW_MOVE] = 0.1;

  widget2->value = widget1->state;

  widget->state_count = widget1->state_count;
  widget->superlabels = widget1->superlabels;

  /** Set up horizontal scroll bar bounds **/

  if (widget3)
  {
    /** Find longest line of text, in terms of characters and pixels **/

    max_width = 0;
    max_length = 0;
    for (i=0; i<widget1->state_count; i++)
    {
      length = strlen (widget1->superlabels[i]);

      if (length > max_length)
        max_length = length;

      XTextExtents (Win->fonts[widget1->fonts[FONT_FIRST]]->font,
		widget1->superlabels[i],
		length, &junk, &junk, &junk, &text);

      if (text.width > max_width)
        max_width = text.width;
    }

    /** If longest line too wide for DATA widget, set offset **/

    if (max_width > (widget1->size.x - 2*widget1->text_offset[0].x))
    {
      widget3->maximum = max_width -
			(widget1->size.x - 2*widget1->text_offset[0].x);

      widget3->maximum = widget3->maximum /
		Win->fonts[widget1->fonts[FONT_FIRST]]->dimension.x + 1;
    }
    else
      widget3->maximum = 0.0;

    widget3->minimum = 0.0;
    widget3->value = 0.0;
    widget1->data[DATA_D_X_OFFSET] = 0;
    if (widget3->maximum != 0.0)
      widget3->data_float[SCROLL_DF_ARROW_MOVE] = 1.0/widget3->maximum + 0.001;
    else
      widget3->data_float[SCROLL_DF_ARROW_MOVE] = 0.01;
  }

  /** Set up scroll bar **/

  list_set_scrollbar (widget1, widget2, widget3);

  /** Draw each component of list widget **/

  if (do_draw)
  {
    widget1->superformat = CONTROL_NONE;
    draw_widget (widget1);
    widget1->superformat = CONTROL_LIST;

    widget2->superformat = CONTROL_NONE;
    draw_widget (widget2);
    widget2->superformat = CONTROL_LIST;

    if (widget3)
    {
      widget3->superformat = CONTROL_NONE;
      draw_widget (widget3);
      widget3->superformat = CONTROL_LIST;
    }

    XSync (Win->display, 0);
  }
}



/******************************************************************************
  list_get_entry (widget, string, index)

  Get the <index>th text entry from list widget <widget> and return in
  character string <string>.  User callable.
******************************************************************************/

list_get_entry (widget, string, index)
  Widget *widget;
  char *string;
  int index;
{
  if (widget->format == CONTROL_LIST)
    widget = (Widget *)(widget->relatives[0]);

  if (index < 0)
  {
    if (widget->next_state >= 0)
      index = widget->next_state;
    else
      index = 0;
  }
  strcpy (string, widget->superlabels[index]);
}



/******************************************************************************
  list_set_entry (widget, string, index)

  Set the <index>th text entry from list widget <widget> using character
  string <string>.  User callable.
******************************************************************************/

list_set_entry (widget, string, index)
  Widget *widget;
  char *string;
  int index;
{
  if (widget->format == CONTROL_LIST)
    widget = (Widget *)(widget->relatives[1]);

  if (index < 0)
  {
    if (widget->next_state >= 0)
      index = widget->next_state;
    else
      index = 0;
  }

  free (widget->superlabels[index]);
  widget->superlabels[index] = strdup (string);
  draw_widget (widget);
  XSync (Win->display, 0);
}




/******************************************************************************
  list_set_scrollbar (widget1, widget2, widget3)

  Set the size and position of the scrollbar bar, given scroll widget
  <widget2> and associated data widget <widget1>.
******************************************************************************/

list_set_scrollbar (widget1, widget2, widget3)
  Widget *widget1, *widget2, *widget3;
{
  double factor;

  if (widget1->state_count)
    factor = ((double)(widget1->data[DATA_D_LINE_COUNT]))/widget1->state_count;
  else
    factor = 1.0;

  if (factor < 1.0)
    widget2->data[SCROLL_D_MARKER_SIZE] =
	(int)((widget2->size.y - 2*DEFAULT_SCROLL_BREADTH - 2)*factor);
  else
    widget2->data[SCROLL_D_MARKER_SIZE] =
	widget2->size.y - 2*DEFAULT_SCROLL_BREADTH - 2;

  widget2->data[SCROLL_D_ACTIVE_LENGTH] = widget2->data[SCROLL_D_LENGTH] - 2 -
		widget2->data[SCROLL_D_MARKER_SIZE] - 2*DEFAULT_SCROLL_BREADTH;

  if (widget3)
  {
    widget3->data[SCROLL_D_MARKER_SIZE] = 
	widget3->size.x - 2*DEFAULT_SCROLL_BREADTH - 2;

    if (widget3->maximum == widget3->minimum)
      factor = 1.0;
    else
      factor = ((double)(widget3->size.x)) / 
		(widget3->size.x + widget3->maximum * 
		Win->fonts[widget1->fonts[FONT_FIRST]]->dimension.x);

    widget3->data[SCROLL_D_MARKER_SIZE] =
		(int)(factor * widget3->data[SCROLL_D_MARKER_SIZE]);

    widget3->data[SCROLL_D_ACTIVE_LENGTH] = widget3->size.x - 2 -
		widget3->data[SCROLL_D_MARKER_SIZE] - 2*DEFAULT_SCROLL_BREADTH;
  }
}



/******************************************************************************
  list_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

list_execute (widget, name)
  Widget *widget;
  char *name;
{
  int old_state;
  Widget *widget1,
         *widget2,
         *widget3;

  if (name == NULL)
    return;

  widget1 = (Widget *)(widget->relatives[0]);
  widget2 = (Widget *)(widget->relatives[1]);
  widget3 = (Widget *)(widget->relatives[2]);

  /** If it's not the data widget **/ 

  if (strcmp (widget2->name, name) == 0)
  {
    old_state = widget1->state;

    /** If it's an arrow press **/

    if (!widget2->status[IS_MOTION_ACTIVE])
    {
      if (widget2->data[SCROLL_D_ARROW_SELECTED] == 0)
        widget1->state--;
      else
        widget1->state++;
    }
    else
    {
      widget1->state = (int)(widget2->value);
    }

    /** Correct out-of-bounds errors **/

    if (widget1->state >
		(widget1->state_count - widget1->data[DATA_D_LINE_COUNT]))
      widget1->state =
		widget1->state_count - widget1->data[DATA_D_LINE_COUNT];
    if (widget1->state < 0)
      widget1->state = 0;

    if (widget1->state == old_state)
      return;

    widget1->superformat = CONTROL_NONE;
    draw_widget (widget1);
    widget1->superformat = CONTROL_LIST;
    XSync (Win->display, 0);
  }
  else if (widget3 && (strcmp (widget3->name, name) == 0))
  {
    widget1->data[DATA_D_X_OFFSET] = widget3->value;

    widget1->superformat = CONTROL_NONE;
    draw_widget (widget1);
    widget1->superformat = CONTROL_LIST;
    XSync (Win->display, 0);
  }
}



/******************************************************************************
  list_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

list_draw (widget, name)
  Widget *widget;
  char *name;
{
  Widget *widget1,
         *widget2,
         *widget3;

  if (name == NULL)
    return;

  widget1 = (Widget *)(widget->relatives[0]);
  widget2 = (Widget *)(widget->relatives[1]);
  widget3 = (Widget *)(widget->relatives[2]);

  if (!strcmp(widget1->name, name))
  {
    widget2->superformat = CONTROL_NONE;
    draw_widget (widget2);
    widget2->superformat = CONTROL_LIST;
  }
  else
  {
    widget1->superformat = CONTROL_NONE;
    draw_widget (widget1);
    widget1->superformat = CONTROL_LIST;
  }
  XSync (Win->display, 0);
}
