#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_plate.h"


/******************************************************************************
  plate_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

plate_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      plate_help (Console->formats[CONTROL_PLATE]);
      break;

    case FUNCTION_FORMAT:
      plate_format (Console->formats[CONTROL_PLATE]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_PLATE], line);
      break;

    case FUNCTION_STORE:
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      plate_initialize (widget);
      break;

    case FUNCTION_PARSE:
      plate_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      plate_clean (widget);
      break;

    case FUNCTION_SETUP:
      plate_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      plate_execute (widget);
      break;

    case FUNCTION_DRAW:
      plate_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to plate", NULL);
      break;
  }
}



/******************************************************************************
  plate_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

plate_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("PLATE\n");
  printf ("    TEXT   <string>                plate label\n");
  printf ("    LOCATE <locate>                relative position of label\n");
  printf ("    ALTER   OUTBULGE               raised appearance\n");
  printf ("    ALTER   INBULGE                sunken appearance\n");
  printf ("    ALTER   FLAT                   homogeneously sunken/raised\n");
  printf (
   "    ALTER   RIM     <int>          rimmed border, width <int> (%d)\n",
			format->data[PLATE_D_RIM_THICK]);
  printf ("    ALTER   TROUGH                 sunken rim\n");
  printf ("\n");
}



/******************************************************************************
  plate_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

plate_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	FALSE;
  format->status[IS_ACTIVATABLE] =	FALSE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	FALSE;
  format->status[IS_BUFFERABLE] =	BUFFER_NORMAL;
  format->size.x =			DEFAULT_PLATE_DIMENSION_X;
  format->size.y =			DEFAULT_PLATE_DIMENSION_Y;
  format->spacing.x =			DEFAULT_PLATE_SPACING_X;
  format->spacing.y =			DEFAULT_PLATE_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	0;

  format->data[PLATE_D_BULGE] =		BULGE_OUT;
  format->data[PLATE_D_RIM] =		FALSE;
  format->data[PLATE_D_RIM_THICK] =	3;

  format->gcs[GC_BODY] =		COLOR_CTRL_SURFACE_1;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_SURFACE_INV_1;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_TEXT_1;

  format->fonts[FONT_FIRST] =		FONT_HUGE;
  format->fonts[FONT_SECOND] =		FONT_HUGE;

  format->locate =			LABEL_SOUTH_EAST;

  format->format =			CONTROL_PLATE;
}



/******************************************************************************
  plate_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

plate_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_PLATE;
}



/******************************************************************************
  plate_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

plate_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char aux_line[100],
       *ptr,
       ch;  

  double factor,
         atof();

  strcpy (aux_line, line);
  strip_first (aux_line);

  factor = atof (aux_line);
  int_factor = atoi (aux_line);
 
  if (line[0] == 'A')
  {
    switch (aux_line[0])
    {
      /** IN **/

      case 'I':

        widget->data[PLATE_D_BULGE] = BULGE_IN;
        break;

      /** OUT **/

      case 'O':

        widget->data[PLATE_D_BULGE] = BULGE_OUT;
        break;

      /** FLAT **/

      case 'F':

        widget->data[PLATE_D_RIM] = FALSE;
        break;

      /** RIM **/

      case 'R':

        widget->data[PLATE_D_RIM] = TRUE;
        strip_first (aux_line);
        if (aux_line[0] >= '0' && aux_line[0] <= '9')
          widget->data[PLATE_D_RIM_THICK] = atoi (aux_line);
        break;

      /** TROUGH **/

      case 'T':

        widget->data[PLATE_D_RIM] = FALSE;
        widget->data[PLATE_D_BULGE] = BULGE_IN;
        break;
    }

    line[0] = '#';
  }
}



/******************************************************************************
  plate_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

plate_clean (widget)
  Widget *widget;
{
  if (!widget->state_count)
    widget->state = -1;
}



/******************************************************************************
  plate_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

plate_setup (widget)
  Widget *widget;
{

  widget->bound_offset.x = widget->thickness[THICKNESS_SHADOW] +
		widget->thickness[THICKNESS_ACTIVE];
  widget->bound_offset.y = widget->thickness[THICKNESS_SHADOW] +
		widget->thickness[THICKNESS_ACTIVE];
  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;  
  widget->bound_size.y = 2*widget->bound_offset.y + widget->size.y;     

  if (widget->state < 0)
  {
     widget->locate = LABEL_NONE;
     return;
  }
  else
    locate_text (widget, widget->label, 0, FONT_FIRST, TRUE);
}



/******************************************************************************
  plate_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

plate_execute (widget)
  Widget *widget;
{
  widget->flag = TRUE;
}



/******************************************************************************
  plate_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

plate_draw (widget)
  Widget *widget;
{
  short color_font_0,
        font_0,
        rim;

  color_font_0 = widget->gcs[GC_FONT_FIRST];
  font_0 = widget->fonts[FONT_FIRST];

  draw_generic_box (widget, widget->data[PLATE_D_BULGE],
				BULGE_KEEP, ACTIVE_HIDE);

  if (widget->data[PLATE_D_RIM])
  {
    rim = widget->data[PLATE_D_RIM_THICK];
    widget->size.x -= (2*widget->thickness[THICKNESS_SHADOW]+2*rim);
    widget->size.y -= (2*widget->thickness[THICKNESS_SHADOW]+2*rim);

    widget->position.x += (widget->thickness[THICKNESS_SHADOW]+rim);
    widget->position.y += (widget->thickness[THICKNESS_SHADOW]+rim);

    draw_generic_box (widget, !(widget->data[PLATE_D_BULGE]),
				BULGE_KEEP, ACTIVE_HIDE);

    widget->size.x += (2*widget->thickness[THICKNESS_SHADOW]+2*rim);
    widget->size.y += (2*widget->thickness[THICKNESS_SHADOW]+2*rim);

    widget->position.x -= (widget->thickness[THICKNESS_SHADOW]+rim);
    widget->position.y -= (widget->thickness[THICKNESS_SHADOW]+rim);
  }

  if (widget->locate != LABEL_NONE)
    draw_string (widget->drawable,
		widget->position.x, widget->position.y,
		widget->text_offset[0].x, widget->text_offset[0].y,
		widget->label, font_0,
		Win->hue_widget[color_font_0], FALSE);
}
