#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_scroll.h"


/******************************************************************************
  scroll_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

scroll_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      scroll_help (Console->formats[CONTROL_SCROLL]);
      break;

    case FUNCTION_FORMAT:
      scroll_format (Console->formats[CONTROL_SCROLL]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_SCROLL], line);
      break;

    case FUNCTION_STORE:
      scroll_store (widget, line);
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      scroll_initialize (widget);
      break;

    case FUNCTION_PARSE:
      scroll_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      scroll_clean (widget);
      break;

    case FUNCTION_SETUP:
      scroll_setup (widget);
      break;

    case FUNCTION_EDIT:
      break;

    case FUNCTION_SELECT:
      scroll_select (widget);
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      scroll_execute (widget);
      break;

    case FUNCTION_DRAW:
      scroll_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to scroll", NULL);
      break;
  }
}



/******************************************************************************
  scroll_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

scroll_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("SCROLL\n");
  printf ("    MIN    <float>                 minimum value\n");
  printf ("    MAX    <float>                 maximum value\n");
  printf ("    INIT   <float>                 initial value\n");
  printf ("    ALTER   LENGTH  <int>          width of vertical scroll (%d)\n",
			format->data[SCROLL_D_LENGTH]);
  printf ("    ALTER      <int>          width of vertical scroll (%d)\n",
			format->data[SCROLL_D_SHADOW_THICK]);
  printf ("    ALTER   COLOR   <gc>           gc index for scroll color (%d)\n",
			format->gcs[SCROLL_G_MARKER]);
  printf ("    ALTER   CLASSIC                use classic format\n");
  printf ("\n"); 
}



/******************************************************************************
  scroll_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

scroll_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	TRUE;
  format->status[IS_ACTIVATABLE] =	TRUE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	TRUE;
  format->status[IS_TIMER_ACTIVE] =	TRUE;
  format->status[IS_BUFFERABLE] =	BUFFER_NORMAL;

  format->size.x =			DEFAULT_SCROLL_DIMENSION_X;
  format->size.y =			DEFAULT_SCROLL_DIMENSION_Y;
  format->spacing.x =  			DEFAULT_SCROLL_SPACING_X;
  format->spacing.y =		   	DEFAULT_SCROLL_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	1;

  format->data[SCROLL_D_SHADOW_THICK] =	DEFAULT_SCROLL_SHADOW_THICK;
  format->data[SCROLL_D_ORIENTATION] =	DEFAULT_SCROLL_ORIENTATION;
  format->data[SCROLL_D_MARKER_SIZE] =	DEFAULT_SCROLL_MARKER_SIZE;
  format->data[SCROLL_D_ARROW_SIZE] =	DEFAULT_SCROLL_ARROW_SIZE;
  format->data[SCROLL_D_LENGTH] =	DEFAULT_SCROLL_LENGTH;
  format->data[SCROLL_D_BREADTH] =	DEFAULT_SCROLL_BREADTH;

  format->data_float[SCROLL_DF_ARROW_MOVE] =	DEFAULT_SCROLL_ARROW_MOVE;

  format->gcs[GC_BODY] =		COLOR_CTRL_BACK;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_BACK;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_DARK;
  format->gcs[SCROLL_G_MARKER] =	COLOR_CTRL_SURFACE;

  format->format =			CONTROL_SCROLL;
}



/******************************************************************************
  scroll_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

scroll_store (widget, line)
  Widget *widget;
  char *line;
{
  char subline[100];

  line[0] = '\0';
  sprintf (subline, "INIT\t%f\n", widget->value);
  strcat (line, subline);
}



/******************************************************************************
  scroll_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

scroll_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_SCROLL;
}



/******************************************************************************
  scroll_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

scroll_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100],
       *ptr,
       ch;  

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (tres_line);
  int_factor = atoi (tres_line);
 
  if (line[0] == 'A')
  {
    switch (dos_line[0])
    {
      /** MARKER **/

      case 'M':

        widget->data[SCROLL_D_MARKER_SIZE] = int_factor;
        break;

      /** ARROW **/

      case 'A':

        widget->data[SCROLL_D_ARROW_SIZE] = int_factor;
        break;

      /** SHADOW **/

      case 'S':

        widget->data[SCROLL_D_SHADOW_THICK] = int_factor;
        break;

      /** HORIZONTAL **/

      case 'H':

        widget->data[SCROLL_D_ORIENTATION] = SCROLL_HORIZONTAL;
        break;

      /** VERTICAL **/

      case 'V':

        widget->data[SCROLL_D_ORIENTATION] = SCROLL_VERTICAL;
        break;

      /** LENGTH **/
 
      case 'L':

        widget->data[SCROLL_D_LENGTH] = int_factor;
        break;

      /** BREADTH **/

      case 'B':

        widget->data[SCROLL_D_BREADTH] = int_factor;
        break;

      /** TICK **/

      case 'T':

        widget->data_float[SCROLL_DF_ARROW_MOVE] = factor;
        break;

      /** COLOR **/

      case 'C':

        widget->gcs[SCROLL_G_MARKER] = int_factor;
        break;
    }

    line[0] = '#';
  }
}



/******************************************************************************
  scroll_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

scroll_clean (widget)
  Widget *widget;
{
  if (widget->label == NULL)
    widget->label = strdup (widget->name);
}



/******************************************************************************
  scroll_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

scroll_setup (widget)
  Widget *widget;
{
  int i,
      arrow_dimension,
      thickness;

  /** Set scrollbar dimensions **/

  if (widget->size.x < 0) 
  {
    if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_HORIZONTAL)
      widget->size.x = widget->data[SCROLL_D_LENGTH];
    else
      widget->size.x = widget->data[SCROLL_D_BREADTH];
  }

  if (widget->size.y < 0) 
  {
    if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_HORIZONTAL)
      widget->size.y = widget->data[SCROLL_D_BREADTH];
    else
      widget->size.y = widget->data[SCROLL_D_LENGTH];
  }

  if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_HORIZONTAL)
  {
    widget->data[SCROLL_D_LENGTH] = widget->size.x;
    widget->data[SCROLL_D_BREADTH] = widget->size.y;
  }
  else
  {
    widget->data[SCROLL_D_LENGTH] = widget->size.y;
    widget->data[SCROLL_D_BREADTH] = widget->size.x;
  }

  if (widget->data[SCROLL_D_ARROW_SIZE] <= 0)
  {
    arrow_dimension = widget->data[SCROLL_D_BREADTH];
    if (arrow_dimension%2 == 0)
      arrow_dimension -= 1;

    widget->data[SCROLL_D_ARROW_SIZE] = arrow_dimension;
  }
  else
    arrow_dimension = widget->data[SCROLL_D_ARROW_SIZE];

  thickness = widget->data[SCROLL_D_SHADOW_THICK];

  /** Set arrow dimensions **/

  switch (widget->data[SCROLL_D_ORIENTATION])
  {
    case SCROLL_HORIZONTAL:

      /** Left arrow **/

      widget->text_offset[0].x = arrow_dimension - 2;
      widget->text_offset[0].y = widget->size.y/2 - arrow_dimension/2;

      widget->text_offset[1].x = arrow_dimension - 2;
      widget->text_offset[1].y = widget->size.y/2 +
					arrow_dimension/2 - thickness + 1;

      /** Right arrow **/

      widget->text_offset[2].x = widget->size.x - arrow_dimension;
      widget->text_offset[2].y = widget->size.y/2 - arrow_dimension/2;

      widget->text_offset[3].x = widget->size.x - arrow_dimension;
      widget->text_offset[3].y = widget->size.y/2 +
					arrow_dimension/2 - thickness + 1;
      break;

    case SCROLL_VERTICAL:

      /** Up arrow **/

      widget->text_offset[0].x = widget->size.x/2 - arrow_dimension/2;
      widget->text_offset[0].y = arrow_dimension - 2;

      widget->text_offset[1].x = widget->size.x/2 +
					arrow_dimension/2 - thickness + 1;
      widget->text_offset[1].y = arrow_dimension - 2;

      /** Down arrow **/

      widget->text_offset[2].x = widget->size.x/2 - arrow_dimension/2;
      widget->text_offset[2].y = widget->size.y - arrow_dimension;

      widget->text_offset[3].x = widget->size.x/2 +
					arrow_dimension/2 - thickness + 1;
      widget->text_offset[3].y = widget->size.y - arrow_dimension;
      break;
  }

  widget->data[SCROLL_D_ACTIVE_LENGTH] = widget->data[SCROLL_D_LENGTH] -
		widget->data[SCROLL_D_MARKER_SIZE] - 2*arrow_dimension - 2;

  widget->bound_offset.x = widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];
  widget->bound_offset.y = widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];

  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;
  widget->bound_size.y = 2*widget->bound_offset.y + widget->size.y;
}



/******************************************************************************
  scroll_select (widget)

  Handle type-specific function of widget <widget> when it is first selected
  by the user.
******************************************************************************/

scroll_select (widget)
  Widget *widget;
{
  int xpoint,
      mouse,
      scroll_width,
      half_scroll_width,
      scroll_edge,
      active_length;

  /** Set convenience variables **/

  scroll_width = widget->data[SCROLL_D_MARKER_SIZE];
  half_scroll_width = scroll_width/2;
  active_length = widget->data[SCROLL_D_ACTIVE_LENGTH];

  if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_VERTICAL)
  {
    mouse = Win->mouse.y;
    scroll_edge = widget->position.y + widget->data[SCROLL_D_ARROW_SIZE] + 1;
  }
  else
  {
    mouse = Win->mouse.x;
    scroll_edge = widget->position.x + widget->data[SCROLL_D_ARROW_SIZE] + 1;
  }

  /** Determine where cursor falls in scrollbar **/

  if (mouse < scroll_edge)
  {
    widget->status[IS_MOTION_ACTIVE] = FALSE;
    widget->data[SCROLL_D_ARROW_SELECTED] = 0;
  }
  else if (mouse > scroll_edge + active_length +
				widget->data[SCROLL_D_MARKER_SIZE])
  {
    widget->status[IS_MOTION_ACTIVE] = FALSE;
    widget->data[SCROLL_D_ARROW_SELECTED] = 1;
  }
  else
  {
    widget->status[IS_MOTION_ACTIVE] = TRUE;

    if (widget->minimum == widget->maximum)
      xpoint = 0;
    else
      xpoint = (int)((widget->value - widget->minimum)*(active_length)/
	(widget->maximum - widget->minimum) + half_scroll_width + 0.5);

    if (mouse - scroll_edge < xpoint-half_scroll_width ||
		mouse - scroll_edge > xpoint+half_scroll_width)
      widget->points[SCROLL_P_OFFSET].x = 0;
    else
      widget->points[SCROLL_P_OFFSET].x = mouse - scroll_edge - xpoint;
  }
}



/******************************************************************************
  scroll_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

scroll_execute (widget)
  Widget *widget;
{
  int i, j,
      shift_pixels,
      xpoint,
      mouse,
      scroll_width,
      half_scroll_width,
      scroll_edge,
      active_length;

  double value,
         value_delta;

  widget->flag = TRUE;
  scroll_width = widget->data[SCROLL_D_MARKER_SIZE];
  half_scroll_width = scroll_width/2;
  active_length = widget->data[SCROLL_D_ACTIVE_LENGTH];

  /** Handle an arrow press **/

  if (!widget->status[IS_MOTION_ACTIVE])
  {
    value_delta = (widget->maximum - widget->minimum) *
			widget->data_float[SCROLL_DF_ARROW_MOVE];

    if (widget->data[SCROLL_D_ARROW_SELECTED] == 0)
    {
      widget->value -= value_delta;
    }
    else
    {
      widget->value += value_delta;
    }

    if (widget->value < widget->minimum)
      widget->value = widget->minimum;
    else if (widget->value > widget->maximum)
      widget->value = widget->maximum;

    return;
  }

  if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_VERTICAL)
  {
    mouse = Win->mouse.y;
    scroll_edge = widget->position.y + widget->data[SCROLL_D_ARROW_SIZE] + 1;
  }
  else
  {
    mouse = Win->mouse.x;
    scroll_edge = widget->position.x + widget->data[SCROLL_D_ARROW_SIZE] + 1;
  }

  /** Determine and set value **/

  if (active_length == 0)
    value = widget->minimum;
  else
    value = (widget->maximum - widget->minimum)*
		(mouse - (scroll_edge + half_scroll_width) -
		widget->points[SCROLL_P_OFFSET].x)/
		(active_length) + widget->minimum;

  if (value < widget->minimum)
    value = widget->minimum;
  else if (value > widget->maximum)
    value = widget->maximum;
 
  Master->last_value = value;
  widget->value = value;
  widget->values[0] = value;
}



/******************************************************************************
  scroll_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

scroll_draw (widget)
  Widget *widget;
{
  int i, xpoint;

  short color_scroll,
        scroll_width,
        half_scroll_width,
        shadow_thickness,
        scroll_edge,
        active_length;

  if (widget->enable == FALSE)
    return;

  /** Set convenience variables **/

  color_scroll =	widget->gcs[SCROLL_G_MARKER];
  scroll_width =	widget->data[SCROLL_D_MARKER_SIZE];
  shadow_thickness =	widget->data[SCROLL_D_SHADOW_THICK];
  half_scroll_width =	scroll_width/2;
  active_length =	widget->data[SCROLL_D_ACTIVE_LENGTH];
  if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_VERTICAL)
    scroll_edge = widget->position.y + widget->data[SCROLL_D_ARROW_SIZE] + 1;
  else
    scroll_edge = widget->position.x + widget->data[SCROLL_D_ARROW_SIZE] + 1;

  /** Draw scrollbar base **/

  draw_generic_box (widget, BULGE_IN, BULGE_KEEP, ACTIVE_SHOW);

  if (widget->minimum == widget->maximum)
    xpoint = 0;
  else
    xpoint = (int)((widget->value - widget->minimum)*(active_length)/
		(widget->maximum - widget->minimum) + 0.5);

  if (xpoint < 0)
    xpoint = 0;
  else if (xpoint > active_length)
    xpoint = active_length;

  /** Draw indicator bar **/

  if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_VERTICAL)
  {
    XFillRectangle (Win->display, widget->drawable,
	Win->hue_widget[color_scroll],
	widget->position.x,
	scroll_edge + xpoint,
	widget->size.x, scroll_width);

    draw_border (widget->drawable,
	widget->position.x + shadow_thickness,
	scroll_edge + xpoint + shadow_thickness,
	widget->size.x-2*shadow_thickness, scroll_width-2*shadow_thickness,
	shadow_thickness,
	Win->hue_widget[COLOR_CTRL_HIGH],
	Win->hue_widget[COLOR_CTRL_SHADOW]);
  }
  else
  {
    XFillRectangle (Win->display, widget->drawable,
	Win->hue_widget[color_scroll],
	scroll_edge + xpoint,
	widget->position.y,
	scroll_width, widget->size.y);

    draw_border (widget->drawable,
	scroll_edge + xpoint + shadow_thickness,
	widget->position.y + shadow_thickness,
	scroll_width-2*shadow_thickness, widget->size.y-2*shadow_thickness,
	shadow_thickness,
	Win->hue_widget[COLOR_CTRL_HIGH],
	Win->hue_widget[COLOR_CTRL_SHADOW]);
  }

  scroll_draw_arrows (widget);
}



/******************************************************************************
  scroll_draw_arrows (widget)

  Draw the arrows specified in the scroll widget <widget>.
******************************************************************************/

scroll_draw_arrows (widget)
  Widget *widget;
{
  short i, orient,
        color_arrow,
        color_arrow_inv,
        color_top,
        color_bottom,
        color_flat,
        xoffset,
        yoffset,
        toffset,
        thickness,
        temp_thickness,
        dimension;

  XPoint start,
         moose,
         end;

  color_arrow = widget->gcs[SCROLL_G_MARKER];

  thickness = widget->data[SCROLL_D_SHADOW_THICK];
  dimension = widget->data[SCROLL_D_ARROW_SIZE];

  for (orient=0; orient<2; orient++)
  {
    if (!widget->status[IS_MOTION_ACTIVE] && widget->mode == MODE_SELECT &&
			orient == widget->data[SCROLL_D_ARROW_SELECTED])
    {
      color_bottom = COLOR_CTRL_HIGH;
      color_top = COLOR_CTRL_SHADOW;
    }
    else
    {
      color_top = COLOR_CTRL_HIGH;
      color_bottom = COLOR_CTRL_SHADOW;
    }

    if (widget->data[SCROLL_D_ORIENTATION] == SCROLL_HORIZONTAL)
    {
      if (orient) /** right arrow **/ 
        xoffset = 2;
      else
        xoffset = -2;
      toffset = xoffset/2;

      /** Draw the arrow body (extends through shadows) **/

      start.x = widget->position.x + widget->text_offset[2*orient+0].x;
      start.y = widget->position.y + widget->text_offset[2*orient+0].y;
      for (i=0; i<dimension/2; i++)
      {
        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_arrow],
			start.x+xoffset*i, start.y+i, 2, dimension-2*i);
      }

      /** Draw the slant shadow **/

      start.x = widget->position.x + widget->text_offset[2*orient+1].x;
      start.y = widget->position.y + widget->text_offset[2*orient+1].y;
      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness >= dimension/2+2)
        {
          temp_thickness--;
          start.y++;
        }

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x, start.y, 2, temp_thickness);
        start.x += xoffset;
        start.y -= 1;
      }
      XDrawPoint (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x+!orient, start.y+1);

      /** Draw the slant highlight **/

      start.x = widget->position.x + widget->text_offset[2*orient+0].x;
      start.y = widget->position.y + widget->text_offset[2*orient+0].y;
      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness > dimension/2)
          temp_thickness--;

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x, start.y, 2, temp_thickness);

        start.x += xoffset;
        start.y += 1;
      }

      /** Draw the flat highlight **/

      start.x = widget->position.x + widget->text_offset[2*orient+0].x;
      start.y = widget->position.y + widget->text_offset[2*orient+0].y;
      if (!orient)
        start.x++;
      for (i=0; i<thickness; i++)
      {
        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x, start.y + i/2,
			1, dimension - 1 - i - i/2);
       start.x += toffset;
      }
    }

    /** Else vertical scrollbar **/

    else
    {
      /** If down arrow **/

      if (orient)
      {
        color_flat = color_top;
        yoffset = 2;
      }
      else
      {
        color_flat = color_bottom;

        yoffset = -2;
      }
      toffset = yoffset/2;

      /** Draw the arrow body (extends through shadows) **/

      start.x = widget->position.x + widget->text_offset[2*orient+0].x;
      start.y = widget->position.y + widget->text_offset[2*orient+0].y;
      for (i=0; i<dimension/2; i++)
      {
        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_arrow],
			start.x+i, start.y+yoffset*i, dimension-2*i, 2);
      }

      /** Draw the slant shadow **/

      start.x = widget->position.x + widget->text_offset[2*orient+1].x;
      start.y = widget->position.y + widget->text_offset[2*orient+1].y;
      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness >= dimension/2+2)
        {
          temp_thickness--;
          start.x++;
        }

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x, start.y, temp_thickness, 2);
        start.x -= 1;
        start.y += yoffset;
      }
      XDrawPoint (Win->display, widget->drawable,
			Win->hue_widget[color_bottom],
			start.x+1, start.y+!orient);

      /** Draw the slant highlight **/

      start.x = widget->position.x + widget->text_offset[2*orient+0].x;
      start.y = widget->position.y + widget->text_offset[2*orient+0].y;
      temp_thickness = thickness;
      for (i=0; i<dimension/2; i++)
      {
        if (i+temp_thickness > dimension/2)
          temp_thickness--;

        XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_top],
			start.x, start.y, temp_thickness, 2);
        start.x += 1;
        start.y += yoffset;
      }

      /** Draw the flat highlight **/

      start.x = widget->position.x + widget->text_offset[2*orient+0].x;
      start.y = widget->position.y + widget->text_offset[2*orient+0].y;
      if (!orient)
        start.y++;
      for (i=0; i<thickness; i++)
      {
        /** up and down arrows have different shadow configuration **/

        if (orient)
          XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_flat],
			start.x + i/2, start.y,
			dimension - 1 - i - i/2, 1);
        else
          XFillRectangle (Win->display, widget->drawable,
			Win->hue_widget[color_flat],
			start.x + i + 1, start.y,
			dimension - 1 - i - i/2, 1);

        start.y += toffset;
      }
    }
  }
}
