#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_slider.h"


/******************************************************************************
  slider_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

slider_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      slider_help (Console->formats[CONTROL_SLIDER]);
      break;

    case FUNCTION_FORMAT:
      slider_format (Console->formats[CONTROL_SLIDER]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_SLIDER], line);
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_STORE:
      slider_store (widget, line);
      break;

    case FUNCTION_INITIALIZE:
      slider_initialize (widget);
      break;

    case FUNCTION_PARSE:
      slider_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      slider_clean (widget);
      break;

    case FUNCTION_SETUP:
      slider_setup (widget);
      break;

    case FUNCTION_EDIT:
      slider_edit (widget, line);
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      slider_execute (widget);
      break;

    case FUNCTION_DRAW:
      slider_draw (widget);
      break;

    default:
      throw_warning ("Function unknown to slider", NULL);
      break;
  }
}



/******************************************************************************
  slider_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

slider_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("SLIDER\n");
  printf ("    TEXT   <string>                slider label\n");
  printf ("    MIN    <float>                 minimum value\n");
  printf ("    MAX    <float>                 maximum value\n");
  printf ("    INIT   <float>                 initial value\n");
  printf ("    LOCATE <locate>                relative position of label\n");
  printf ("    ALTER   BAR     <int>          width of vertical bar (%d)\n",
			format->data[SLIDER_D_BAR]);
  printf ("    ALTER   TICKS   <bool>         (don't) use tickmarks (%d)\n",
			format->data[SLIDER_D_TICKS]);
  printf ("    ALTER   POINT   <bool>         (don't) use pointer (%d)\n",
			format->data[SLIDER_D_POINT]);
  printf ("    ALTER   DIGIT   <char>         formating of digits (%c)\n",
			(char)(format->data[SLIDER_D_DIGIT]));
  printf ("    ALTER   COLOR   <gc>           gc index for bar color (%d)\n",
			format->gcs[SLIDER_G_BAR]);
  printf ("    ALTER   GPOINT  <gc>           gc index for pointer (%d)\n",
			format->gcs[SLIDER_G_POINT]);
  printf ("    ALTER   CLASSIC                use classic format\n");
  printf ("\n"); 
}



/******************************************************************************
  slider_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

slider_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	TRUE;
  format->status[IS_ACTIVATABLE] =	TRUE;
  format->status[IS_EDITABLE] =		EDITABLE_FULL;
  format->status[IS_MOTION_ACTIVE] =	TRUE;
  format->status[IS_BUFFERABLE] =	BUFFER_NORMAL;

  format->size.x =			DEFAULT_SLIDER_DIMENSION_X;
  format->size.y =			DEFAULT_SLIDER_DIMENSION_Y;
  format->spacing.x =			DEFAULT_SLIDER_SPACING_X;
  format->spacing.y =	  		DEFAULT_SLIDER_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] =	2;
  format->thickness[THICKNESS_ACTIVE] =	1;

  format->data[SLIDER_D_DIGIT] =	DEFAULT_SLIDER_DIGIT;
  format->data[SLIDER_D_POINT] =	DEFAULT_SLIDER_POINT;
  format->data[SLIDER_D_TICKS] =	TRUE;
  format->data[SLIDER_D_BAR] =		FALSE;
  format->data[SLIDER_D_DIGIT_DECIMAL] = DEFAULT_SLIDER_DIGIT_DECIMAL;
  format->data[SLIDER_D_DIGIT_INTEGER] = DEFAULT_SLIDER_DIGIT_INTEGER;

  format->gcs[GC_BODY] =		COLOR_CTRL_AUX_2;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_DARK_INV;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_DARK;
  format->gcs[SLIDER_G_BAR] =		COLOR_CTRL_AUX_3;
  format->gcs[SLIDER_G_POINT] =		COLOR_CTRL_DARK;
 
  format->fonts[FONT_FIRST] =		FONT_MEDIUM;
  format->fonts[FONT_SECOND] =		FONT_MEDIUM;

  format->locate =			LABEL_WEST;

  format->format =			CONTROL_SLIDER;
}



/******************************************************************************
  slider_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

slider_store (widget, line)
  Widget *widget;
  char *line;
{
  char subline[100];

  line[0] = '\0';
  sprintf (subline, "INIT\t%f\n", widget->value);
  strcat (line, subline);
}



/******************************************************************************
  slider_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

slider_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_SLIDER;
}



/******************************************************************************
  slider_parse (widget, line)

  Parse any type-specific commands in string <line> for widget <widget>,
  setting internal values as needed.
******************************************************************************/

slider_parse (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor;

  char dos_line[100],
       tres_line[100],
       *ptr,
       ch;  

  double factor,
         atof();

  strcpy (dos_line, line);
  strip_first (dos_line);

  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (tres_line);
  int_factor = atoi (tres_line);
 
  if (line[0] == 'A')
  {
    switch (dos_line[0])
    {
      /** BAR **/

      case 'B':

        widget->data[SLIDER_D_BAR] = int_factor;
        break;

      /** TICKS **/

      case 'T':

        widget->data[SLIDER_D_TICKS] = int_factor;
        break;

      /** POINT **/

      case 'P':

        widget->data[SLIDER_D_POINT] = int_factor;
        break;

      /** GPOINT **/

      case 'G':

        widget->gcs[SLIDER_G_POINT] = int_factor;
        break;


      /** DIGIT **/

      case 'D':

        if (tres_line[0] >= '0' && tres_line[0] <= '9')
        {
          widget->data[SLIDER_D_DIGIT_INTEGER] = tres_line[0] - '0';
          widget->data[SLIDER_D_DIGIT_DECIMAL] = tres_line[2] - '0';
          widget->data[SLIDER_D_DIGIT] = 'Z';
        }
        else
          widget->data[SLIDER_D_DIGIT] = tres_line[0];
        break;

      /** COLOR or CLASSIC **/

      case 'C':

        if (dos_line[1] == 'L')
        {
          widget->subformat = CONTROL_SLIDER_CLASSIC;

          widget->gcs[GC_BODY] =       COLOR_CTRL_DARK;
          widget->gcs[GC_BODY_INV] =   COLOR_CTRL_DARK_INV;
          widget->gcs[GC_FONT_FIRST] = COLOR_CTRL_TEXT_2;
          widget->gcs[SLIDER_G_BAR] =  COLOR_CTRL_TEXT_3;

          widget->data[SLIDER_D_BAR] = DEFAULT_SLIDER_BAR_WIDTH;
          widget->data[SLIDER_D_TICKS] = FALSE;
          widget->data[SLIDER_D_POINT] = FALSE;

          widget->locate = LABEL_CENTER;
        }
        else
          widget->gcs[SLIDER_G_BAR] = int_factor;
        break;
    }

    line[0] = '#';
  }
}



/******************************************************************************
  slider_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

slider_clean (widget)
  Widget *widget;
{
  if (widget->label == NULL)
    widget->label = strdup (widget->name);
}



/******************************************************************************
  slider_setup (widget)

  Set up derived parameters (those not directly accessible to user) for
  widget <widget>.
******************************************************************************/

slider_setup (widget)
  Widget *widget;
{
  int height,
      junk,
      limit1, limit2;

  char line[100],
       string[100],
       number[20];

  XPoint text_start,
         text_size;

  XCharStruct text_information;

  /** Copy label to string **/

  if (widget->string == NULL && widget->label)
    widget->string = strdup (widget->label);

  /** If displaying a number **/

  if (widget->data[SLIDER_D_DIGIT] != 'N')
  {
    if (widget->data[SLIDER_D_DIGIT] != 'Z')
      throw_error ("Old ALTER DIGIT format no longer supported", NULL);

    /** Set label using formatting specification, using routine found	**/
    /** in pip_c_data.c.						**/

    data_set_label (widget, widget->data[SLIDER_D_DIGIT_INTEGER],
                widget->data[SLIDER_D_DIGIT_DECIMAL], TRUE);

    /** Position descriptive label **/

    if (widget->string)
      sprintf (line, "%s %s", widget->string, widget->label);
    else
      sprintf (line, "%s", widget->label);

    locate_text (widget, line, 0, FONT_FIRST, FALSE); 
  }
  else
  {
    if (widget->string)
      locate_text (widget, widget->string, 0, FONT_FIRST, FALSE); 
  }

  if (widget->locate != LABEL_CENTER)
    widget->status[IS_EDITABLE] = EDITABLE_NEVER;

  /** If slider has anything protuberances, make sure background is clear **/

  if (widget->data[SLIDER_D_TICKS] || widget->data[SLIDER_D_POINT] || 
	(widget->locate != LABEL_NONE && widget->locate != LABEL_CENTER))
    widget->status[IS_BUFFERABLE] = BUFFER_CLEAR;


  /** Set widget bounds, taking into account ticks and value marker **/
 
  limit1 = widget->size.y/2;
  limit2 = widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];

  if (widget->data[SLIDER_D_POINT] && limit1 > limit2)
    widget->bound_offset.x = limit1;
  else
    widget->bound_offset.x = limit2;

  widget->bound_offset.y = widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];

  if (widget->data[SLIDER_D_TICKS])
    widget->bound_offset.y += 6;

  widget->bound_size.x = 2*widget->bound_offset.x + widget->size.x;

  if (widget->data[SLIDER_D_POINT])
    widget->bound_size.y = widget->bound_offset.y + widget->size.y +
		(widget->size.y/2 - widget->thickness[THICKNESS_SHADOW]) + 1;
  else
    widget->bound_size.y = widget->bound_offset.y + widget->size.y +
			widget->thickness[THICKNESS_SHADOW] +
			widget->thickness[THICKNESS_ACTIVE];

  if (widget->locate != LABEL_NONE && widget->locate != LABEL_CENTER)
  {
    XTextExtents (Win->fonts[widget->fonts[FONT_FIRST]]->font,
                        widget->string, strlen (widget->string),
                        &junk, &junk, &junk, &text_information);

    text_start.x = widget->text_offset[0].x - text_information.lbearing;
    text_start.y = widget->text_offset[0].y - text_information.ascent;
    text_size.x = text_information.lbearing + text_information.rbearing;
    text_size.y = text_information.ascent + text_information.descent;

    if (text_start.x < 0)
    {
      widget->bound_offset.x = -text_start.x;
      widget->bound_size.x = widget->bound_offset.x + widget->size.x +
                widget->thickness[THICKNESS_SHADOW] +
                widget->thickness[THICKNESS_ACTIVE];
    }
    else if (text_start.x > widget->size.x)
    {
      widget->bound_size.x = widget->bound_offset.x +
                text_start.x + text_size.x;
    }

    if (text_start.y < 0)
    {
      widget->bound_offset.y = -text_start.y;
      widget->bound_size.y = widget->bound_offset.y + widget->size.y +
                widget->thickness[THICKNESS_SHADOW] +
                widget->thickness[THICKNESS_ACTIVE];
    }
    else if (text_start.y > widget->size.y)
    {
      widget->bound_size.y = widget->bound_offset.y +
                text_start.y + text_size.y;
    }
  }
}



/******************************************************************************
  slider_edit (widget, line)

  Manage editing of widget <widget>, based on input string <line>.  The
  internal value <widget->reserve> determines editing status.
******************************************************************************/

slider_edit (widget, line)
  Widget *widget;
  char *line;
{
  static char string_store[30];
  static short digit_store;
  char string_temp[100];
  double atof();

  switch (widget->reserve)
  {
    case EDIT_BEGIN:

      strcpy (string_store, widget->string);
      digit_store = widget->data[SLIDER_D_DIGIT];
      widget->data[SLIDER_D_DIGIT] = 'N';
      widget->reserve = EDIT_PROCESS;
      edit_text (widget);
      break;

    case EDIT_PROCESS:

      sprintf (string_temp, "%s %s", string_store, line);
      set_string (widget, string_temp);
      break;

    case EDIT_FINISH:

      widget->value = atof (line);
      set_string (widget, string_store);
      set_label (widget, line);
      widget->data[SLIDER_D_DIGIT] = digit_store;
      widget->reserve = EDIT_BEGIN;
      widget->flag = TRUE;
      break;
  }
}



/******************************************************************************
  slider_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

slider_execute (widget)
  Widget *widget;
{
  int i, j,
      split,
      done,
      minidone,
      slider_bar_width,
      newcommand;

  double value,
         yoffset;

  done = FALSE;

  widget->flag = TRUE;
  slider_bar_width = widget->data[SLIDER_D_BAR];

  value = widget->value;

  /** If right button, just set to last value **/

  if (Win->event.xbutton.button == Button3)
  {
    value = Master->last_value;
 
    if (value < widget->minimum)
      value = widget->minimum;
    else if (value > widget->maximum)
      value = widget->maximum;
  }

  /** Else normal button, determine value **/

  else
  {
    if (Win->mouse.x - widget->position.x > widget->size.x)
      value = widget->maximum;
    else if (slider_bar_width <= 1)
      value = (widget->maximum - widget->minimum)*
		(Win->mouse.x - widget->position.x)/widget->size.x +
		widget->minimum;
    else if (slider_bar_width >= widget->size.x)
      value = widget->minimum;
    else
      value = (widget->maximum - widget->minimum)*
		(Win->mouse.x - widget->position.x)/
		(widget->size.x - slider_bar_width) + widget->minimum;
  }
 
  Master->last_value = value;
 
  widget->value = value;
  widget->values[0] = value;
}



/******************************************************************************
  slider_draw (widget)

  Draw the widget <widget>.
******************************************************************************/

slider_draw (widget)
  Widget *widget;
{
  int i,
      xpoint;

  char line[100],
       string[100],
       number[20];

  short color_bar,
	color_font_0,
	color_point,
	color_background,
        font_0,
        slider_bar_width,
        use_slider_bar_width,
        shadow_thickness,
        active_thickness;

  double one_sixteenth;

  XPoint points[3],
         tick_position,
         tick_base;

  if (widget->enable == FALSE)
    return;

  /** Set convenience variables **/

  color_font_0 =	widget->gcs[GC_FONT_FIRST];
  color_bar =		widget->gcs[SLIDER_G_BAR];
  color_point =		widget->gcs[SLIDER_G_POINT];
  color_background =	widget->gcs[GC_BACKGROUND];
  font_0 =		widget->fonts[FONT_FIRST];

  slider_bar_width =	widget->data[SLIDER_D_BAR];
  if (slider_bar_width <= 1)
    slider_bar_width =	0;

  shadow_thickness =	widget->thickness[THICKNESS_SHADOW];
  active_thickness =	widget->thickness[THICKNESS_ACTIVE];

  /** Draw slider base **/

  draw_generic_box (widget, BULGE_IN, BULGE_KEEP, ACTIVE_SHOW);

  /** Set internal label **/

  if (widget->data[SLIDER_D_DIGIT] != 'N')
  {
    /** ALTER: Joint label and numeric string won't work **/

    data_set_label (widget, widget->data[SLIDER_D_DIGIT_INTEGER],
                widget->data[SLIDER_D_DIGIT_DECIMAL], FALSE);
    if (widget->string)
      sprintf (line, "%s %s", widget->string, widget->label);
    else
      sprintf (line, "%s", widget->label);
  }
  else
  {
    if (widget->string)
      sprintf (line, "%s", widget->string);
    else
      sprintf (line, "");
  }

  /** Determine value marker position **/

  if (widget->minimum == widget->maximum)
    xpoint = 0;
  else
    xpoint = (int)((widget->value - widget->minimum)*
		(widget->size.x-slider_bar_width)/
		(widget->maximum - widget->minimum) + 0.5);

  if (xpoint < 0)
    xpoint = 0;
  else if (xpoint > (widget->size.x-slider_bar_width))
    xpoint = widget->size.x-slider_bar_width;

  /** Draw value marker **/

  if (widget->data[SLIDER_D_BAR])
    XFillRectangle (Win->display, widget->drawable,
		Win->hue_widget[color_bar],
		widget->position.x + xpoint, widget->position.y,
		widget->data[SLIDER_D_BAR], widget->size.y);
  else
    XFillRectangle (Win->display, widget->drawable,
		Win->hue_widget[color_bar],
		widget->position.x, widget->position.y,
		xpoint, widget->size.y);

  /** Draw label **/

  if (widget->locate == LABEL_CENTER)
    draw_string (widget->drawable,
		widget->position.x, widget->position.y,
		widget->text_offset[0].x, widget->text_offset[0].y, line,
		font_0, Win->hue_widget[color_font_0], FALSE);
  else if (widget->locate != LABEL_NONE)
    draw_string (widget->drawable,
		widget->position.x, widget->position.y,
		widget->text_offset[0].x, widget->text_offset[0].y, line,
		font_0, Win->hue_widget[color_font_0], FALSE);


  /** Draw triangular value marker **/

  if (widget->data[SLIDER_D_POINT])
  {
    points[0].x = widget->position.x + xpoint;
    points[0].y = widget->position.y + widget->size.y/2;
  
    points[1].x = -widget->size.y/2;
    points[1].y = widget->size.y;

    points[2].x = widget->size.y;
    points[2].y = 0;

    XFillRectangle (Win->display, widget->drawable,
		Win->hue_widget[color_background],
		widget->position.x + points[1].x,
		widget->position.y + widget->size.y +
			shadow_thickness + active_thickness,
		widget->size.x+points[2].x, widget->size.y/2 + 2);

    XFillPolygon (Win->display, widget->drawable, Win->hue_widget[color_point],
		  points, 3, Convex, CoordModePrevious);
  }

  /** Draw ticks **/

  if (widget->data[SLIDER_D_TICKS])
  {
    one_sixteenth = ((double)(widget->size.x))/16.0;

    tick_base.y = widget->position.y - shadow_thickness - 1;
    tick_base.x = widget->position.x;
  
    for (i=0; i<=16; i++)
    {
      if (i%8 == 0)
        tick_position.y = tick_base.y - 6;
      else if (i%4 == 0)
        tick_position.y = tick_base.y - 5;
      else if (i%2 == 0)
        tick_position.y = tick_base.y - 3;
      else
        tick_position.y = tick_base.y - 2;

      tick_position.x = tick_base.x + (int)(one_sixteenth*i + 0.5);

      XDrawLine (Win->display, widget->drawable,
		Win->hue_widget[color_point],
		tick_position.x, tick_position.y,
		tick_position.x, tick_base.y);
    }
  }
}
