#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"
#include "pip_generic.h"

#include "pip_c_data.h"
#include "pip_c_slider.h"
#include "pip_c_slipper.h"


/******************************************************************************
  slipper_function (widget, line, function)

  Run the type-specific routine for widget <widget> that is associated with
  command <function>, using supplemental information from string <line>.
******************************************************************************/

slipper_function (widget, line, function)
  Widget *widget;
  char *line;
  short function;
{
  switch (function)
  {
    case FUNCTION_HELP:
      slipper_help (Console->formats[CONTROL_SLIPPER]);
      break;

    case FUNCTION_FORMAT:
      slipper_format (Console->formats[CONTROL_SLIPPER]);
      break;

    case FUNCTION_LOAD:
      parse_format_defaults (Console->formats[CONTROL_SLIPPER], line);
      break;

    case FUNCTION_STORE:
      slipper_store (widget, line);
      break;

    case FUNCTION_DUMP:
      generic_dump (widget);
      break;

    case FUNCTION_INITIALIZE:
      slipper_initialize (widget);
      break;

    case FUNCTION_PARSE:
      slider_parse (widget, line);
      break;

    case FUNCTION_CLEAN:
      slipper_clean (widget);
      break;

    case FUNCTION_SETUP:
      break;

    case FUNCTION_EDIT:
      slipper_edit (widget, line);
      break;

    case FUNCTION_SELECT:
      break;

    case FUNCTION_MOTION:
      break;
 
    case FUNCTION_RELEASE:
      break;

    case FUNCTION_EXECUTE:
      slipper_execute (widget, line);
      break;

    case FUNCTION_DRAW:
      slipper_draw (widget, line);
      break;

    default:
      throw_warning ("Function unknown to slipper", NULL);
      break;
  }
}



/******************************************************************************
  slipper_help (format)

  Print a help message to stdout detailing the usage of any type-specific
  arguments for the widget <format>, including default values when possible.
******************************************************************************/

slipper_help (format)
  Widget *format;
{
  printf ("\n");
  printf ("SLIPPER\n");
  printf ("    TEXT   <string>                slider label\n");
  printf ("    MIN    <float>                 minimum value\n");
  printf ("    MAX    <float>                 maximum value\n");
  printf ("    INIT   <float>                 initial value\n");
  printf ("    LOCATE <locate>                relative position of label\n");
  printf ("    ALTER   BAR     <int>          width of vertical bar (%d)\n",
                        format->data[SLIPPER_D_BAR]);
  printf ("    ALTER   TICKS   <bool>         (don't) use tickmarks (%d)\n",
                        format->data[SLIPPER_D_TICKS]);
  printf ("    ALTER   POINT   <bool>         (don't) use pointer (%d)\n",
                        format->data[SLIPPER_D_POINT]);
  printf ("    ALTER   DIGIT   <char>         formating of digits (%c)\n",
                        (char)(format->data[SLIPPER_D_DIGIT]));
  printf ("\n");
}



/******************************************************************************
  slipper_format (format)

  Set type-specific default for canonical widget <format>.
******************************************************************************/

slipper_format (format)
  Widget *format;
{
  format->status[IS_SELECTABLE] =	FALSE;
  format->status[IS_ACTIVATABLE] =	FALSE;
  format->status[IS_EDITABLE] =		EDITABLE_NEVER;
  format->status[IS_MOTION_ACTIVE] =	FALSE;
  format->status[IS_BUFFERABLE] =	FALSE;
  format->status[IS_COMPOSITE] =	TRUE;

  format->size.x =			DEFAULT_SLIPPER_DIMENSION_X;
  format->size.y =			DEFAULT_SLIPPER_DIMENSION_Y;
  format->spacing.x =			DEFAULT_SLIPPER_SPACING_X;
  format->spacing.y =			DEFAULT_SLIPPER_SPACING_Y;

  format->thickness[THICKNESS_SHADOW] = 2;
  format->thickness[THICKNESS_ACTIVE] = 1;

  format->data[SLIPPER_D_DIGIT] =	DEFAULT_SLIPPER_DIGIT;
  format->data[SLIPPER_D_POINT] =	DEFAULT_SLIPPER_POINT;
  format->data[SLIPPER_D_TICKS] =	TRUE;
  format->data[SLIPPER_D_BAR] =		FALSE;

  format->gcs[GC_BODY] =		COLOR_CTRL_AUX_2;
  format->gcs[GC_BODY_INV] =		COLOR_CTRL_DARK_INV;
  format->gcs[GC_FONT_FIRST] =		COLOR_CTRL_TEXT_2;
  format->gcs[SLIPPER_G_BAR] =		COLOR_CTRL_AUX_3;
  format->gcs[SLIPPER_G_POINT] =	COLOR_CTRL_DARK;

  format->fonts[FONT_FIRST] =		FONT_MEDIUM;
  format->fonts[FONT_SECOND] =		FONT_MEDIUM;

  format->format = CONTROL_SLIPPER;
}



/******************************************************************************
  slipper_store (widget, line)

  Copy type-specific initialization information for widget <widget> onto the
  end of string <line>, for later storage into a dump file.
******************************************************************************/

slipper_store (widget, line)
  Widget *widget;
  char *line;
{
  char subline[100];

  line[0] = '\0';
  sprintf (subline, "INIT\t%f\n", widget->value);
  strcat (line, subline);
}



/******************************************************************************
  slipper_initialize (widget)

  Initialize widget <widget> with non-canonical information.
******************************************************************************/

slipper_initialize (widget)
  Widget *widget;
{
  widget->format = CONTROL_SLIPPER;
}



/******************************************************************************
  slipper_clean (widget)

  Clear up any parameter discrepancies in widget <widget>.
******************************************************************************/

slipper_clean (widget)
  Widget *widget;
{
  char line[100];

  Widget *widget1,
         *widget2;

  Panel *panel;

  panel = Console->panels[Console->panel_count];

  /** Initialize slider widget **/

  widget1 = create_widget();
  widget1->panel = panel;
  panel->widgets[panel->widget_count] = widget1;
  panel->widget_count++;

  widget1->format = CONTROL_SLIDER;
  sprintf (line, "%s_slider", widget->name);
  widget1->name = strdup (line);
  copy_defaults (widget1, NULL);
  widget1->superformat = CONTROL_SLIPPER;
  widget1->relatives[0] = (void *)(widget);

  /** Initialize data widget **/

  widget2 = create_widget();
  widget2->panel = panel;
  panel->widgets[panel->widget_count] = widget2;
  panel->widget_count++;

  widget2->format = CONTROL_DATA;
  sprintf (line, "%s_data", widget->name);
  widget2->name = strdup (line);
  copy_defaults (widget2, NULL);
  widget2->superformat = CONTROL_SLIPPER;
  widget2->relatives[0] = (void *)(widget);

  /** Set up relatives **/

  widget->relatives[0] = (void *)(widget1);
  widget->relatives[1] = (void *)(widget2);

  /** Set up slider widget **/

  load_partial_widget (widget1, widget);
  widget1->data[SLIDER_D_DIGIT] = 'N';
  widget1->fonts[0] = widget->fonts[0];
  widget1->gcs[SLIPPER_G_BAR] = widget->gcs[SLIPPER_G_BAR];
  widget1->gcs[SLIPPER_G_POINT] = widget->gcs[SLIPPER_G_POINT];

  /** Set up data widget **/

  load_partial_widget (widget2, widget);
  free (widget2->link_names[LINK_BASIC]);
  widget2->link_names[LINK_BASIC] = strdup (widget1->name);
  widget2->fonts[0] = widget->fonts[0];
  widget2->size.y = -1;
  widget2->size.x = 50;
  widget2->position.x = LOCATE_RIGHT;
  widget2->position.y = LOCATE_SAME;
  set_label (widget2, NULL);
  set_string (widget2, NULL);
  widget2->gcs[GC_FONT_FIRST] = widget->gcs[GC_FONT_FIRST];

  widget2->data[DATA_D_DIGIT] = 'Z';
  widget2->data[DATA_D_DIGIT_INTEGER] = widget1->data[SLIDER_D_DIGIT_INTEGER];
  widget2->data[DATA_D_DIGIT_DECIMAL] = widget1->data[SLIDER_D_DIGIT_DECIMAL];
  widget2->data[DATA_D_OUTLINE] = FALSE;
  widget1->data[SLIDER_D_DIGIT_INTEGER] = -1;
  widget1->data[SLIDER_D_DIGIT_DECIMAL] = -1;

  widget_function (widget1, NULL, FUNCTION_CLEAN);
  widget_function (widget2, NULL, FUNCTION_CLEAN);
}



/******************************************************************************
  slipper_edit (widget, line)

  Manage editing of widget <widget>, based on input string <line>.  The
  internal value <widget->reserve> determines editing status.
******************************************************************************/

slipper_edit (widget, name)
  Widget *widget;
  char *name;
{
  Widget *widget1,
         *widget2;

  if (name == NULL)
    return;

  widget1 = (Widget *)(widget->relatives[0]);
  widget2 = (Widget *)(widget->relatives[1]);

  if (!strcmp(widget2->name, name))
  {
    if (widget2->reserve == EDIT_BEGIN)
    {
      widget1->value = widget2->value;
      widget1->flag = TRUE;
    }
  }
}



/******************************************************************************
  slipper_execute (widget)

  Execute the function of widget <widget>, as when mouse button is released
  or mouse motion occurs.
******************************************************************************/

slipper_execute (widget, name)
  Widget *widget;
  char *name;
{
  Widget *widget1,
         *widget2;

  if (name == NULL)
    return;

  widget1 = (Widget *)(widget->relatives[0]);
  widget2 = (Widget *)(widget->relatives[1]);

  /** If widget is slider **/

  if (!strcmp(widget1->name, name))
  {
    widget2->value = widget1->value;
    data_set_label (widget2, widget2->data[DATA_D_DIGIT_INTEGER],
		widget2->data[DATA_D_DIGIT_DECIMAL], FALSE);
    widget2->superformat = CONTROL_NONE;
    draw_widget (widget2);
    widget2->superformat = CONTROL_SLIPPER;
    XSync (Win->display, 0);
    widget2->flag = TRUE;
  }
  else
  {
    widget1->value = widget2->value;
    widget1->flag = TRUE;
  }
}



/******************************************************************************
  slipper_draw (widget, name)

  Draw the widget <widget>.
******************************************************************************/

slipper_draw (widget, name)
  Widget *widget;
  char *name;
{
  Widget *widget1,
         *widget2;

  if (name == NULL)
    return;

  widget1 = (Widget *)(widget->relatives[0]);
  widget2 = (Widget *)(widget->relatives[1]);

  if (!strcmp(widget1->name, name))
  {
    widget2->superformat = CONTROL_NONE;
    draw_widget (widget2);
    widget2->superformat = CONTROL_SLIPPER;
  }
  else
  {
    widget1->superformat = CONTROL_NONE;
    draw_widget (widget1);
    widget1->superformat = CONTROL_SLIPPER;
  }
  XSync (Win->display, 0);
}
