#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"


/******************************************************************************
  Widget *
  handle_widget_search (panel)

  Search <panel> for the smallest widget which the cursor is currently in,
  returning a pointer to that widget.
******************************************************************************/

Widget *
handle_widget_search (panel)
  Panel *panel;
{
  int i, j,
      x_offset, y_offset,
      min_size,
      height;

  Widget *widget,
         *best_widget;

  /** Set the mouse position **/

  set_mouse ();

  /** Find the smallest widget which the cursor is in **/

  min_size = 100000000;
  best_widget = NULL;
  for (i=0; i<panel->widget_count; i++)
  {
    widget = panel->widgets[i];

    /** Ignore disabled and unselectable widgets **/

    if (widget->enable == FALSE)
      continue;

    if (!widget->status[IS_SELECTABLE])
      continue;

    /** Composite widgets don't really exist **/

    if (widget->status[IS_COMPOSITE])
      continue;

    /** Check to see if cursor within widget **/

    if (widget->position.x <= Win->mouse.x &&
		widget->position.x+widget->size.x >= Win->mouse.x &&
		widget->position.y <= Win->mouse.y &&
		widget->position.y+widget->size.y >= Win->mouse.y)
    {
      if (widget->size.x <= min_size)
      {
        min_size = widget->size.x;
        best_widget = widget;
      }
    }
  }

  /** Set internal cursor position **/

  if (best_widget != NULL)
  {
    best_widget->cursor.x = Win->mouse.x;
    best_widget->cursor.y = Win->mouse.y;
  }

  /** Return widget pointer **/

  return (best_widget);
}



/******************************************************************************
  Widget *
  strong_widget_search (panel, x, y)

  Find the smallest widget containing (<x>,<y>), regardless of that widgets
  selectability, and return a pointer to the widget.
******************************************************************************/

Widget *
strong_widget_search (panel, x, y)
  Panel *panel;
  int x, y;
{
  int i, j,
      min_size;

  Widget *widget,
         *best_widget;

  /** Set cursor position **/

  set_mouse ();

  /** Find the smallest containing widget **/

  min_size = 100000000;
  best_widget = NULL;
  for (i=0; i<panel->widget_count; i++)
  {
    widget = panel->widgets[i];

    /** Still can't select disabled widget **/

    if (widget->enable == FALSE)
      continue;

    /** Composite widgets don't really exist **/

    if (widget->status[IS_COMPOSITE])
      continue;

    if (widget->position.x <= x &&
                widget->position.x+widget->size.x >= x &&
                widget->position.y <= y &&
                widget->position.y+widget->size.y >= y)
    {
      if (widget->size.x * widget->size.y <= min_size)
      {
        min_size = widget->size.x * widget->size.y;
        best_widget = widget;
      }
    }
  }

  /** Return widget **/

  return (best_widget);
}



/******************************************************************************
  Widget *
  handle_button_press (panel, widget)

  Manage a button press within widget <widget> of <panel>, appropriately
  setting active and select status.
******************************************************************************/

Widget *
handle_button_press (panel, widget)
  Panel *panel;
  Widget *widget;
{
  if (widget == NULL)
    return (widget);

  /** If selection by button onset, just return widget **/

  if (!Master->use_offset_activate)
    return (widget);

  /** Can't select unselectable widget **/

  if (!widget->status[IS_SELECTABLE])
    return (NULL);

  /** Set selected widget **/

  Console->active_widget = NULL;
  Console->select_widget = widget;
  Console->select_panel = panel;
  Console->select_mode = MODE_SELECT;
  widget = Console->select_widget;

  /** Run select function for widget **/

  widget_function (widget, "B", FUNCTION_SELECT);

  /** Set the type of event **/

  widget->status[IS_EVENT] = EVENT_PRESS;

  /** Set the widget mode to MODE_SELECT **/

  set_widget_mode (Console->select_widget, Console->select_mode);

  /** Return the widget if sensitive to motion (like SLIDER) **/

  if (widget->status[IS_MOTION_ACTIVE])
  {
    if (widget->status[IS_MOTION_ACTIVE] == MOTION_UPDATE)
    {
      draw_widget (widget);
      XSync (Win->display, 0);
      return (NULL);
    }
    else
      return (Console->select_widget);
  }
  else
    return (NULL);
}



/******************************************************************************
  Widget *
  handle_button_release (panel, widget)

  Manage a button release within widget <widget> of <panel>, appropriately
  setting active and select status.
******************************************************************************/

Widget *
handle_button_release (panel, widget)
  Panel *panel;
  Widget *widget;
{
  /** Don't do anything if widgets are set with button onset **/

  if (!Master->use_offset_activate)
    return (NULL);

  if (Console->select_widget == NULL)
    return (NULL);

  /** Set widget mode **/

  Console->select_widget->mode = MODE_NORMAL;

  /** If current widget same as previously selected widget **/

  if (Console->select_widget == widget)
  {
    Console->select_widget->status[IS_EVENT] = EVENT_RELEASE;
    Console->select_widget = NULL;
    return (widget);
  }

  /** Else button release on a different widget **/

  else
  {
    if (Console->select_widget->status[IS_ACTIVATABLE])
    {
      draw_widget (Console->select_widget);
      XSync (Win->display, 0);
    }
    Console->select_widget = NULL;
    return (NULL);
  }
}



/******************************************************************************
  Widget *
  handle_button_motion (panel, widget, event_type)

  Manage a mouse motion within widget <widget> of <panel>, appropriately
  setting active and select status.
******************************************************************************/

Widget *
handle_button_motion (panel, widget, event_type)
  Panel *panel;
  Widget *widget;
  int event_type;
{
  /** If no mouse dragging allowed, just return NULL)

  if (!Master->allow_drag)
    return (NULL);

  /** If set by button onset, just return NULL **/

  if (!Master->use_offset_activate)
    return (NULL);

  /** If no currently active widget, check for widget highlighting **/

  if (Console->select_widget == NULL)
  {
    /** If highlighting disabled, just return NULL **/

    if (!Master->use_motion_highlights)
      return (NULL);

    /** If there's an active widget, and it's the same as current widget **/

    if (Console->active_widget != NULL)
    {
      if (Console->active_widget != widget)
      {
        Console->active_mode = MODE_NORMAL;
        set_widget_mode (Console->active_widget, Console->active_mode);
        Console->active_widget->status[IS_EVENT] = event_type;
        Console->active_widget = NULL;
      }
    }

    /** If there is no active widget **/

    if (Console->active_widget == NULL)
    {
      if (widget == NULL)
        return (NULL);
        
      Console->active_widget = widget;
      Console->active_mode = MODE_ACTIVE;
      set_widget_mode (Console->active_widget, Console->active_mode);
      Console->active_widget->status[IS_EVENT] = event_type;
      return (NULL);
    }

    return (NULL);
  }

  /** If selected widget not same as current (only get here if there is	**/
  /** a currently selected widget.)					**/

  if (Console->select_widget != widget)
  {
    if (Console->select_mode == MODE_SELECT)
    {
      /** Set to active after first motion **/

      Console->select_mode = MODE_ACTIVE;
      set_widget_mode (Console->select_widget, Console->select_mode);
      Console->select_widget->status[IS_EVENT] = event_type;
    }
    return (NULL);
  }

  /** If the widget is active **/

  if (Console->select_mode == MODE_ACTIVE)
  {
    Console->select_mode = MODE_SELECT;
    set_widget_mode (Console->select_widget, Console->select_mode);
    Console->select_widget->status[IS_EVENT] = event_type;
  }

  /** Return widget if has some type of motion/timer sensitivity **/

  widget = Console->select_widget;
  if (event_type == EVENT_MOTION && widget->status[IS_MOTION_ACTIVE])
  {
    widget_function (widget, "B", FUNCTION_MOTION);

    /** Only return widget if application execution required **/

    if (widget->status[IS_MOTION_ACTIVE] == MOTION_UPDATE)
    {
      draw_widget (widget);
      XSync (Win->display, 0);
      return (NULL);
    }
    else
      return (Console->select_widget);
  }
  else if (event_type == EVENT_TIMER && widget->status[IS_TIMER_ACTIVE])
    return (Console->select_widget);
  else
    return (NULL);
}



/******************************************************************************
  Widget *
  handle_keypress (panel, widget)

  Determine what to do when widget <widget> of <panel> is under the mouse
  when a key is pressed.  Return a pointer to the widget if it is editable,
  else restore keypress event and run the edit function of the widget.
******************************************************************************/

Widget *
handle_keypress (panel, widget)
  Panel *panel;
  Widget *widget;
{
  if (widget == NULL)
    return (NULL);

  if (Console->select_widget != NULL)
    return (NULL);

  if (widget->status[IS_EDITABLE] == EDITABLE_ACTION)
    return (widget);
  if (widget->status[IS_EDITABLE] != EDITABLE_FULL)
    return (NULL);

  XPutBackEvent (Win->display, &Win->event);

  widget_function (widget, NULL, FUNCTION_EDIT);

  return (NULL);
}



/******************************************************************************
  set_widget_mode  (widget, mode)

  Set the mode of the widget specified by <widget> to <mode>.
******************************************************************************/

set_widget_mode  (widget, mode)
  Widget *widget;
  int mode;
{
  widget->mode = mode;

  /** CHECK: handle pixmaps **/

  if (widget->status[IS_ACTIVATABLE])
  {
    draw_widget (widget);
    XSync (Win->display, 0);
  }
}



/******************************************************************************
  draw_named_widget (name)

  Draw the widget specfied by <name>.
******************************************************************************/

draw_named_widget (name)
  char *name;
{
  Widget *widget;

  widget = get_named_widget (name);
  draw_widget (widget);
  XSync (Win->display, 0);
}



/******************************************************************************
  set_widget_state (widget, state)

  Set the state of the widget specified by <widget> to <state>.
******************************************************************************/

set_widget_state (widget, state)
  Widget *widget;
  int state;
{
  widget->state = state;
  draw_widget (widget);
  XSync (Win->display, 0);
}



/******************************************************************************
  set_named_widget_state (name, state)

  Set the state of the widget specified by <name> to <state>.
******************************************************************************/

set_named_widget_state (name, state)
  char name[];
  int state;
{
  Widget *widget;

  widget = get_named_widget (name);
  set_widget_state (widget, state);
}



/******************************************************************************
  set_widget_label (widget, new_label)

  Set the label of the widget specified by <name> to string <new_label>,
  centering and redrawing as necessary.
******************************************************************************/

set_widget_label (widget, new_label)
  Widget *widget;
  char new_label[];
{
  set_label (widget, new_label);

  if (widget->labels[0] != NULL)
  {
    free (widget->labels[0]);
    if (new_label != NULL)
      widget->labels[0] = strdup (new_label);
    else
      widget->labels[0] = NULL;
  }

  center_text (0, 0, widget->size.x, widget->size.y, widget->label,
		&(widget->text_offset[0].x), &(widget->text_offset[0].y),
		widget->fonts[FONT_0]);
  draw_widget (widget);
  XSync (Win->display, 0);
}



/******************************************************************************
  set_named_widget_label (name, new_label)

  Set the label of the widget specified by <name> to string <new_label>.
******************************************************************************/

set_named_widget_label (name, new_label)
  char name[],
       new_label[];
{
  Widget *widget;

  widget = get_named_widget (name);
  set_widget_label (widget, new_label);
}



/******************************************************************************
  set_widget_value (widget, value)

  Set double value of widget specified by <widget> to <value>, redrawing
  widget afterwards.
******************************************************************************/

set_widget_value (widget, value)
  Widget *widget;
  double value;
{
  if (value < widget->minimum)
    widget->value = widget->minimum;
  else if (value > widget->maximum)
    widget->value = widget->maximum;
  else
    widget->value = value;

  draw_widget (widget);
  XSync (Win->display, 0);
}



/******************************************************************************
  set_named_widget_value (name, value)

  Set double value of widget specified by <name> to <value>.
******************************************************************************/

set_named_widget_value (name, value)
  char name[];
  double value;
{
  Widget *widget;

  widget = get_named_widget (name);
  set_widget_value (widget, value);
}



/******************************************************************************
  set_widget_value_int (widget, value)

  Set integer value of widget specified by <name> to <value>.
******************************************************************************/

set_widget_value_int (widget, value)
  Widget *widget;
  int value;
{
  set_widget_value (widget, (double)(value));
}



/******************************************************************************
  set_named_widget_value_int (name, value)

  Set integer value of widget specified by <name> to <value>.
******************************************************************************/

set_named_widget_value_int (name, value)
  char name[];
  int value;
{
  set_named_widget_value (name, (double)(value));
}



/******************************************************************************
  double
  get_named_widget_value (name)

  Return value of widget specified by <name>.
******************************************************************************/

double
get_named_widget_value (name)
  char name[];
{
  Widget *widget;

  widget = get_named_widget (name);
  return (widget->value);
}



/******************************************************************************
  int
  get_named_widget_state (name)

  Return state of widget specified by <name>.
******************************************************************************/

int
get_named_widget_state (name)
  char name[];
{
  Widget *widget;

  widget = get_named_widget (name);
  return (widget->state);
}



/******************************************************************************
  Widget *
  get_named_widget_label (name, string, state_index)

  Return widget specified by <name>, putting state label <state_index> into
  <string>.
******************************************************************************/

Widget *
get_named_widget_label (name, string, state_index)
  int state_index;
  char name[],
       string[];
{
  int i;

  Widget *widget;

  widget = get_named_widget (name);

  if (state_index > 100)
  {
    if (widget->label == NULL)
      string = NULL;
    else
      strcpy (string, widget->label);
  }
  else if (state_index < 0)
    strcpy (string, widget->labels[widget->state]);
  else
    strcpy (string, widget->labels[state_index]);

  return (widget);
}



/******************************************************************************
  Widget *
  get_named_widget (name)

  Return pointer to widget specified by <name>.  If no widget, return NULL.
******************************************************************************/

Widget *
get_named_widget (name)
  char name[];
{
  int i;

  for (i=0; i<Console->panel->widget_count; i++)
    if (!strcmp (name, Console->panel->widgets[i]->name))
      return (Console->panel->widgets[i]);

  throw_error ("Request for unknown widget %s", name);
  return (NULL);
}



/******************************************************************************
  Widget *
  get_flagged_widget (name)

  Return pointer to widget specified by <name>, if an only if that widget's
  flag is set.
******************************************************************************/

Widget *
get_flagged_widget (name)
  char *name;
{
  int i;

  for (i=0; i<Console->panel->widget_count; i++)
  {
    if (Console->panel->widgets[i]->flag)
    {
      if (!strcmp (name, Console->panel->widgets[i]->name))
        return (Console->panel->widgets[i]);
    }
  }

  return (NULL);
}



/******************************************************************************
  Widget *
  find_flagged_widget ()

  Return pointer to the first widget of current panel which is flagged.  If
  not flags are set, return NULL.
******************************************************************************/

Widget *
find_flagged_widget ()
{
  int i;

  for (i=0; i<Console->panel->widget_count; i++)
  {
    if (Console->panel->widgets[i]->flag)
      return (Console->panel->widgets[i]);
  }

  return (NULL);
}



/******************************************************************************
  set_widget_flag (widget)

  Set flag of widget specified by <widget>.
******************************************************************************/

set_widget_flag (widget)
  Widget *widget;
{
  widget->flag = TRUE;
}



/******************************************************************************
  set_named_widget_flag (name)

  Set flag of widget specified by <name>.
******************************************************************************/

set_named_widget_flag (name)
  char *name;
{
  Widget *widget;

  widget = get_named_widget (name);
  widget->flag = TRUE;
}



/******************************************************************************
  reset_widget_flag (widget)

  Reset flag of widget specified by <widget>.
******************************************************************************/

reset_widget_flag (widget)
  Widget *widget;
{
  widget->flag = FALSE;
}



/******************************************************************************
  reset_named_widget_flag (name)

  Reset flag of widget specified by <name>.
******************************************************************************/

reset_named_widget_flag (name)
  char *name;
{
  Widget *widget;

  widget = get_named_widget (name);
  widget->flag = FALSE;
}



/******************************************************************************
  reset_all_flags ()

  Reset all widget flags of current panel.
******************************************************************************/

reset_all_flags ()
{
  int i;

  for (i=0; i<Console->panel->widget_count; i++)
    Console->panel->widgets[i]->flag = FALSE;
}



/******************************************************************************
  int
  check_widget_flag (widget)

  If widget <widget> has its flag set return TRUE, else return FALSE.
******************************************************************************/

int
check_widget_flag (widget)
  Widget *widget;
{
  if (widget != NULL && widget->flag)
    return (TRUE);
  else
    return (FALSE);
}



/******************************************************************************
  int
  check_named_widget_flag (widget, widget_name)

  If the widget specified by <widget_name> has its flag set return TRUE,
  else return FALSE.
******************************************************************************/

int
check_named_widget_flag (widget, widget_name)
  Widget **widget;
  char *widget_name;
{
  if ((*widget = get_flagged_widget (widget_name)) != NULL)
  {
    (*widget)->flag = FALSE;
    return (TRUE);
  }
  else
    return (FALSE);
}



/******************************************************************************
  Panel *
  get_named_panel (name)

  Search all panels for panel labeled <name>, returning either a pointer to
  that panel, or NULL.
******************************************************************************/

Panel *
get_named_panel (name)
  char name[];
{
  int i;

  for (i=0; i<Console->panel_count; i++)
    if (!strcmp (name, Console->panels[i]->name))
      return (Console->panels[i]);

  throw_error ("Request for unknown panel %s", name);
  return (NULL);
}



/******************************************************************************
  set_active_panel (panel)

  Set active (selected) panel to <panel>, as well as current panel if <panel>
  is not NULL.
******************************************************************************/

set_active_panel (panel)
  Panel *panel;
{
  Console->select_panel = Console->panel;
  if (panel != NULL)
    Console->panel = panel;
}



/******************************************************************************
  restore_panel ()

  Restore current panel to selected panel.
******************************************************************************/

restore_panel ()
{
  Console->panel = Console->select_panel;
}



/******************************************************************************
  set_mouse ()

  Set the mouse coordinates from global event structure.
******************************************************************************/

set_mouse ()
{
  Win->mouse.x = Win->event.xbutton.x;
  Win->mouse.y = Win->event.xbutton.y;
}
