#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"

#define MAX_FPS		10

/******************************************************************************
  load_panel_info ()

  Given a list of panel initialization files in <Console->panel_files[]>,
  load each of the panels sequentially.
******************************************************************************/

load_panel_info ()
{
  int i, j;

  char line[200],
       filename[100];

  double atof();

  FILE *fp,
       *fopen();

  /** NULL out each panel pointer **/

  for (i=0; i<MAX_PANELS; i++)
    Console->panels[i] = NULL;

  Console->panel_files[Console->panel_count][0] = '\0';
  Console->panel_count = 0;

  /** While there are still panels left to load **/

  while (Console->panel_files[Console->panel_count][0] != '\0')
  {
    /** Exit if too many panels **/

    if (Console->panel_count >= MAX_PANELS)
      throw_error ("Too many panels", NULL);

    /** Allocate new panel **/

    Console->panels[Console->panel_count] = (Panel *)(malloc(sizeof(Panel)));
    Console->panels[Console->panel_count]->index = Console->panel_count;

    /** Open panel file, if possible **/

    if ((fp = fopen (Console->panel_files[Console->panel_count], "r")) == NULL)
    {
      sprintf (line, "Unable to open panel %d widget file %s",
			Console->panel_count,
			Console->panel_files[Console->panel_count]);
      throw_error (line, NULL);
    }

    /** Load the widgets for this panel **/

    load_single_panel_info (Console->panels[Console->panel_count], fp);

    fclose (fp);

    Console->panel_count++;
  }

  /** Set base and current panels **/

  Console->base_panel = Console->panels[0];
  Console->panel = Console->base_panel;
}



/******************************************************************************
  load_single_panel_info (panel, fp)

  Load widgets for <panel> from file pointer <fp>.
******************************************************************************/

load_single_panel_info (panel, fp)
  Panel *panel;
  FILE *fp;
{
  int i, j,
      done;

  char line[100],
       dos_line[100],
       *ptr,
       *load_data_line(),
       ch;

  double factor,
         atof();

  Widget *widget;

  int fp_count;
  FILE *fps[MAX_FPS];

  /** Initialize panel to default settings **/

  initialize_panel (panel);
  sprintf (panel->name, "panel_%d", panel->index);

  /** Create enough of the base widget to get by  **/

  panel->widgets[0] = create_widget();
  panel->base_widget = panel->widgets[0];
  panel->base_widget->panel = panel;
  panel->base_widget->name = strdup ("base");
  panel->base_widget->format = CONTROL_X_BASE;
  copy_defaults (panel->base_widget, NULL);

  panel->widget_count++;

  /** Set file pointer stack **/

  fp_count = 1;
  fps[0] = fp;

  /** While there are valid widget lines left **/

  done = FALSE;

  while (!done)
  {
    if (load_data_line (line, fp) == NULL || line[0] == '~')
    {
      if (fp_count == 1)
      {
        done = TRUE;
      }
      else
      {
        fclose (fp);
        fp_count--;
        fp = fps[fp_count-1];
      }

      continue;
    }

    if (panel->widget_count >= MAX_CONTROLS)
      throw_error ("Too many widgets on panel %s", panel->name);

    /** Handle DEFAULT setting **/

    if (line[0] == 'D' && line[1] == 'E')
    {
      subload_widget_setup (line);
      continue;
    }

    /** Handle COLOR setting **/

    if (line[0] == 'C' && line[1] == 'O')
    {
      subload_widget_setup (line);
      continue;
    }

    /** Handle FONT setting **/

    if (line[0] == 'F' && line[1] == 'O')
    {
      subload_widget_setup (line);
      continue;
    }

    /** Handle FILE setting **/

    if (line[0] == 'F' && line[1] == 'I')
    {
      if (fp_count >= MAX_FPS)
        throw_error ("Too many nested subfiles", NULL);

      strip_first (line);
      if ((fps[fp_count] = fopen (line, "r")) == NULL)
        throw_error ("Unable to open subfile %s", line);

      fp = fps[fp_count++];
      continue;
    }

    /** Allocate a widget **/

    panel->widgets[panel->widget_count] = create_widget();
    widget = panel->widgets[panel->widget_count];
    widget->panel = panel;
    panel->widget_count++;

    /** Determine the name of the widget **/

    strip_string (line);
    widget->name = strdup (line);

    /** Load the widget type **/

    load_data_line (line, fp);

    /** If widget type isn't the PANEL itself **/

    if (widget_function (widget, line, FUNCTION_INITIALIZE))
    {
      /** If widget is the base (usually a PLATE) **/

      if (widget->format == CONTROL_X_BASE)
      {
        /** Parse and clean widget **/

        parse_entire_widget (panel->base_widget, fp);
        widget_function (panel->base_widget, NULL, FUNCTION_CLEAN);

        /** The base doesn't count as a normal widget **/

        free (widget);
        panel->widget_count--;
        panel->widgets[panel->widget_count] = NULL;
      }

      /** Else widget is just a normal widget **/

      else
      {
        /** Load defaults, parse, and clean widget **/

        copy_defaults (widget, NULL);
        parse_entire_widget (widget, fp);
        widget_function (widget, NULL, FUNCTION_CLEAN);
      }
    }

    /** Else widget is PANEL itself **/

    else
    {
      /** Parse panel **/

      strcpy (panel->name, widget->name);
      parse_entire_panel (panel, fp);

      /** Panel doesn't count as a normal widget **/

      free (widget);
      panel->widget_count--;
      panel->widgets[panel->widget_count] = NULL;
    }
  }
}


 
/******************************************************************************
  Widget *
  create_widget ()

  Create a new widget.
******************************************************************************/

Widget *
create_widget ()
{
  Widget *widget;

  widget = (Widget *)(malloc(sizeof(Widget)));
  widget->label = NULL;
  widget->string = NULL;

  return (widget);
}


/******************************************************************************
  copy_defaults (dest_widget, raw_src_widgets)

  Load the defeault settings for widget <dest_widget> from widget
  <raw_src_widget>.  If the latter is NULL, copy from the formatting default.
******************************************************************************/

copy_defaults (dest_widget, raw_src_widget)
  Widget *dest_widget,
         *raw_src_widget;
{
  int i;
  Widget *src_widget;

  if (raw_src_widget == NULL)
    src_widget = Console->formats[dest_widget->format];
  else
    src_widget = raw_src_widget;

  for (i=0; i<MAX_LINKS; i++)
    dest_widget->link_names[i] =	NULL;
  dest_widget->comment =		strdup ("NONE");

  dest_widget->format =			src_widget->format;
  dest_widget->subformat =		src_widget->subformat;
  dest_widget->superformat =		src_widget->superformat;
  dest_widget->mode =			src_widget->mode;
  dest_widget->flag =			src_widget->flag;
  dest_widget->enable =			src_widget->enable;
  dest_widget->reserve =		src_widget->reserve;
  dest_widget->user =			src_widget->user;
  dest_widget->locate =			src_widget->locate;
  for (i=0; i<MAX_STATUS; i++)
    dest_widget->status[i] =		src_widget->status[i];

  for (i=0; i<MAX_RELATIVES; i++)
    dest_widget->relatives[i] =	NULL;

  dest_widget->size.x =			src_widget->size.x;
  dest_widget->size.y =			src_widget->size.y;
  dest_widget->position.x =		src_widget->position.x;
  dest_widget->position.y =		src_widget->position.y;
  dest_widget->real_position.x =	src_widget->real_position.x;
  dest_widget->real_position.y =	src_widget->real_position.y;
  dest_widget->spacing.x =		src_widget->spacing.x;
  dest_widget->spacing.y =		src_widget->spacing.y;
  dest_widget->delta.x =		src_widget->delta.x;
  dest_widget->delta.y =		src_widget->delta.y;
  dest_widget->cursor.x =		src_widget->cursor.x;
  dest_widget->cursor.y =		src_widget->cursor.y;
  dest_widget->locate_offset.x =	src_widget->locate_offset.x;
  dest_widget->locate_offset.y =	src_widget->locate_offset.y;
  for (i=0; i<MAX_STATES; i++)
  {
    dest_widget->text_offset[i].x =	src_widget->text_offset[i].x;
    dest_widget->text_offset[i].y =	src_widget->text_offset[i].y;
  }
  for (i=0; i<MAX_POINTS; i++)
  {
    dest_widget->points[i].x =		src_widget->points[i].x;
    dest_widget->points[i].y =		src_widget->points[i].y;
  }
  dest_widget->bound_offset.x =		src_widget->bound_offset.x;
  dest_widget->bound_offset.y =		src_widget->bound_offset.y;
  dest_widget->bound_size.x =		src_widget->bound_size.x;
  dest_widget->bound_size.y =		src_widget->bound_size.y;

  for (i=0; i<MAX_THICKNESS; i++)
    dest_widget->thickness[i] =		src_widget->thickness[i];

  /** Free up old label and set new one **/

  set_label (dest_widget, src_widget->label);

  /** Free up old string and set new one **/

  set_string (dest_widget, src_widget->string);

  for (i=0; i<MAX_LABELS; i++)
    dest_widget->labels[i] =		NULL;
  dest_widget->superlabels = 		NULL;

  for (i=0; i<MAX_FONTS; i++)
    dest_widget->fonts[i] =		src_widget->fonts[i];
  for (i=0; i<MAX_GCS; i++)
    dest_widget->gcs[i] =		src_widget->gcs[i];

  dest_widget->state_count =		src_widget->state_count;
  dest_widget->state =			src_widget->state;
  dest_widget->next_state =		src_widget->next_state;
  for (i=0; i<MAX_STATES; i++)
    dest_widget->states[i] =		src_widget->states[i];

  dest_widget->value_count =		src_widget->value_count;
  dest_widget->value =			src_widget->value;
  dest_widget->next_value =		src_widget->next_value;
  for (i=0; i<MAX_VALUES; i++)
    dest_widget->values[i] =		src_widget->values[i];
  dest_widget->minimum =		src_widget->minimum;
  dest_widget->maximum =		src_widget->maximum;

  for (i=0; i<MAX_DATA; i++)
  {
    dest_widget->data[i] =		src_widget->data[i];
    dest_widget->data_float[i] =	src_widget->data_float[i];
  }

  dest_widget->graphic_count =		src_widget->graphic_count;
  for (i=0; i<MAX_GRAPHICS; i++)
    dest_widget->graphics[i] =		NULL;
}



/******************************************************************************
  load_partial_widget (dest_widget, src_widget)

  Load most elements of widget <src_widget> into widget <dest_widget>,
  omitting only certain font and gc information.
******************************************************************************/

load_partial_widget (dest_widget, src_widget)
  Widget *dest_widget,
         *src_widget;
{
  int i;

  for (i=0; i<MAX_LINKS; i++)
  {
    if (src_widget->link_names[i] != NULL)
      dest_widget->link_names[i] = strdup (src_widget->link_names[i]);
    else
      dest_widget->link_names[i] = NULL;
  }
  if (src_widget->comment != NULL)
    dest_widget->comment = strdup (src_widget->comment);

  dest_widget->enable =		src_widget->enable;
  dest_widget->locate =		src_widget->locate;

  dest_widget->size.x =		src_widget->size.x;
  dest_widget->size.y =		src_widget->size.y;
  dest_widget->position.x =		src_widget->position.x;
  dest_widget->position.y =		src_widget->position.y;
  dest_widget->real_position.x =	src_widget->real_position.x;
  dest_widget->real_position.y =	src_widget->real_position.y;
  dest_widget->spacing.x =		src_widget->spacing.x;
  dest_widget->spacing.y =		src_widget->spacing.y;
  dest_widget->delta.x =		src_widget->delta.x;
  dest_widget->delta.y =		src_widget->delta.y;
  dest_widget->locate_offset.x =	src_widget->locate_offset.x;
  dest_widget->locate_offset.y =	src_widget->locate_offset.y;

  for (i=0; i<MAX_POINTS; i++)
  {
    dest_widget->points[i].x =		src_widget->points[i].x;
    dest_widget->points[i].y =		src_widget->points[i].y;
  }

  for (i=0; i<3; i++)
    dest_widget->thickness[i] =	src_widget->thickness[i];

  /** Free up old label and set new one **/

  set_label (dest_widget, src_widget->label);

  /** Free up old string and set new one **/

  set_string (dest_widget, src_widget->string);

/**
  for (i=0; i<MAX_FONTS; i++)
    dest_widget->fonts[i] =		src_widget->fonts[i];
  for (i=0; i<MAX_GCS; i++)
    dest_widget->gcs[i] =		src_widget->gcs[i];
**/

  dest_widget->state_count =		src_widget->state_count;
  dest_widget->state =			src_widget->state;
  dest_widget->next_state =		src_widget->next_state;
  for (i=0; i<MAX_STATES; i++)
    dest_widget->states[i] =		src_widget->states[i];

  dest_widget->value_count =		src_widget->value_count;
  dest_widget->value =			src_widget->value;
  dest_widget->next_value =		src_widget->next_value;
  for (i=0; i<MAX_VALUES; i++)
    dest_widget->values[i] =		src_widget->values[i];
  dest_widget->minimum =		src_widget->minimum;
  dest_widget->maximum =		src_widget->maximum;

  for (i=0; i<MAX_DATA; i++)
  {
    dest_widget->data[i] =		src_widget->data[i];
    dest_widget->data_float[i] =	src_widget->data_float[i];
  }
}



/******************************************************************************
  parse_entire_widget (widget, fp)

  Parse all the widget elements in file <fp>, initializing widget <widget>
  accordingly.
******************************************************************************/

parse_entire_widget (widget, fp)
  Widget *widget;
  FILE *fp;
{
  char line[200],
       *ptr;

  ptr=fgets (line, 200, fp);
  while (line[0] != ' ' && line[0] != '\n' && line[0] != '\0' && ptr != NULL)
  {
    parse_configuration (widget, line);
    ptr=fgets (line, 200, fp);
  }
}



/******************************************************************************
  parse_configuration (widget, line)

  Parse string <line>, initializing widget <widget> accordingly.
******************************************************************************/

parse_configuration (widget, line)
  Widget *widget;
  char *line;
{
  int i, j,
      int_factor,
      pix_index,
      gc_index,
      font_index;

  char dos_line[100],
       tres_line[100],
       quad_line[100],
       ch;

  double factor,
         atof();

  XColor color;

  /** Parse any widget-specific parameters **/

  widget_function (widget, line, FUNCTION_PARSE);

  /** Create some convenience strings and values **/

  strcpy (dos_line, line);
  strip_first (dos_line);
  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (dos_line);
  int_factor = atoi (dos_line);

  /** Determine the argument type **/

  switch (line[0])
  {
    /** BITMAP **/

    case 'B':

      if (dos_line[0] < '0' || dos_line[0] > '9')
        pix_index = 0;
      else
        pix_index = dos_line[0] - '0';

      if (pix_index >= MAX_STATES || pix_index >= MAX_GRAPHICS)
        break;

      if (widget->graphics[pix_index] == NULL)
      {
        widget->graphics[pix_index]= (Graphic *)(malloc(sizeof(Graphic)));
        if (pix_index+1 > widget->state_count)
        {
          widget->graphics[pix_index]->gc =	1;
          widget->graphic_count =		pix_index+1;
          widget->state_count =			pix_index+1;
        }
      }

      if (dos_line[0] < '0' || dos_line[0] > '9')
        strcpy (widget->graphics[pix_index]->filename, dos_line);
      else
        strcpy (widget->graphics[pix_index]->filename, tres_line);
      widget->graphics[pix_index]->is_bitmap = TRUE;

      break;


    /** COMMENT or COLUMN **/

    case 'C':

      if (line[3] == 'M')	/** COMMENT **/
      {
        free (widget->comment);
        widget->comment = strdup (dos_line);
        break;
      }
      else			/** COLUMN **/
      {
        switch (dos_line[0])
        {
          case 'I':
          case 'R':
            strip_first (dos_line);
            if (dos_line[0] == 'O' || dos_line[0] == 'P')
              widget->position.x = LOCATE_OVERLAY_RIGHT;
            else
              widget->position.x = LOCATE_RIGHT;
            break;

          case 'D':
          case 'L':
            strip_first (dos_line);
            if (dos_line[0] == 'O' || dos_line[0] == 'P')
              widget->position.x = LOCATE_OVERLAY_LEFT;
            else
              widget->position.x = LOCATE_LEFT;
            break;

          case 'S':
            widget->position.x = LOCATE_SAME;
            break;

          case 'P':
            widget->position.x = LOCATE_OVERLAY_RIGHT;
            break;

          case 'Q':
            widget->position.x = LOCATE_OVERLAY_LEFT;
            break;

          case 'C':
            strip_first (dos_line);
            if (dos_line[0] == 'O' || dos_line[0] == 'P')
              widget->position.x = LOCATE_OVERLAY_CENTER;
            else
              widget->position.x = LOCATE_CENTER;
            break;

          default:
            widget->position.x = (int)(factor + 0.5);
            break;
        }
      }
      break;

    /** column (outdated) **/

    case 'c':

      widget->position.x = (int)(widget->spacing.x +
		factor*(widget->spacing.x+widget->size.x) + 0.5);
      break;

    /** DATA or DELTA **/

    case 'D':
    case 'd':

      if (line[1] == 'A' || line[1] == 'a')	/** DATA **/
      {
        if (dos_line[0] == 'F')
        {
          int_factor = atoi (tres_line);
          if (int_factor >= MAX_DATA)
            break;
          strip_first (tres_line);
          widget->data_float[int_factor] = atof (tres_line);
          break;
        }

        if (int_factor >= MAX_DATA)
          break;

        if (tres_line[0] < '0' || tres_line[0] > '9')
          widget->data[int_factor] = (short)(tres_line[0]);
        else
          widget->data[int_factor] = atoi (tres_line);
      }
      else
      {
        if (dos_line[0] == 'X')
          widget->delta.x = atoi (tres_line);
        else if (dos_line[0] == 'Y')
          widget->delta.y = atoi (tres_line);
      }

      break;

    /** FONT **/

    case 'F':

      if (tres_line[0] < '0' || tres_line[0] > '9')
      {
        font_index = load_typeset (tres_line);
        if (font_index < 0)
          break;
      }
      else
        font_index = atoi (tres_line);

      switch (dos_line[0])
      {
        case 'F':
        case 'P':
          widget->fonts[0] = font_index;
          break;

        case 'S':
        case 'A':
          widget->fonts[1] = font_index;
          break;

        case 'T':
          widget->fonts[2] = font_index;
          break;

        default:
          if (int_factor < MAX_FONTS)
            widget->fonts[int_factor] = font_index;
      }
      break;

    /** GRAPHIC **/
    /** PIXMAP **/

    case 'G':
    case 'P':

      if (dos_line[0] < '0' || dos_line[0] > '9')
        pix_index = 0;
      else
        pix_index = dos_line[0] - '0';

      if (pix_index >= MAX_STATES || pix_index >= MAX_GRAPHICS)
        break;

      if (widget->graphics[pix_index] == NULL)
      {
        widget->graphics[pix_index]= (Graphic *)(malloc(sizeof(Graphic)));

        if (pix_index+1 > widget->state_count)
        {
          widget->graphic_count = pix_index+1;
          widget->state_count = pix_index+1;
        }
      }

      if (dos_line[0] < '0' || dos_line[0] > '9')
        strcpy (widget->graphics[pix_index]->filename, dos_line);
      else
        strcpy (widget->graphics[pix_index]->filename, tres_line);
      widget->graphics[pix_index]->is_bitmap = FALSE;

      break;

    /** HEIGHT **/

    case 'H':

      if (dos_line[0] == 'S')
        widget->size.y = DIMENSION_SAME;
      else if (dos_line[0] == 'T')
        widget->size.y = DIMENSION_TAFFY;
      else if (dos_line[0] == 'P')
        widget->size.y = DIMENSION_PANEL;
      else if (dos_line[0] < '0' || dos_line[0] > '9')
        widget->size.y = DIMENSION_SAME;
      else
        widget->size.y = (int)(factor+0.5);
      break;

    /** height (outdated) **/

    case 'h':

      widget->size.y = int_factor*Console->formats[0]->size.y +
			(int_factor-1)*Console->formats[0]->spacing.y;
      break;

    /** INIT **/

    case 'I':

      int_factor = atoi (tres_line);
      if (int_factor >= MAX_STATES)
        break;

      switch (dos_line[0])
      {
        case 'S':
          strip_first (tres_line);
          widget->states[int_factor] = atoi (tres_line);
          break;

        case 'V':
          strip_first (tres_line);
          widget->values[int_factor] = atof (tres_line);
          break;

        default:
          widget->states[int_factor] = 1;
          widget->value = factor;
          widget->values[0] = factor;
          widget->value_count = 1;
          break;
      }
      break;

    /** LABEL or LOCATE **/

    case 'L':

      switch (dos_line[0])
      {
        case 'X':
          widget->locate_offset.x = atoi (tres_line);
          break;

        case 'Y':
          widget->locate_offset.y = atoi (tres_line);
          break;

        case 'R':
            widget->locate = LABEL_EAST;
            break;
        case 'L':
            widget->locate = LABEL_WEST;
            break;
        case '0':
            widget->locate = LABEL_NORTH;
            break;
        case '1':
            widget->locate = LABEL_SOUTH;
            break;
        case '2':
            widget->locate = LABEL_NORTH_WEST;
            break;
        case '3':
            widget->locate = LABEL_NORTH_EAST;
            break;
        case '4':
            widget->locate = LABEL_SOUTH_WEST;
            break;
        case '5':
            widget->locate = LABEL_SOUTH_EAST;
            break;

        case 'N':
          if (dos_line[1] == 'W' || tres_line[0] == 'W')
            widget->locate = LABEL_NORTH_WEST;
          else if (dos_line[1] == 'E' || tres_line[0] == 'E')
            widget->locate = LABEL_NORTH_EAST;
          else
            widget->locate = LABEL_NORTH;
          break;

        case 'S':
          if (dos_line[1] == 'W' || tres_line[0] == 'W')
            widget->locate = LABEL_SOUTH_WEST;
          else if (dos_line[1] == 'E' || tres_line[0] == 'E')
            widget->locate = LABEL_SOUTH_EAST;
          else
            widget->locate = LABEL_SOUTH;
          break;

        case 'E':
          widget->locate = LABEL_EAST;
          break;

        case 'W':
          widget->locate = LABEL_WEST;
          break;

        case 'C':
          widget->locate = LABEL_CENTER;
          break;

        case 'Z':
          widget->locate = LABEL_NONE;
          break;

        case 'G':
          widget->locate = LABEL_GENERIC;
          break;

        default:
          throw_warning ("Invalid locate in %s", widget->name);
          break;
      }
      break;

    /** MIN or MAX **/

    case 'M':

      if (line[1] == 'I')
        widget->minimum = factor;
      else
        widget->maximum = factor;
      break;

    /** NEXUS **/

    case 'N':

      switch (dos_line[0])
      {
        case 'W':
          if (dos_line[1] == 'I')
            widget->link_names[LINK_WIDTH] = strdup (tres_line);
          else
            widget->link_names[LINK_BASIC] = strdup (dos_line);
          break;

        case 'H':
          if (dos_line[1] == 'E')
            widget->link_names[LINK_HEIGHT] = strdup (tres_line);
          else
            widget->link_names[LINK_BASIC] = strdup (dos_line);
          break;

        case 'X':
          widget->link_names[LINK_X] = strdup (tres_line);
          break;

        case 'Y':
          widget->link_names[LINK_Y] = strdup (tres_line);
          break;

        default:
          widget->link_names[LINK_BASIC] = strdup (dos_line);
          break;
      }
      break;

    /** OVER **/

    case 'O':
    case 'o':

      switch (dos_line[0])
      {
        case 'A':
        case 'a':
          widget->status[IS_ACTIVATABLE] = atoi (tres_line);
          break;

        case 'B':
        case 'b':
          widget->status[IS_BUFFERABLE] = atoi (tres_line);
          break;

        case 'E':
        case 'e':
          widget->status[IS_EDITABLE] = atoi (tres_line);
          break;

        case 'M':
        case 'm':
          widget->status[IS_MOTION_ACTIVE] = atoi (tres_line);
          break;

        case 'T':
        case 't':
          widget->status[IS_TIMER_ACTIVE] = atoi (tres_line);
          break;

        case 'S':
          widget->status[IS_SELECTABLE] = atoi (tres_line);
          break;
       }
       break;

    /** ROW **/

    case 'R':

      switch (dos_line[0])
      {
        case 'I':
        case 'B':
          strip_first (dos_line);
          if (dos_line[0] == 'O' || dos_line[0] == 'P')
            widget->position.y = LOCATE_OVERLAY_BELOW;
          else
            widget->position.y = LOCATE_BELOW;
          break;

        case 'D':
        case 'A':
          strip_first (dos_line);
          if (dos_line[0] == 'O' || dos_line[0] == 'P')
            widget->position.y = LOCATE_OVERLAY_ABOVE;
          else
            widget->position.y = LOCATE_ABOVE;
          break;

        case 'S':
          widget->position.y = LOCATE_SAME;
          break;

        case 'P':
          widget->position.y = LOCATE_OVERLAY_BELOW;
          break;

        case 'Q':
          widget->position.y = LOCATE_OVERLAY_ABOVE;
          break;

        case 'C':
          strip_first (dos_line);
          if (dos_line[0] == 'O' || dos_line[0] == 'P')
            widget->position.y = LOCATE_OVERLAY_CENTER;
          else
            widget->position.y = LOCATE_CENTER;
          break;

        default:
          widget->position.y = (int)(factor + 0.5);
          break;
      }
      break;

    /** row (outdated) **/

    case 'r':

      widget->position.y = (int)(widget->spacing.y +
		factor*(widget->spacing.y + widget->size.y) + 0.5);
      break;

    /** SPACE or STRING **/

    case 'S':
      switch (line[1])
      {
        case 'P':
          if (dos_line[0] == 'X')
            widget->spacing.x = atoi (tres_line);
          else if (dos_line[0] == 'Y')
            widget->spacing.y = atoi (tres_line);
          break;

        default:
          strip_first (line);
          set_string (widget, line);
          break;
      }
      break;

    /** TEXT or THICK **/

    case 'T':

      switch (line[1])
      {
        case 'H':		/** THICK **/
          switch (dos_line[0])
          {
            case 'S':
              if (dos_line[1] == 'H')
                widget->thickness[THICKNESS_SHADOW] = atoi (tres_line);
              else
                widget->thickness[THICKNESS_SELECT] = atoi (tres_line);
              break;

            case 'A':
              widget->thickness[THICKNESS_ACTIVE] = atoi (tres_line);
              break;

            default:
              if (int_factor < MAX_THICKNESS)
                widget->thickness[int_factor] = atoi (tres_line);
          }
          break;

        case 'E':		/** TEXT **/
        default:

          if (widget->state_count >= MAX_STATES)
            break;

          strip_first (line);
          widget->labels[widget->state_count] = strdup (line);
          if (widget->state_count == 0)
            set_label (widget, line);
          widget->state_count++;
          break;
      }
      break;

    /** WIDTH **/

    case 'W':

      if (dos_line[0] == 'S')
        widget->size.x = DIMENSION_SAME;
      else if (dos_line[0] == 'T')
        widget->size.x = DIMENSION_TAFFY;
      else if (dos_line[0] == 'P')
        widget->size.x = DIMENSION_PANEL;
      else if (dos_line[0] < '0' || dos_line[0] > '9')
        widget->size.x = DIMENSION_SAME;
      else
        widget->size.x = (int)(factor+0.5);
      break;

    /** width (outdated) **/

    case 'w':

      widget->size.x = int_factor*Console->formats[0]->size.x +
			(int_factor-1)*Console->formats[0]->spacing.x;
      break;

    /** XOUT **/

    case 'X':
      widget->enable = FALSE;
      break;

    /** ZCOLOR **/

    case 'Z':

      if (dos_line[0] == 'P' || (dos_line[0] == 'B' && dos_line[1] == 'I'))
      {
        strcpy (quad_line, tres_line);
        strip_first (quad_line);
        gc_index = load_triplet (quad_line, &color);
      }
      else
        gc_index = load_triplet (tres_line, &color);

      for (i=0; gc_index<0 && i<Win->color_widget_limit; i++)
      {
        if (color.red == Win->color_widget[i].red &&
		color.green == Win->color_widget[i].green &&
		color.blue == Win->color_widget[i].blue)
          gc_index = i;
      }

      if (gc_index < 0)
      {
        Win->color_widget[Win->color_widget_limit].red = color.red;
        Win->color_widget[Win->color_widget_limit].green = color.green;
        Win->color_widget[Win->color_widget_limit].blue = color.blue;
        gc_index = Win->color_widget_limit++;

        strcpy (Win->spectrum_names[Win->spectrum_count], tres_line);
        Win->spectrum_indices[Win->spectrum_count] = gc_index;
        Win->spectrum_count++;
      }

      switch (dos_line[0])
      {
        case 'A':
          widget->gcs[GC_FONT_SECOND] = gc_index;
          break;

        case 'B':
          if (dos_line[1] == 'I')
          {
            pix_index = tres_line[0] - '0';
            if (widget->graphics[pix_index] == NULL)
              break;

            strip_first (tres_line);
            widget->graphics[pix_index]->gc = gc_index;
          }
          else
            widget->gcs[GC_BODY] = gc_index;
          break;

        case 'F':
          widget->gcs[GC_FONT_FIRST] = gc_index;
          break;

        case 'I':
          widget->gcs[GC_BODY_INV] = gc_index;
          break;

        case 'P':
          pix_index = tres_line[0] - '0';
          if (widget->graphics[pix_index] == NULL)
            break;

          strip_first (tres_line);
          widget->graphics[pix_index]->gc = gc_index;
          break;

        default:
          if (int_factor < MAX_GCS)
            widget->gcs[int_factor] = gc_index;
          break;
      }
      break;

    /** Widget comment (@) **/

    case '@':

      free (widget->comment);
      widget->comment = strdup (&line[1]);
      break;

    /** File comment (#,!,%) **/

    case '#':
    case '!':
    case '%':

      break;

    /** Bad widget parameter **/

    default:

      throw_error ("Invalid widget definition: %s", line);
      break;
  }
}



/******************************************************************************
  parse_entire_panel (panel, fp)

  Parse entire panel entry in <fp>, setting panel <panel> accordingly.
******************************************************************************/

parse_entire_panel (panel, fp)
  Panel *panel;
  FILE *fp;
{
  char line[200],
       *ptr;

  ptr=fgets (line, 200, fp);
  while (line[0] != ' ' && line[0] != '\n' && line[0] != '\0' && ptr != NULL)
  {
    parse_panel_configuration (panel, line);
    ptr=fgets (line, 200, fp);
  }
}



/******************************************************************************
  parse_panel_configuration (panel, line)

  Parse string given by <line>, setting <panel> accordingly.
******************************************************************************/

parse_panel_configuration (panel, line)
  Panel *panel;
  char *line;
{
  int int_factor;

  char dos_line[100],
       tres_line[100];

  double factor,
         atof();

  /** Set some convenience strings and values **/

  strcpy (dos_line, line);
  strip_first (dos_line);
  strcpy (tres_line, dos_line);
  strip_first (tres_line);

  factor = atof (dos_line);
  int_factor = atoi (dos_line);

  /** Determine parameter **/

  switch (line[0])
  {
    /** WIDTH **/

    case 'W':
    case 'w':

      panel->size.x = int_factor;
      break;

    /** HEIGHT **/

    case 'H':
    case 'h':

      panel->size.y = int_factor;
      break;

    /** COLUMN **/

    case 'C':
    case 'c':

      panel->position.x = int_factor;
      break;

    /** ROW **/

    case 'R':
    case 'r':

      panel->position.y = int_factor;
      break;

    /** OPEN **/

    case 'O':
    case 'o':

      panel->is_open = TRUE;
      panel->is_active = TRUE;
      break;

    /** File comments (#,!,%) **/

    case '#':
    case '!':
    case '%':

      break;

    /** Bad parameter **/

    default:

      throw_error ("Invalid panel definition: %s", line);
      break;
  }
}



/******************************************************************************
  load_widget_setup ()

  Open master widget setup file and load defaults values, colors, fonts, and
  such.
******************************************************************************/

load_widget_setup ()
{
  char line[200];

  FILE *fp, *fopen();

  /** Open the setup file, if possible **/

  if ((fp = fopen (Master->widget_setup_file, "r")) == NULL)
    return;

  Console->panel_count = 0;

  /** While there are lines left to read **/

  while (fgets (line, 100, fp) != NULL)
  {
    subload_widget_setup (line);
  }

  fclose (fp);
}



/******************************************************************************
  subload_widget_setup (line)

  Handle the actual loading of default values.
******************************************************************************/

subload_widget_setup (line)
  char *line;
{
  int i,
      gc_index,
      index;

  char dos_line[200],
       data_line[200],
       triplet_line[200],
       quad_line[200],
       dummy_line[200];

  double atof();

  XColor color;

  /** Skip over comments **/

  if (line[0] == '#' || line[0] == '@' || line[0]=='!')
    return;

  /** Set some convenience strings **/

  strcpy (dos_line, line);
  strip_first (dos_line);
  strcpy (data_line, dos_line);
  strip_first (data_line);
  strcpy (triplet_line, data_line);
  strip_first (triplet_line);

  /** Determine the setup command **/

  switch (line[0])
  {
    /** AUX **/

    case 'A':

      mini_parse (dos_line);
      break;

    /** COLOR **/

    case 'C':	
    case 'c':

      index = atoi (data_line);
      switch (dos_line[0])
      {
        /** REDEF **/

        case 'R':

          copy_first (dummy_line, data_line);
          gc_index = load_triplet (dummy_line, &color);
          if (gc_index < 0)
          {
            throw_warning ("Cannot redefine nonexistent color", NULL);
            break;
          }
          load_triplet (triplet_line, &(Win->color_widget[gc_index]));
          break;

        /** DEF **/

        case 'D':
          index = Win->color_widget_limit++;
          load_triplet (triplet_line, &(Win->color_widget[index]));

          copy_first (Win->spectrum_names[Win->spectrum_count], data_line);
          Win->spectrum_indices[Win->spectrum_count] = index;
          Win->spectrum_count++;
          break;

        /** USABLE **/

        case 'U':
          load_triplet (triplet_line, &(Win->color_usable[index]));
          break;

        /** ORABLE **/

        case 'O':
          load_triplet (triplet_line, &(Win->color_orable[index]));
          break;

        /** CONTROL **/
        /** WIDGET **/

        case 'C':
        case 'W':
          load_triplet (triplet_line, &(Win->color_widget[index]));
          break;
      }
      break;

    /** DEFAULT **/

    case 'D':
    case 'd':

      if (dos_line[0] == 'P' && dos_line[1] == 'U')
        push_formats ();
      else if (dos_line[0] == 'P' && dos_line[1] == 'O')
        pop_formats ();
      else
        widget_function (NULL, dos_line, FUNCTION_LOAD);
      break;

    /** FONT **/

    case 'F':
    case 'f':

      switch (dos_line[0])
      {
        /** REDEF **/

        case 'R':
          copy_first (dummy_line, data_line);
          index = load_typeset (dummy_line);
          if (index < 0)
          {
            throw_warning ("Cannot redefine nonexistent font", NULL);
            break;
          }
          free (Win->fonts[index]->fontname);
          Win->fonts[index]->fontname = strdup (triplet_line);
          break;

        /** DEF **/

        case 'D':
          index = Win->font_limit++;
          Win->fonts[index]->fontname = strdup (triplet_line);

          copy_first (Win->typeset_names[Win->typeset_count], data_line);
          Win->typeset_indices[Win->typeset_count] = index;
          Win->typeset_count++;
          break;

        /** Oldstyle index definition **/

        default:
          index = atoi (dos_line);
          if (index >= MAX_FONTS)
            break;

          free (Win->fonts[index]->fontname);
          Win->fonts[index]->fontname = strdup (data_line);
          break;
       }
       break;

    /** INFO or IF **/

    case 'I':

      strcpy (Console->panel_files[Console->panel_count++], dos_line);
      break;

    /** LIMIT **/

    case 'L':

      switch (dos_line[0])
      {
        case 'F':
          Win->font_limit = atoi (data_line);
          break;

        case 'O':
          Win->color_orable_limit = atoi (data_line);
          if (Win->color_orable_limit < COLOR_ORABLE_MIN_LIMIT)
            Win->color_orable_limit = COLOR_ORABLE_MIN_LIMIT;
          break;

        case 'U':
          Win->color_usable_limit = atoi (data_line);
          if (Win->color_usable_limit < COLOR_USABLE_MIN_LIMIT)
            Win->color_usable_limit = COLOR_USABLE_MIN_LIMIT;
          break;

        case 'C':
        case 'W':
          Win->color_widget_limit = atoi (data_line);
          if (Win->color_widget_limit < COLOR_CONTROL_MIN_LIMIT)
            Win->color_widget_limit = COLOR_CONTROL_MIN_LIMIT;
          break;
      }
      break;

    /** SPECIAL **/

    case 'S':
    case 's':

      widget_function (NULL, dos_line, FUNCTION_LOAD);
      break;
  }
}



/******************************************************************************
  parse_format_defaults (format, line)

  Parse the default commands for widget <format>, as store in string <line>.
******************************************************************************/

parse_format_defaults (format, line)
  Widget *format;
  char *line;
{
  int index,
      int_factor;

  char dos_line[100],
       tres_line[100],
       quattro_line[100];

  double factor,
         atof();

  /** Send the second argument onward to the normal parser **/

  strcpy (dos_line, line);
  strip_first (dos_line);
  parse_configuration (format, dos_line);
}



/******************************************************************************
  push_formats ()

  Pushes all formats into back-up storage for subsequent (temporary)
  reassigning of default values.
******************************************************************************/

push_formats ()
{
  int i;

  for (i=0; i<MAX_FORMATS; i++)
  {
    if (Console->formats[i]->format != CONTROL_X_NONE)
      copy_defaults (Console->back_formats[i], Console->formats[i]);
  }
}



/******************************************************************************
  pop_formats ()

  Pops all formats from back-up storage for restoration of temporary default
  value reassignments.
******************************************************************************/

pop_formats ()
{
  int i;

  for (i=0; i<MAX_FORMATS; i++)
  {
    if (Console->formats[i]->format != CONTROL_X_NONE)
      copy_defaults (Console->formats[i], Console->back_formats[i]);
  }
}


/** ALTER: probably doesn't work **/

/******************************************************************************
  update_widget_info (filename)

  Update all widgets from parameters stored in file <filename>.
******************************************************************************/

update_widget_info (filename)
  char *filename;
{
  /** ALTER: of specific panel index? **/

  int panel_index;

  char line[100];

  Widget *widget;
  
  FILE *fp;

  if ((fp = fopen (filename, "r")) == NULL)
    return;

  printf ("Updating button/slider states from: %s\n", filename); 
  
  panel_index = 0;

  while (fgets (line, 100, fp) != NULL)
  {
    strip_blank_first (line);
    if (line[0] == '\0' || line[0] == '\n')
    {
    }
    else if (line[0] == 'P' && line[1] == 'A')
    {
      strip_first (line);
      panel_index = atoi (line);
    }
    else
    {
      widget = get_named_widget (line);
      parse_configuration (widget, fp);
    }
  }

  draw_panel (Console->panel, TRUE);
  XSync (Win->display, 0);

  fclose (fp);
}



/** ALTER: probably doesn't work **/

/******************************************************************************
  dump_widget_info (filename)

  Dump all widget parameters to file <filename>.
******************************************************************************/

dump_widget_info (filename)
{
  int panel_index,
      widget_index;

  char line[2000];

  Widget *widget;
  
  FILE *fp, *fopen();

  if ((fp = fopen (filename, "w")) == NULL)
  {
    throw_warning ("Unable to open widget info storage file %s", filename);
    return;
  }

  printf ("Saving button/slider states as: %s\n", filename); 
  
  for (panel_index=0; panel_index<Console->panel_count; panel_index++)
  {
    if (panel_index != 0)
      fprintf (fp, "\n");
    fprintf (fp, "PANEL\t%d\n", panel_index);
    for (widget_index=0;
	widget_index<Console->panels[panel_index]->widget_count; widget_index++)
    {
      widget = Console->panels[panel_index]->widgets[widget_index];

      line[0] = '\0';
      widget_function (widget, line, FUNCTION_STORE);
      if (line[0] != '\0')
      {
        fprintf (fp, "\n");
        fprintf (fp, "%s\n", widget->name);
        fprintf (fp, "%s", line);
      }
    }
  }

  fclose (fp);
}
