#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"


/******************************************************************************
  char *
  load_data_line (line, fp)

  Keep reading through lines of file <fp> until a valid widget specifying
  command is encountered.  Put valid command in string <line> and return
  pointer to line.
******************************************************************************/

char *
load_data_line (line, fp)
  char line[];
  FILE *fp;
{
  char *ptr;

  ptr = fgets (line, 100, fp);
  while (ptr != NULL && (line[0] == ' ' || line[0] == '\0' || line[0] == '\n' ||
                   line[0] == '\t' || line[0] == '*' || line[0] == '#'))
    ptr = fgets (line, 100, fp);

  return (ptr);
}



/******************************************************************************
  load_triplet (line, color)

  Load a color triplet in string <line> into XColor pointer <color>, also
  handling parsable color names.
******************************************************************************/

load_triplet (line, color)
  char line[];
  XColor *color;
{
  int i, j,
      valid_color,
      ptr;
  Display *display;

  ptr = 0;

  /** Load color using X color tables **/

  if (line[0] < '0' || line[0] > '9')
  {
    for (i=0; i<Win->spectrum_count; i++)
    {
      if (!strcasecmp (line, Win->spectrum_names[i]))
      {
        if (Win->spectrum_indices[i] >= 0)
          return (Win->spectrum_indices[i]);
        else
        {
          color->red = Win->spectrum_defs[i][0] << 8;
          color->green = Win->spectrum_defs[i][1] << 8;
          color->blue = Win->spectrum_defs[i][2] << 8;
          return (-1);
        }
      }
    }

    display = XOpenDisplay (Win->display_name);
    if (XParseColor (display, DefaultColormap (display, DefaultScreen(display)),
		line, color))
      valid_color = TRUE;
    else
      valid_color = FALSE;
    XCloseDisplay (display);

    if (!valid_color)
    {
      color->red = 0 << 8; 
      color->green = 0 << 8; 
      color->blue = 0 << 8; 
    }

    return (-1);
  }

  /** Assign RGB colors **/

  for (i=0; i<3; i++)
  {
    for (; !(line[ptr] >= '0' && line[ptr] <= '9'); ptr++);
    if (i==0)
      color->red = atoi (&line[ptr])<<8;
    else if (i==1)
      color->green = atoi (&line[ptr])<<8;
    else 
      color->blue = atoi (&line[ptr])<<8;
    for (; line[ptr] != '\0' && line[ptr] >= '0' && line[ptr] <= '9'; ptr++);

    if (line[ptr] == '\0')
    {
      if (i == 0)
        return (atoi (line));
      else
        return (-1);
    }
  }

  return (-1);
}



/******************************************************************************
  load_typeset (line)

******************************************************************************/

load_typeset (line)
  char line[];
{
  int i;

  /** Check for typeset **/

  for (i=0; i<Win->typeset_count; i++)
  {
    if (!strcasecmp (line, Win->typeset_names[i]))
    {
      return (Win->typeset_indices[i]);
    }
  }

  return (-1);
}



/******************************************************************************
  strip_first (line)

  Strip the first non-white-space sequence from string <line>.
******************************************************************************/

strip_first (line)
  char line[];
{
  int i, j;

  for (i=0; line[i] != '\n' && line[i] != '\0' && line[i] != ' ' &&
			line[i]!= ',' && line[i] != '\t'; i++);
  for (;!(line[i] != '\n' && line[i] != ' ' && line[i] != ',' &&
			 line[i] != '\t'); i++);

  for (j=0; line[i] != '\n' && line[i] != '\0'; i++,j++)
    line[j] = line[i];
  line[j] = '\0';
}



/******************************************************************************
  strip_blank_first (line)

  Strip the first white-space sequence from string <line>.
******************************************************************************/

strip_blank_first (line)
  char line[];
{
  int i, j;

  if (line[0] == '\n' || line[0] == '\0')
  {
    line[0] = '\0';
    return;
  }

  for (i=0; line[i] == '\n' || line[i] == '\0' || line[i] == ' ' ||
						 line[i] == '\t'; i++);

  for (j=0; line[i] != '\n' && line[i] != '\0'; i++,j++)
    line[j] = line[i];
  line[j] = '\0';
}



/******************************************************************************
  copy_first (dest, src)

  Copy first non-white-space sequence from string <src> to string <dest>.
******************************************************************************/

copy_first (dest, src)
  char src[], dest[];
{
  int i;

  for (i=0; src[i] != ' ' && src[i] != '\t' && src[i] != '\0'
			&& src[i] != '\n'; i++)
    dest[i] = src[i];

  dest[i] = '\0';
}



/******************************************************************************
  strip_string (line)

  Strip the newline off the end of string <line>.
******************************************************************************/

strip_string (line)
  char line[];
{
  int i;

  for (i=0; line[i] != '\n' && line[i] != '\0' &&
				line[i] != ' ' && line[i] != '\t'; i++);

  line[i] = '\0';
}



/******************************************************************************
  throw_error (widget_string, parameter_string)

  Print an error message to stderr and exit.
******************************************************************************/

throw_error (widget_string, parameter_string)
  char *widget_string,
       *parameter_string;
{
  throw_message ("ERROR: ", widget_string, parameter_string);
  exit (0);
}



/******************************************************************************
  throw_warning (widget_string, parameter_string)

  Print a warning message to stderr.
******************************************************************************/

throw_warning (widget_string, parameter_string)
  char *widget_string,
       *parameter_string;
{
  throw_message ("WARNING: ", widget_string, parameter_string);
}



/******************************************************************************
  throw_message (type_string, widget_string, parameter_string)

  Print a generic message to stderr.
******************************************************************************/

throw_message (type_string, widget_string, parameter_string)
  char *type_string,
       *widget_string,
       *parameter_string;
{
  char full_string[200],
       temp_string[200];

  if (parameter_string == NULL)
    sprintf (full_string, "%s%s\n", type_string, widget_string);
  else
  {
    sprintf (temp_string, "%s%s\n", type_string, widget_string);
    sprintf (full_string, temp_string, parameter_string);
  }

  fprintf (stderr, full_string);
  fflush (stderr);
}

