#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "pip_constant.h"
#include "pip_extern.h"


/******************************************************************************
  open_panel (panel)
  
  Open <panel> if possible.
******************************************************************************/

open_panel (panel)
  Panel *panel;
{
  if (panel == NULL)
    throw_error ("Cannot open NULL panel", NULL);

  if (panel->window != (Window)(0))
  {
    XRaiseWindow (Win->display, panel->window);
    XSync (Win->display, 0);
    return;
  }

  open_derived_window (panel);
  init_widgets_window (panel);
  panel->is_open = TRUE;
}



/******************************************************************************
  close_panel (panel)
  
  Close <panel> if already open (and not base panel).
******************************************************************************/

close_panel (panel)
  Panel *panel;
{
  if (panel->window == (Window)(0))
    return;

  if (panel == NULL)
    throw_error ("Cannot close NULL panel", NULL);

  if (panel == Console->panels[0])
  {
    throw_warning ("Cannot close base window", NULL);
    return;
  }

  XDestroyWindow (Win->display, panel->window);
  XSync (Win->display, 0);

  panel->is_open = FALSE;
  panel->window = (Window)(0);
  init_widgets_window (panel);
}



/******************************************************************************
  open_all_windows ()
  
  Open all panels that are supposed to be open.
******************************************************************************/

open_all_windows ()
{
  int i;

  if (Console->panels[0]->is_open)
  {
    Console->panels[0]->window = Win->window;
    init_widgets_window (Console->panels[0]);
  }

  for (i=1; i<Console->panel_count; i++)
  {
    if (Console->panels[i]->is_open)
    {
      open_derived_window (Console->panels[i]);
      init_widgets_window (Console->panels[i]);
    }
  }
}



/******************************************************************************
  open_base_window ()
  
  Open the base window (for panels[0]), to which all GCs and such are attached.
******************************************************************************/

open_base_window ()
{
  int i, j,
      shift;

  unsigned long dummy;

  unsigned long valuemask;

  unsigned long plane_masks[4],
                orable_pixel,
                pixels[255];

  XVisualInfo vinfo;

  Pixmap stipple[5];

  int foreground,
      background,
      nplanes,
      actual_depth,
      junk;

  XCharStruct text;

  /** Get display **/

  Win->display = XOpenDisplay (Win->display_name);
  if (Win->display == NULL)
    throw_error ("Cannot open display %s", Win->display_name);

  Win->screen  = DefaultScreen (Win->display);

  /** Set foreground and background **/

  foreground = WhitePixel (Win->display, Win->screen);
  background = BlackPixel (Win->display, Win->screen);

  /** Set window dimensions **/

  if (Win->hint.x < 0 || Win->hint.y < 0)
    Win->hint.flags = ( PPosition | PSize);
  else
    Win->hint.flags = (USPosition | PSize);

  Win->depth = DefaultDepth (Win->display, Win->screen);
  actual_depth = Win->depth;

  /** Get visual **/

  Win->visual = DefaultVisual (Win->display, Win->screen);

  /** Try to force the display to 8 planes **/

  if (Win->depth != 8)
  {
    if (XMatchVisualInfo (Win->display, Win->screen, 8, PseudoColor, &vinfo))
    {
      Win->visual = vinfo.visual;
      Win->depth = 8;
    }
  }

  /** If couldn't find an 8 bit visual, try a 16 bit visual **/

  if (Win->depth != 8)
  {
    if (XMatchVisualInfo
                (Win->display, Win->screen, 16, PseudoColor, &vinfo))
    {
      Win->visual = vinfo.visual;
      Win->depth = 16;
    }
  }  

  /** If have less than 8 planes, just use a single plane **/

  if (Win->depth < 8)
  {
    throw_warning ("XPIP designed for full color systems.  Expect trouble",
				NULL);
    Win->depth = 1;
  }

  /** Open window **/

  if (Win->depth >= 8)
  {
    /** Set colormap **/

    if (Win->use_dedicated_colormap)
      Win->colormap = XCreateColormap
			(Win->display, DefaultRootWindow(Win->display),
			Win->visual, AllocNone);
    else if (actual_depth > 8)
    {
      /** XPIP won't work for 24 bit displays with no optional visuals	**/
      /** with 8 or 16 bits.  This may be the case on some PCs. In	**/
      /** these cases, it may be possible to reconfigure the display to	**/
      /** 8 or 16 bits.							**/

      if (Win->depth > 16)
      {
        throw_warning ("No PseudoColor visual available\n%s",
		"         Try changing display to <= 16 bits");
        throw_error ("Unable to continue without colormap", NULL);
      }

      Win->colormap = XCreateColormap
			(Win->display, DefaultRootWindow(Win->display),
			Win->visual, AllocNone);
    }
    else
      Win->colormap = DefaultColormap (Win->display, Win->screen);
    
    /** Set window attributes **/

    valuemask = CWBackPixel | CWBorderPixel | CWBitGravity | CWColormap;
    Win->attrib.background_pixel = background;
    Win->attrib.border_pixel = background;
    Win->attrib.bit_gravity = CenterGravity;
    Win->attrib.colormap = Win->colormap;

    /** Create the window **/

    Win->window = XCreateWindow (Win->display,
		RootWindow (Win->display, Win->screen),
		Win->hint.x, Win->hint.y,
		Win->hint.width, Win->hint.height, 0,
		Win->depth, InputOutput, Win->visual, valuemask, &Win->attrib);
  }
  else
  {
    Win->window = XCreateSimpleWindow (Win->display,
		DefaultRootWindow (Win->display),
		Win->hint.x, Win->hint.y,
		Win->hint.width, Win->hint.height, 0,
		background,
		foreground);
  }


  if (Win->depth >= 8)
  {
    XSetStandardProperties (Win->display, Win->window, Win->window_title,
			Win->window_title, None, NULL, None, &Win->hint);
  }
  else if (Win->depth == 1)
    XSetStandardProperties (Win->display, Win->window, Win->window_title,
			Win->window_title, None, NULL, None, &Win->hint);

  /** Set window manager hints **/

  Win->xwmh.flags = (InputHint|StateHint);  
  Win->xwmh.input = TRUE;
  Win->xwmh.initial_state = NormalState;
  XSetWMHints (Win->display, Win->window, &(Win->xwmh));

  /** make window sensitive to specified events **/

  if (Master->use_motion_highlights)
    XSelectInput (Win->display, Win->window,
		ButtonPressMask|ButtonReleaseMask|ButtonReleaseMask|
		Button1MotionMask|Button2MotionMask|Button3MotionMask|
		KeyPressMask|ExposureMask|PointerMotionMask|StructureNotifyMask);
  else
    XSelectInput (Win->display, Win->window,
		ButtonPressMask|ButtonReleaseMask|ButtonReleaseMask|
		Button1MotionMask|Button2MotionMask|Button3MotionMask|
		KeyPressMask|ExposureMask|StructureNotifyMask);

  /** Load color map **/

  if (Win->depth >= 8)
  {
    if (Win->color_orable_limit)
    {
      if (Win->color_orable_limit < 3)
        nplanes = 1;
      else if (Win->color_orable_limit < 5)
        nplanes = 2;
      else if (Win->color_orable_limit < 9)
        nplanes = 3;
      else
        nplanes = 4;

      if (!XAllocColorCells (Win->display, Win->colormap,
			TRUE, plane_masks, nplanes, &orable_pixel, 1))
      {
        show_alloc_error ();
      }

      for (shift=0; plane_masks[0] != (1<<shift); shift++);

      for (i=0; i<Win->color_orable_limit; i++)
      {
        Win->color_orable[i].pixel = (i<<shift) | orable_pixel;
        Win->color_orable[i].flags = DoRed | DoGreen | DoBlue;
      }
      XStoreColors (Win->display, Win->colormap,
				Win->color_orable, Win->color_orable_limit);
    }

    if (!XAllocColorCells (Win->display, Win->colormap,
			FALSE, NULL, 0, pixels, Win->color_widget_limit))
    {
      show_alloc_error ();
    }

    for (i=0; i<Win->color_widget_limit; i++)
    {
      if (i==COLOR_CTRL_SURFACE && Win->color_orable_limit)
        Win->color_widget[i].pixel = Win->color_orable[0].pixel;
      else
        Win->color_widget[i].pixel = pixels[i];

      Win->color_widget[i].flags = DoRed | DoGreen | DoBlue;
    }
    XStoreColors (Win->display, Win->colormap,
				Win->color_widget, Win->color_widget_limit);

    if (!XAllocColorCells (Win->display, Win->colormap,
			FALSE, NULL, 0, pixels, Win->color_usable_limit))
    {
      show_alloc_error ();
    }

    for (i=0; i<Win->color_usable_limit; i++)
    {
      Win->color_usable[i].pixel = pixels[i];
      Win->color_usable[i].flags = DoRed | DoGreen | DoBlue;
    }
    XStoreColors (Win->display, Win->colormap,
				Win->color_usable, Win->color_usable_limit);
  }

  /** Raise window **/

  XMapRaised(Win->display, Win->window);

  /** Set orable GCs **/

  for (i=1; i<Win->color_orable_limit; i++)
  {
    Win->hue_orable[i] = XCreateGC (Win->display, Win->window, 0, 0);
    if (i==0)
      XSetFunction (Win->display, Win->hue_orable[i], GXcopy);
    else
      XSetFunction (Win->display, Win->hue_orable[i], GXor);
    XSetGraphicsExposures (Win->display, Win->hue_orable[i], FALSE);
  }

  /** Set widget GCs **/

  for (i=0; i<Win->color_widget_limit; i++)
  {
    Win->hue_widget[i] = XCreateGC (Win->display, Win->window, 0, 0);
    XSetFunction (Win->display, Win->hue_widget[i], GXcopy);
    XSetGraphicsExposures (Win->display, Win->hue_widget[i], FALSE);
  }

  /** Set usable GCs **/

  for (i=0; i<Win->color_usable_limit; i++)
  {
    Win->hue_usable[i] = XCreateGC (Win->display, Win->window, 0, 0);
    XSetFunction (Win->display, Win->hue_usable[i], GXcopy);
    XSetGraphicsExposures (Win->display, Win->hue_usable[i], FALSE);
  }

  /** Set foreground and background GCs for all colors **/

  if (Win->depth == 1)
  { 
    for (i=0; i<5; i++)
      stipple[i] = XCreateBitmapFromData (Win->display, Win->window,
					Grey_Palette[i], 8, 8);

/**
    XSetStipple (Win->display,
		Win->hue_widget[COLOR_CTRL_BACK], stipple[0]);
    XSetFillStyle (Win->display,
		Win->hue_widget[COLOR_CTRL_BACK], FillOpaqueStippled);
**/

    for (i=0; i<Win->color_usable_limit; i++)
    {
      XSetForeground (Win->display, Win->hue_usable[i], foreground);
      XSetBackground (Win->display, Win->hue_usable[i], background);
    }
  }
  else if (Win->depth >= 8)
  {
    for (i=1; i<Win->color_orable_limit; i++)
    {
      XSetForeground (Win->display,
		Win->hue_orable[i], Win->color_orable[i].pixel);
      XSetBackground (Win->display,
		Win->hue_orable[i], Win->color_orable[0].pixel);
    }

    for (i=0; i<Win->color_widget_limit; i++)
    {
      XSetForeground (Win->display,
		Win->hue_widget[i], Win->color_widget[i].pixel);
      XSetBackground (Win->display, Win->hue_widget[i],
		Win->color_widget[COLOR_CTRL_SURFACE].pixel);
    }

    for (i=0; i<Win->color_usable_limit; i++)
    {
      XSetForeground (Win->display,
		Win->hue_usable[i], Win->color_usable[i].pixel);
      XSetBackground (Win->display, Win->hue_usable[i], 
			 Win->color_usable[0].pixel);
    }
  }

  /** Load fonts **/

  for (i=0; i<Win->font_limit; i++)
    Win->fonts[i]->font =
	 XLoadQueryFont (Win->display, Win->fonts[i]->fontname);

  /** Determine font dimensions **/

  for (i=0; i<Win->font_limit; i++)
  {
    if (Win->fonts[i]->font == 0)
    {
      throw_warning ("Font %s not found", Win->fonts[i]->fontname);
      Win->fonts[i]->font = XQueryFont (Win->display,
			XGContextFromGC(DefaultGC(Win->display,Win->screen)));
    }
    if (i==0)
      XSetFont (Win->display, Win->hue_usable[1], Win->fonts[i]->font->fid);
    else if (i==1)
      XSetFont (Win->display, Win->hue_usable[3], Win->fonts[i]->font->fid);
    else if (i==2)
      XSetFont (Win->display, Win->hue_usable[2], Win->fonts[i]->font->fid);
    else 
      XSetFont (Win->display, Win->hue_usable[2], Win->fonts[i]->font->fid);

    XTextExtents (Win->fonts[i]->font, "T", 1, &junk, &junk, &junk, &text);
 
    Win->fonts[i]->dimension.x = text.width;
    Win->fonts[i]->dimension.y = text.ascent;
    Win->fonts[i]->is_fixed_width = TRUE;

    XTextExtents (Win->fonts[i]->font, "r", 1, &junk, &junk, &junk, &text);
    if (Win->fonts[i]->dimension.x != text.width)
      Win->fonts[i]->is_fixed_width = FALSE;

    XTextExtents (Win->fonts[i]->font, ",", 1, &junk, &junk, &junk, &text);
    if (Win->fonts[i]->dimension.x != text.width)
      Win->fonts[i]->is_fixed_width = FALSE;

    if (Win->fonts[i]->dimension.y <= 0)
      Win->fonts[i]->dimension.y = 8;

    if (Win->fonts[i]->dimension.x <= 0)
      Win->fonts[i]->dimension.x = 8;
  }

  /** Map window to make it visible **/

  if (Win->depth >= 8)
    XMapWindow (Win->display, Win->window);

  XFlush (Win->display);
}



/******************************************************************************
  open_derived_window (panel)
  
  Open the derived window specified by <panel>.
******************************************************************************/

open_derived_window (panel)
  Panel *panel;
{
  unsigned long valuemask;
  XSizeHints hint;

  /** Set window dimenions **/

  hint.x = panel->position.x;
  hint.y = panel->position.y;
  hint.width = panel->size.x;
  hint.height = panel->size.y;

  if (hint.x < 0 || hint.y < 0)
    hint.flags = ( PPosition | PSize);
  else
    hint.flags = (USPosition | PSize);

  valuemask = CWBackPixel | CWBorderPixel | CWBitGravity | CWColormap;

  /** Create window **/

  panel->window = XCreateWindow (Win->display,
		RootWindow (Win->display, Win->screen),
		hint.x, hint.y,
		hint.width, hint.height, 0,
		Win->depth, InputOutput, Win->visual, valuemask, &Win->attrib);

  XSetStandardProperties (Win->display, panel->window,
			panel->name, panel->name,
			None, NULL, None, &hint);

  XSetWMHints (Win->display, panel->window, &(Win->xwmh));

  /** Make sensitive to specific events **/

  XSelectInput (Win->display, panel->window,
		ButtonPressMask|ButtonReleaseMask|ButtonReleaseMask|
		Button1MotionMask|Button2MotionMask|Button3MotionMask|
		KeyPressMask|ExposureMask);

  /** Raise and map window **/

  XMapRaised(Win->display, panel->window);

  XMapWindow (Win->display, panel->window);

  XFlush (Win->display);
}



/******************************************************************************
  show_alloc_error ()
  
  Print color allocation errors to stderr and exit.
******************************************************************************/

show_alloc_error ()
{
  throw_warning ("Could not allocate all color cells\n%s",
	"         Kill other colorful applications or run with -pip_colormap");
  throw_error ("Insufficient colormap resources", NULL);
}
