#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

/** Mandatory include files **/

#include "pip_constant.h"
#include "pip_extern.h"


/******************************************************************************
  file_run ()
******************************************************************************/

file_run ()
{
  check_file_widgets ();
}


/******************************************************************************
  check_file_widgets ()

  Check the file panel widgets for user control, updating display when
  necessary.
******************************************************************************/

check_file_widgets ()
{
  int is_ok,
      is_cancel;

  Panel *panel;

  Widget *widget,
         *widget_callback;

  if (check_named_widget_flag (&widget, "file_list_text"))
    file_change_file ();

  if (check_named_widget_flag (&widget, "directory_list_text"))
    file_change_directory (TRUE);

  if (check_named_widget_flag (&widget, "directory_data"))
    file_change_directory (FALSE);

  if (check_named_widget_flag (&widget, "suffix_data"))
    file_change_directory (FALSE);

  if (check_named_widget_flag (&widget, "ok_button"))
    is_ok = TRUE;
  else
    is_ok = FALSE;

  if (check_named_widget_flag (&widget, "cancel_button"))
    is_cancel = TRUE;
  else
    is_cancel = FALSE;

  if (is_ok || is_cancel)
  {
    widget = get_named_widget ("ok_button");

    close_panel (Console->event_panel);

    /** Send the appropriate event to the control widget **/

    widget_callback = widget->relatives[0];

    if (widget_callback != NULL)
    {
      if (is_ok)
        widget_callback->user = TRUE;
      else
        widget_callback->user = FALSE;

      Win->event.xbutton.x = widget_callback->position.x + 1;
      Win->event.xbutton.y = widget_callback->position.y + 1;
      Win->event.xexpose.window = widget_callback->panel->window;

      Win->event.type = ButtonPress;
      XSendEvent (Win->display, widget_callback->panel->window, TRUE,
                        ButtonPressMask, &(Win->event));
      Win->event.type = ButtonRelease;
      XSendEvent (Win->display, widget_callback->panel->window, TRUE,
                        ButtonReleaseMask, &(Win->event));
      XSync (Win->display, 0);

      widget->relatives[0] = NULL;
    }
  }

  reset_all_flags ();
}



/******************************************************************************
  file_change_file ()

  Handle selection of a file from the file list widget.
******************************************************************************/

file_change_file ()
{
  char string[200];
  Widget *file_list_widget;

  file_list_widget = get_named_widget ("file_list_text");

  list_get_entry (file_list_widget, string, -1);
  set_named_widget_label ("file_data", string);
}



/******************************************************************************
  file_change_directory (use_update)

  Handle selection of a directory from the directory list widget or setting of
  new directory or search suffix, updating both the file and directory lists.
  If <use_update>, user has selected directory list entry.
******************************************************************************/

file_change_directory (use_update)
  int use_update;
{
  int i,
      file_count,
      directory_count,
      length;

  char line[200],
       string[200],
       command[500],
       *file_strings[500],
       *directory_strings[500],
       *strdup();

  FILE *pp;

  Widget *file_widget,
         *file_list_widget,
         *directory_widget,
         *directory_list_widget,
         *suffix_widget;

  /** Get all the necessary widget pointers **/

  file_widget = get_named_widget ("file_data");
  file_list_widget = get_named_widget ("file_list");
  directory_widget = get_named_widget ("directory_data");
  directory_list_widget = get_named_widget ("directory_list");
  suffix_widget = get_named_widget ("suffix_data");

  /** If user selected directory list entry **/

  if (use_update)
  {
    /** Get the new directory **/

    list_get_entry (directory_list_widget, line, -1);

    /** If new directory is ".." for up-a-level **/

    if (strcmp (line, "..") == 0)
    {
      /** If at current level, directly set "..", else see if we need	**/
      /** to erase old directory.					**/

      if (directory_widget->label[1] == '\0')
        sprintf (string, "../");
      else
      {
        /** Get the current full directory name **/

        sprintf (string, "%s", line);
        strcpy (line, directory_widget->label);
        length = strlen (line) - 2;

        /** If last entry is a "..", just add another **/

        if (line[length] == '.')
          sprintf (string, "%s../", directory_widget->label);
        else
        {
          /** Else we need to erase a directory entry **/

          for (i=length; i>=0 && line[i] != '/'; i--);

          /** If back at current level, just set to ".", else add "../" **/

          if (i <= 1)
            sprintf (string, ".");
          else
          {
            line[i] = '\0';
            sprintf (string, "%s/", line);
          }
        }
      }
    }
    else if (directory_widget->label[1] == '\0')
      sprintf (string, "%s", line);
    else
      sprintf (string, "%s%s", directory_widget->label, line);

    /** Set the full directory name **/

    set_widget_label (directory_widget, string);
  }

  /** Open a command pipe for directory list **/

  sprintf (command, "/bin/ls -F %s", directory_widget->label);

  if ((pp = popen (command, "r")) == NULL)
    throw_error ("Cannot open /bin/ls pipe", NULL);

  file_count = 0;
  directory_count = 0;
  directory_strings[directory_count++] = strdup ("..");

  /** For each /bin/ls output line, add to directory or file list **/

  while (fgets (line, 200, pp) != NULL)
  {
    length = strlen (line);  
    line[length-1] = '\0';

    if (length > 1 && line[length-2] == '/')
      directory_strings[directory_count++] = strdup (line);
    else
    {
      if (length > 1 && line[length-2] == '*')
        line[length-2] == '\0';

      if (file_match_any_template (line, suffix_widget->label))
        file_strings[file_count++] = strdup (line);
    }
  }

  /** Load directory and file lists **/

  list_load (file_list_widget, file_strings, file_count, TRUE, FALSE);
  list_load (directory_list_widget,
		directory_strings, directory_count, TRUE, FALSE);

  /** Free temporary storage **/

  for (i=0; i<file_count; i++)
    free (file_strings[i]);
  for (i=0; i<directory_count; i++)
    free (directory_strings[i]);
}



/******************************************************************************
  file_init_widgets (directory, suffix)

  Initialize the file panel directory and suffix widgets, then use these fields
  to initialize entire panel.
******************************************************************************/

file_init_widgets (directory, suffix)
  char *directory,
       *suffix;
{
  Panel *panel;

  panel = get_named_panel ("file_panel");
  set_active_panel (panel);

  set_named_widget_label ("directory_data", directory);
  set_named_widget_label ("suffix_data", suffix);
  file_change_directory (FALSE);

  restore_panel ();
}



/******************************************************************************
  file_set_callback_widget (widget)

  Set "callback" <widget> to be activated when OK button is pressed.
******************************************************************************/

file_set_callback_widget (widget)
  Widget *widget;
{
  Panel *panel;
  Widget *widget_ok;

  panel = get_named_panel ("file_panel");
  set_active_panel (panel);

  widget_ok = get_named_widget ("ok_button");
  widget_ok->relatives[0] = widget;

  restore_panel ();
}



/******************************************************************************
  file_get_filename (string, filename, directory)

  Get just the filename of the currently active file from the file panel (not
  the directory).
******************************************************************************/

file_get_filename (fullpath, filename, directory)
  char *fullpath,
       *filename,
       *directory;
{
  char name_directory[200],
       name_file[200],
       string[400];
  Panel *panel;

  panel = get_named_panel ("file_panel");
  set_active_panel (panel);

  get_named_widget_label ("file_data", name_file, 200);
  get_named_widget_label ("directory_data", name_directory, 200);

  if (name_directory[1] == '\0')
    sprintf (string, "%s", name_file);
  else
    sprintf (string, "%s%s", name_directory, name_file);

  if (fullpath) 
    strcpy (fullpath, string);

  if (filename) 
    strcpy (filename, name_file);

  if (directory) 
    strcpy (directory, name_directory);

  restore_panel ();
}



/******************************************************************************
  int
  file_match_any_template (string, full_template)

  Determine if <string> matches any of the wildcards given in <template>,
  returning TRUE if it does, else returning FALSE.
******************************************************************************/

int
file_match_any_template (string, full_template)
  char *string,
       *full_template;
{
  int i;

  char *template_ptr,
       template[100];

  template_ptr = full_template;

  while (1)
  {
    for (i=0; template_ptr[i] != ',' &&
		template_ptr[i] != ' ' && template_ptr[i] != '\0'; i++)
      template[i] = template_ptr[i];

    template[i] = '\0';

    if (file_match_single_template (string, template))
      return (TRUE);

    if (template_ptr[i] == ',')
      template_ptr += i+1;
    else
      return (FALSE);
  }
}



/******************************************************************************
  int
  file_match_single_template (string, template)

  Determine if <string> matches the wildcard <template>, returning TRUE if
  it does, else returning FALSE.
******************************************************************************/

int
file_match_single_template (string, template)
  char *string,
       *template;
{
  int i,
      length,
      allow_skip;
  char ch,
       matcher[2],
       *string_ptr,
       *strstr();

  string_ptr = string;
  allow_skip = FALSE;

  /** Cut off comments after first space in template **/

  for (i=0; template[i] != ' ' && template[i] != '\0'; i++);
  template[i] = '\0';

  length = strlen(template);

  /** For each character of template **/

  for (i=0; i<length; i++)
  {
    ch = template[i];

    /** If a generic wildcard, go to next character **/

    if (ch == '*')
    {
      allow_skip = TRUE;
      continue;
    }

    /** If last template character was wildcard **/

    if (allow_skip)
    {
      /** Determine if template character occurs in remaining string **/

      matcher[0] = ch;
      matcher[1] = '\0';
      string_ptr = strstr (string, matcher);

      if (string_ptr == NULL)
        return (FALSE);
    }

    /** If characters matach, keep going, else return FALSE **/

    if (string_ptr[0] == ch)
      string_ptr++;
    else
      return (FALSE);

    /** If at the end of string **/

    if (string_ptr[0] == '\0')
    {
      if (i != length-1)
        return (FALSE);
      else
        return (TRUE);
    }

    allow_skip = FALSE;
  }

  /** Only return TRUE if last template character was wildcard **/

  if (allow_skip)
    return (TRUE);
  else
    return (FALSE);
}



/******************************************************************************
  file_init ()
******************************************************************************/

file_init ()
{
}



/******************************************************************************
  file_exit ()
******************************************************************************/

file_exit ()
{
}
